unit FIDUnit;

interface

uses
  Graphics,ExtCtrls,sysUtils,classes,math,
 // GraphUtil,  // Gradient drawing
  StdCtrls, // for Tlabel
  ComCtrls; // for TProgressBar

type TplotMode = (pmNormal,pmPolar);
type Tprecision = (prSingle, prDouble);
type TsaveMode = (smAsci, smBinary, smBoth, smAbs);

type TFIDPart = (FIDReal,FIDImag,FIDAbs);
type TPlotFIDOption = set of TFIDPart;


//type TReturnValues = set of (X,Y,deltaX,deltaY,integral,
//                      average,standardDeviation);


type TDomain = (dmTime, dmFrequency);



//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  type THalfFID = class
    private
      FAL: cardinal;
      procedure setAL(a:cardinal);

    public
      sig: array of double;

      procedure initialize;
      function Max: double; overload;
      function Max(kini,kfin:integer): double; overload;
      function Min: double; overload;
      function Min(kini,kfin:integer): double; overload;
      function AbsMax: double; overload;
      function AbsMax(kini,kfin:integer):double; overload;

      function AbsMaxPoint(kini,kfin:integer):integer;
      function Average(kini,kfin:integer):double; overload;
      function Average: double;overload;
      function Sum(kini,kfin:integer): double;overload;
      function Sum: double;overload;

      function standardDeviation:double;overload;
      function standardDeviation(kini,kfin:integer):double;overload;

      procedure DCOffsetCut;
      procedure squareRoot;


      constructor create(k:cardinal);
      destructor free;

      property AL: cardinal read FAL write setAL;

    end;
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   type TFID=class
     private
       FAL: cardinal;
       FDW: single;

     //  FXCursor:cardinal;
     //  procedure setXCursor(a:cardinal);
       procedure setAL(a:cardinal);
       procedure setDW(r:single);
//       procedure polarDraw(ti:TPaintBox);



     public
       real:THalfFID;
       imag:THalfFID;
       abs:THalfFID;

       domain:TDomain;

       phase0:double;
       pivot: cardinal;   // P̈ʑ킹̒S_

       actualNA: integer; // 'actual' number of scans
    //   fileStream:TFileStream;
    //   property xCursor:cardinal read FXCursor write setXCursor;

       function Max:double; overload;
       function Max(kini,kfin:integer):double; overload;

       function Min:double; overload;
       function Min(kini,kfin:integer):double; overload;

       function AbsMax:double; overload;
       function AbsMax(kini,kfin:integer):double; overload;

       function phase(k:integer):double;
       function averagePhase:double;overload;
       function averagePhase(kini,kfin:integer):double;overload;
       function phaseStandardDeviation:double;overload;
       function phaseStandardDeviation(kini,kfin:integer):double;overload;

       function averageIntensity:double;overload;
       function averageIntensity(kini,kfin:integer):double;overload;
       function intensityStandardDeviation:double;overload;
       function intensityStandardDeviation(kini,kfin:integer):double;overload;

       function intensityCount(i1,i2:double):integer;overload;
       function intensityCount(i1,i2:double; kini,kfin:integer):integer;overload;

       property AL: cardinal read FAL write setAL;
       property DW: single read FDW write setDW;

       procedure initialize;

       procedure updateAbs;
       procedure rotate(phi_deg:double);
       procedure FFT;
       procedure inverseFFT;

       procedure LT;
       procedure InverseLT;

//       procedure draw(ti:TPaintBox); overload;
//       procedure draw(ti:TPaintBox; plotMode:TplotMode);overload;

       constructor create(k:cardinal);
       destructor free;
   end;

  type FpenStyle=(line,circle);
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  type TFID_2D = class
    private
      FAL: cardinal;
      FDW: single;
      FSF1: single;
      FAcqPhase: single;
      oldx: integer;
      FAxisFontSize:integer;

      procedure cartesianDraw(ti:TPaintBox; k:cardinal);
      procedure halfFIDDraw(hFID:THalfFID;ti:TPaintBox;color:TColor;
                                 dx:single; vCenter:word; scale:single);

      procedure polarDraw(ti:TPaintBox; k:cardinal);
      procedure setAL(a:cardinal);
      procedure setSF1(s:single);
      procedure setDW(r:single);
      procedure setAcqPhase(p:single);

    public
      lMargin, rMargin, tMargin, bMargin:word;
      cursorOn,cursorWasOff: boolean;
      busy: boolean;
//      resized: boolean;

      length_2D: cardinal;
      FID: array of TFID;

      Fxini:cardinal;
      Fxfin:cardinal;
      FxRange:integer;
      xPointLeft: cardinal;
      latestClickedXPosition: cardinal;
      currentFID: cardinal;  // ݕ\ijFID͉Ԗڂ̃f[^H

      FvOffset:single;

      pivot: cardinal;   // P̈ʑ킹̒S_

      cursorX: word;
      autoscale: boolean;
      receiverScale: boolean;
      repeatScan:boolean;

      plotScale:single;
      plotOffset:single;
      plotOption: TPlotFIDOption;
      plotMode: TplotMode;

      saveMode: TSaveMode;
      saveReImSeparately: boolean;

      fileName: string;

      backgroundColor:TColor;
      realColor:TColor;
      imagColor:TColor;
      absColor:TColor;
      polarPlotColor:TColor;
      cursorColor:TColor;
      axisWidth:integer;
      axisColor:TColor;
      penWidth:integer;
      cursorPenWidth:integer;
      canvasFontName:string;
      canvasFontStyle:TFontStyles;
      canvasFontSize:integer;
      canvasFontColor:TColor;
      axisFontName: string;
      axisFontStyle:TFontStyles;
      axisFontSize:integer;
      axisFontColor:TColor;
      penStyle:FpenStyle;

      procedure setAxisFontSize(k:integer);
      property axisFonrSize:integer read FAxisFontSize write setAxisFontSize;
      procedure setXini(x:cardinal);
      procedure setXfin(x:cardinal);
      property xini:cardinal read Fxini write setXini;
      property xfin:cardinal read Fxfin write setXfin;
      property xRange:integer read FXrange;
      procedure setVoffset(v:single);
      property vOffset:single read Fvoffset write setVOffset;


      procedure add;
      procedure delete(k:cardinal);
      procedure SaveToFile(fn:string);
      procedure Readsm2pFile(fn:string);
      procedure Readsm2dFile(fn:string);overload;
      procedure Readsm2dFile(fn:string;progress:TProgressBar);overload;

      procedure ReadAvanceParFile(fn:string);
      procedure ReadAvanceDataFile(fn:string);overload; // test
      procedure ReadAvanceDataFile(fn:string;progress:TProgressBar);overload; // test

      procedure ReadVnmrFile(fn:string);   // not complete


      procedure drawCursor(ti:TPaintBox);
      procedure overWriteCursor(ti:TPaintBox);


      procedure draw(ti:TPaintBox; k:cardinal);

      procedure swap(i,j: cardinal);
      procedure reverse;



      function horizontalFFT(k:cardinal):integer; overload;
      function horizontalFFT:integer; overload;
      function longitudinalFFT(k:cardinal):integer; overload;
      function longitudinalFFT:integer; overload;

      constructor create(r:single; k:cardinal); // r <- dw, k <- FAL
      destructor free;
      property  al:cardinal read FAL write setAL;
      property  dw:single read FDW write setDW;
      property sf1: single read FSF1 write setSF1;
      property acqPhase: single read FAcqPhase write setAcqPhase;
  end;






  TFIDFFT=class
  public
    error: boolean;
    errorMessage: string;
    function Process(FID:TFID):integer; overload;
    function Process(FID2D:TFID_2D):integer; overload;
  end;

  TFIDLongitudinalFFT=class
  public
    error: boolean;
    errorMessage: string;
    function Process(FID2D:TFID_2D):integer;
  end;



  TFIDZeroFill=class
  public
    error: boolean;
    errorMessage: string;
    points:cardinal;
    function Process(FID:TFID):integer; overload;
    function Process(FID2D:TFID_2D):integer; overload;
  end;

  TFIDLongitudinalZeroFill=class
  public
    error: boolean;
    errorMessage: string;
    points:cardinal;
    function Process(FID2D:TFID_2D):integer;
  end;

  TFIDCovariance=class
  public
    error: boolean;
    errorMessage: string;
    function Process(FID2D:TFID_2D):integer;
  end;

  TFIDCrossCorrelation=class
  public
    error: boolean;
    errorMessage: string;
    function Process(FID2D:TFID_2D):integer;
  end;

  TFIDSort_riri_To_rrii=class
  public
    error: boolean;
    errorMessage: string;
    function Process(FID2D:TFID_2D):integer;
  end;







//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
implementation

//------------------------------------------------------------------------------
procedure TFID_2D.setAcqPhase(p: Single);
begin
  FAcqPhase:=p;
end;
//------------------------------------------------------------------------------
procedure TFID_2D.setSF1(s:single);
begin
  FSF1:=s;
end;
//------------------------------------------------------------------------------
procedure TFID_2D.setAL(a:cardinal);
var k:cardinal;
begin
  for k:=0 to Length_2D-1 do begin
    FID[k].AL:=a;
  end;
  FAL:=a;
end;
//------------------------------------------------------------------------------
procedure TFID_2D.setDW(r:single);
var k:cardinal;
begin
  for k:=0 to Length_2D-1 do begin
    FID[k].DW:=r;
  end;
  FDW:=r;
end;
//------------------------------------------------------------------------------
function TFIDSort_riri_To_rrii.Process(FID2D:TFID_2D):integer;
var auxFID2D:TFID_2D;
    i,j,k:cardinal;
begin

    // łȂƂ߁I
  if FID2D.length_2D mod 2 <> 0 then begin
    result:=1; error:=true;
    errorMessage:= 'Array length must be an even number in '
        +'re,im,re,im,... -> re,re,...im,im,... sorting.';
    exit;
  end;


  auxFID2D:=TFID_2D.create(1,FID2D.al);



    i:=0; k:=0;
    while i<FID2D.length_2D do begin
      for j:=0 to FID2D.al-1 do begin
        auxFID2D.add;
        auxFID2D.FID[k].real.sig[j]:=FID2D.FID[i].real.sig[j];
        auxFID2D.FID[k].imag.sig[j]:=FID2D.FID[i].imag.sig[j];
        auxFID2D.FID[k].abs.sig[j]:=FID2D.FID[i].abs.sig[j];
      end; //j
      k:=k+1;
      i:=i+2;
    end;  // while

    i:=1;
    while i<FID2D.length_2D do begin
      for j:=0 to FID2D.al-1 do begin
        auxFID2D.add;
        auxFID2D.FID[k].real.sig[j]:=FID2D.FID[i].real.sig[j];
        auxFID2D.FID[k].imag.sig[j]:=FID2D.FID[i].imag.sig[j];
        auxFID2D.FID[k].abs.sig[j]:=FID2D.FID[i].abs.sig[j];
      end; // j
      k:=k+1;
      i:=i+2;
    end;  // while

    for i:=0 to FID2D.al-1 do begin
      for j:=0 to FID2D.al-1 do begin
        FID2D.FID[i].real.sig[j]:=auxFID2D.FID[k].real.sig[j];
        FID2D.FID[i].imag.sig[j]:=auxFID2D.FID[k].imag.sig[j];
        FID2D.FID[i].abs.sig[j]:=auxFID2D.FID[k].abs.sig[j];
      end;
    end;

  auxFID2D.free;
  error:=false;
  errorMessage:='';
  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDCrossCorrelation.Process(FID2D:TFID_2D):integer;
var r: array of array of double;
    i,j:cardinal;
begin
  //
  //  ŝݎ󂯕tBCovariance͕KȂĂ͂B
  //
  if FID2D.al<>FID2D.length_2D then begin
    result:=1; error:=true;
    errorMessage:='Cross correlation requires the data to be nxn matrix.';
    exit;
  end;

  setLength(r,FID2D.al);
  for i:=0 to FID2D.length_2D-1 do setLength(r[i],FID2D.length_2D-1);
  for i:=0 to FID2D.length_2D-1 do for j:=0 to FID2D.length_2D-1 do
    r[i][j]:= FID2D.FID[i].real.sig[j]
                / sqrt(FID2D.FID[i].real.sig[i] * FID2D.FID[j].real.sig[j]);

  for i:=0 to FID2D.length_2D-1 do for j:=0 to FID2D.length_2D-1 do
    FID2D.FID[i].real.sig[j] := r[i][j];


  error:=false;
  errorMessage:='';
  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDCovariance.Process(FID2D:TFID_2D):integer;
var i,j,k:cardinal;
    a: double;
    c: array of array of double;
    s: array of double;
begin
  if FID2D.length_2D<2 then begin result:=1; exit; end;
  if FID2D.al<2 then begin result:=1; exit; end;

  //
  //  c[i][j]̃m
  //
  SetLength(c,FID2D.al);
  for j:=0 to FID2D.al-1 do SetLength(c[j],FID2D.al);

  //
  //  c[i][j]̏
  //
  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do c[i][j]:=0.0;



  //
  //  s[j]ic̕ρj̏ƌvZ
  //
  setLength(s,FID2D.al);
  for i:=0 to FID2D.al-1 do s[i]:=0.0;

  for j:=0 to FID2D.al-1 do begin
    for i:=0 to FID2D.length_2D-1 do begin
      s[j]:=s[j]+FID2D.FID[i].real.sig[j];
    end; //i
    s[j]:=s[j]/FID2D.length_2D;
  end; //j



  for i:=0 to FID2D.al-1 do begin
    for j:=0 to FID2D.al-1 do begin

      a:=0.0;
      for k:=0 to FID2D.al-1 do begin
        a:=a+(FID2D.FID[k].real.sig[i]-s[i])*(FID2D.FID[k].real.sig[j]-s[j]);
      end;
      c[i][j]:= a/(FID2D.al-1);

    end; // j
  end; //i


  //
  //  c[i][j] -> FID2DRealɑ
  //

  if FID2D.length_2D<FID2D.al then repeat FID2D.add; until FID2D.length_2D=FID2D.al;
      // longitudinal zero fill!
  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do FID2D.FID[i].real.sig[j]:=c[i][j];

  error:=false;
  errorMessage:='';
  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDLongitudinalZeroFill.Process(FID2D:TFID_2D):integer;
var j:cardinal;
begin
  if points+FID2D.length_2D>65536 then result:=1 else begin

    for j:=1 to points do begin
      FID2D.add;
    end; // j
    error:=false;
    errorMessage:='';
    result:=0;
  end;
end;
//------------------------------------------------------------------------------

function TFIDLongitudinalFFT.Process(FID2D:TFID_2D):integer;
begin
  FID2D.longitudinalFFT;
  error:=false;
  errorMessage:='';
  result:=0;
end;

//------------------------------------------------------------------------------
function TFIDZeroFill.Process(FID:TFID):integer;
var k:cardinal;
begin
  if points+FID.AL>65536 then result:=1 else begin
    FID.AL:=FID.AL+points;
    for k:=FID.AL-1 downto FID.AL-points do begin
      FID.real.sig[k]:=0.0;
      FID.imag.sig[k]:=0.0;
      FID.abs.sig[k]:=0.0;
    end; // k
    error:=false;
    errorMessage:='';
    result:=0;
  end; // if;

end;

function TFIDZeroFill.Process(FID2D:TFID_2D):integer;
var j:cardinal;
begin
  for j:=0 to FID2D.length_2D-1 do Process(FID2D.FID[j]);
  error:=false;
  errorMessage:='';
  result:=0;
end;

//------------------------------------------------------------------------------
function TFIDFFT.Process(FID:TFID):integer;
begin
  FID.FFT;
  error:=false;
  errorMessage:='';
  result:=0;
end;

function TFIDFFT.Process(FID2D:TFID_2D):integer;
var k:cardinal;
begin
  for k:=0 to FID2D.length_2D-1 do Process(FID2D.FID[k]);
  error:=false;
  errorMessage:='';
  result:=0;
end;
{----------------------------------------------------------}
constructor TFID_2D.Create(r:single; k:cardinal);
begin


  FAcqPhase:=0.0;
  FAL:=k;
  FDW:=r;
  SetLength(FID,0);
  length_2D:=0;
//  resized:=false;
  Fxini:=0;
  Fxfin:=FAL-1;
  Fvoffset:=0.0;
  cursorOn:=false;
  autoscale:=true;
  receiverScale:=false;
  plotScale:=1.0;
  plotOffset:=0.0;
  plotOption:=[FIDReal,FIDImag,FIDAbs];

  latestClickedXPosition:=0;

  plotMode:=pmNormal;
  saveMode:=smBinary;
  saveReImSeparately:=false;

  lMargin:=20; rMargin:=20; tMargin:=45; bMargin:=45;


  backGroundColor:=$00000000;
  realColor:=$000000FF;
  imagColor:=$0000FF00;
  absColor:=$0000FFFF;
  polarPlotColor:=$00FFFF00;

  cursorColor:=$00FFFFFF;
  axisWidth:=1;
  axisColor:=$00555555;

  penWidth:=2;
  cursorPenWidth:=1;
  canvasFontName:='Tahoma';
//  canvasFontStyle;
  canvasFontSize:=8;
  canvasFontColor:=$00FFFF00;

  axisFontName:='Tahoma';
//  axisFontStyle;
  axisFontSize:=11;
  axisFontColor:=$FFFF00;

  penstyle:=line;

  busy:=false;
  repeatScan:=false;

end;
{----------------------------------------------------------}
destructor TFID_2D.free;
begin
  while length_2D>0 do begin
    delete(length_2D-1);
  end;

end;

//-----------------------------------------------------------
procedure TFID_2D.setAxisFontSize(k:integer);
begin
  FAxisFontSize:=k;
  bMargin:=Round(5.0*k);
end;

{----------------------------------------------------------}
procedure TFID_2D.delete(k:cardinal);
var q:integer;
begin
  for q:=k to length_2D-2 do begin
    FID[k]:=FID[k+1];
  end; // q
  FID[length_2D-1].Free;
  dec(length_2D);
end;
{----------------------------------------------------------}
procedure TFID_2D.add;
begin
  Inc(length_2D);
  SetLength(FID,length_2D);
  FID[length_2D-1]:=TFID.create(AL);
end;
//------------------------------------------------------------------------------
procedure TFID_2D.swap(i,j: cardinal);
var auxFID:TFID;
    k: integer;
begin
  if length_2D<1 then exit;
  if i=j then exit;
  if (i>length_2D-1) then exit;
  if (j>length_2D-1) then exit;

  auxFID:=TFID.create(al);

  for k:=0 to al-1 do begin
    auxFID.real.sig[k] :=  FID[i].real.sig[k];
    auxFID.imag.sig[k] := FID[i].imag.sig[k];
    auxFID.abs.sig[k] := FID[i].abs.sig[k];

    FID[i].real.sig[k] := FID[j].real.sig[k];
    FID[i].imag.sig[k] := FID[j].imag.sig[k];
    FID[i].abs.sig[k] := FID[j].abs.sig[k];

    FID[j].real.sig[k] := auxFID.real.sig[k];
    FID[j].imag.sig[k] := auxFID.imag.sig[k];
    FID[j].abs.sig[k] := auxFID.abs.sig[k];
  end; // k

  auxFID.free;
end;
//------------------------------------------------------------------------------
procedure TFID_2D.reverse;
var  k,n:cardinal;
begin
  if length_2D<2 then exit;
  n:=length_2D div 2;
  for k:=0 to n-1 do begin
    swap(k,length_2D-1-k);
  end;
end;
//------------------------------------------------------------------------------
function TFID_2D.horizontalFFT:integer;
var
  j: integer;
begin
  for j:=0 to Length_2D-1 do begin
    FID[j].FFT;
  end; // j
  result:=0;
end;
//------------------------------------------------------------------------------
function TFID_2D.horizontalFFT(k:cardinal):integer;
begin
  if k>Length_2D-1 then begin
    result:=-1;
    exit;
  end;
  FID[k].FFT;
  result:=0;
end;
//------------------------------------------------------------------------------
function TFID_2D.longitudinalFFT:integer;
var
  j,p:integer;
begin
  for j:=0 to FAL-1 do begin
    p:=longitudinalFFT(j);
    if p<>0 then begin result:=p; exit; end; // error?
  end; // j
  result:=0;
end;
//------------------------------------------------------------------------------
function TFID_2D.longitudinalFFT(k:cardinal):integer;
var transFID: TFID;
    i,n: cardinal;
begin
  // Length_2D <= 1    ---> stop
  // k>FAL-1           ---> stop

  if Length_2D <= 1 then begin result:=-1; exit; end;
  if k>FAL-1 then begin result:=-1; exit; end;

  n:=1;
  while(n<Length_2D) do begin n:=n shl 1; end;   // 2̉H

  transFID:=TFID.create(n);
  for i:=0 to Length_2D-1 do begin
    transFID.real.sig[i]:=FID[i].real.sig[k];
    transFID.imag.sig[i]:=FID[i].imag.sig[k];
  end;

  for i:=Length_2D to n-1 do begin  // array length 2^nłȂ[ǉ
    transFID.real.sig[i]:=0;
    transFID.imag.sig[i]:=0;
    transFID.abs.sig[i]:=0;
  end;

  transFID.FFT;

  for i:=0 to Length_2D-1 do begin
    FID[i].real.sig[k]:=transFID.real.sig[i];
    FID[i].imag.sig[k]:=transFID.imag.sig[i];
    FID[i].abs.sig[k]:=transFID.abs.sig[i];
  end;

  transFID.free;
  result:=0;
end;
{----------------------------------------------------------}
procedure TFID_2D.ReadVnmrFile(fn: string);
var
  F1:TextFile;
  vnpFile,s,s2:string;
  k: integer;
  fr: double;
begin
 // sm2pFile:=ChangeFileExt(fn,'.sm2p');
  vnpFile:=fn;
  AssignFile(F1, vnpFile);
  Reset(F1);

  fileName:=fn;

  repeat
    Readln(F1,s); //ShowMessage(buf);
    { 󔒂}
    s:=Trim(LowerCase(s));

    if Copy(s,1,2)='np' then begin
      Readln(F1,s2);
      k:=1;
      repeat inc(k) until s2[k]=' ';
      inc(k);
      s2:=Copy(s2,k,length(s2)-k);
      FAL:=StrToInt(s2);

      FAL:=FAL div 2;
    end else if Copy(s,1,2)='sw' then begin
      Readln(F1,s2);
      k:=1;
      repeat inc(k) until s2[k]=' ';
      inc(k);
      s2:=Copy(s2,k,length(s2)-k);
      fr:=StrToFloat(s2);
      FDW:=1000000/fr;
    end else if Copy(s,1,4)='sfrq' then begin
      Readln(F1,s2);
      k:=1;
      repeat inc(k) until s2[k]=' ';
      inc(k);
      s2:=Copy(s2,k,length(s2)-k);
      FSF1:=StrToFloat(s2);
    end; // if
  until EOF(F1);

  CloseFile(F1);


end;

{----------------------------------------------------------}
procedure TFID_2D.Readsm2pFile(fn:string);
var
  F1:TextFile;
  sm2pFile,s,s2:string;
  k: integer;
begin
  sm2pFile:=ChangeFileExt(fn,'.sm2p');
  AssignFile(F1, sm2pFile);
  Reset(F1);

  fileName:=fn;

  repeat
    Readln(F1,s); //ShowMessage(buf);
    { 󔒂}
    s:=Trim(LowerCase(s));
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
//    while Pos(' ', s) > 0 do system.Delete(s,Pos(' ',s),1);
    if Copy(s,1,6)='point=' then begin
      s2:=Copy(s,7,Length(s));
      FAL:=StrToInt(s2);
    end else if Copy(s,1,3)='dw=' then begin
      s2:=Copy(s,4,Length(s));
      FDW:=StrToFloat(s2);
    end else if Copy(s,1,4)='sf1=' then begin
      s2:=Copy(s,5,Length(s));
      FSF1:=StrToFloat(s2);
    end; // if
  until EOF(F1);

  CloseFile(F1);

end;
{----------------------------------------------------------}
procedure TFID_2D.Readsm2dFile(fn:string;progress:TProgressBar);
var  sm2dFile :string;
     fs:TFileStream;
     ri:array of single;
     k,m,c,n:integer;
     sr:TSearchRec;

begin
  sm2dFile:=ChangeFileExt(fn,'.sm2d');
  fs:=TFileStream.Create(sm2dFile,fmOpenRead);

  FindFirst(sm2dFile,faAnyFile,sr);
  c:=sr.FindData.nFileSizeLow;
  n:=c div(2*integer(AL)*sizeof(single));

  setLength(ri,2*AL);

  progress.Max:=n;
  progress.Visible:=true;
  //info.Caption:='';
  //info.Visible:=true;

  for k := 0 to n - 1 do begin

      progress.Position:=k;
    //  info.Caption:='Reading data: '+IntToStr(k+1)+'/'+IntToStr(n);

      Add; // FIDǉ
      FID[length_2D-1].AL:=AL;
      FID[length_2D-1].DW:=DW;

        fs.Read(ri[0], 2*AL*sizeof(single));

        for m := 0 to al-1 do begin
          FID[length_2D-1].real.sig[m]:=ri[2*m];
          FID[length_2D-1].imag.sig[m]:=ri[2*m+1];
          FID[length_2D-1].abs.sig[m]:=
                 sqrt(ri[2*m]*ri[2*m]+ri[2*m+1]*ri[2*m+1]);

        end;// for m
      end;  //for k

  fs.Free;

  progress.Visible:=false;
 // info.Visible:=false;
end;

{----------------------------------------------------------}
procedure TFID_2D.Readsm2dFile(fn:string);
var  sm2dFile :string;
     fs:TFileStream;
     ri:array of single;
     k,m,c,n:integer;
     sr:TSearchRec;

begin
  sm2dFile:=ChangeFileExt(fn,'.sm2d');
  fs:=TFileStream.Create(sm2dFile,fmOpenRead);

  FindFirst(sm2dFile,faAnyFile,sr);
  c:=sr.FindData.nFileSizeLow;
  n:=c div(2*integer(AL)*sizeof(single));

  setLength(ri,2*AL);

  for k := 0 to n - 1 do begin

      Add; // FIDǉ
      FID[length_2D-1].AL:=AL;
      FID[length_2D-1].DW:=DW;

        fs.Read(ri[0], 2*AL*sizeof(single));

        for m := 0 to al-1 do begin
          FID[length_2D-1].real.sig[m]:=ri[2*m];
          FID[length_2D-1].imag.sig[m]:=ri[2*m+1];
          FID[length_2D-1].abs.sig[m]:=
                 sqrt(ri[2*m]*ri[2*m]+ri[2*m+1]*ri[2*m+1]);

        end;// for m
      end;  //for k

  fs.Free;

end;
{----------------------------------------------------------}
procedure TFID_2D.ReadAvanceParFile(fn:string);
var
  F1:TextFile;
  s,s2:string;
  k: integer;
begin
  AssignFile(F1, fn);
  Reset(F1);

  fileName:=fn;

  repeat
    Readln(F1,s);
    { 󔒂}
    s:=Trim(s);
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
    if Copy(s,1,6)='##$TD=' then begin
      s2:=Copy(s,7,Length(s));
      FAL:=StrToInt(s2) div 2; // TD carries re + im data points
    end else if Copy(s,1,8)='##$SW_h=' then begin
      s2:=Copy(s,9,Length(s));
      FDW:=1000000/StrToFloat(s2);
    end else if Copy(s,1,7)='##$BF1=' then begin
      s2:=Copy(s,8,Length(s));
      FSF1:=StrToFloat(s2);
    end; // if
  until EOF(F1);

  CloseFile(F1);

end;


//
//--------------------------------------------------------------
procedure TFID_2D.ReadAvanceDataFile(fn: string;progress:TProgressBar);
var
     fs:TFileStream;
     ri:array of integer;
     k,m,c,n,q:integer;
     sr:TSearchRec;
     re,im:single;

begin

  fs:=TFileStream.Create(fn,fmOpenRead);

  FindFirst(fn,faAnyFile,sr);
  c:=sr.FindData.nFileSizeLow;
  n:=c div((2*integer(AL))*sizeof(integer) + 0);

  setLength(ri,2*AL+256);

  progress.Max:=n;
  progress.Position:=0;
  progress.Visible:=true;

  for k := 0 to n - 1 do begin

    progress.Position:=k;

    Add; // FIDǉ
    FID[length_2D-1].AL:=AL;
    FID[length_2D-1].DW:=DW;

    fs.Read(ri[0], (2*FAL)*sizeof(integer) + 0);

    m:=0; q:=0;
    repeat
      re:=ri[q];
      FID[length_2D-1].real.sig[m]:=re;
      inc(q);
      im:=ri[q];
      FID[length_2D-1].imag.sig[m]:=im;
      inc(q);
      FID[length_2D-1].abs.sig[m]:=sqrt(re*re+im*im);
      inc(m);
    until m=integer(al)-1;

  end;  //for k

  fs.Free;
  progress.Visible:=false;

end;

//--------------------------------------------------------------
procedure TFID_2D.ReadAvanceDataFile(fn: string);
var
     fs:TFileStream;
     ri:array of integer;
     k,m,c,n,q:integer;
     sr:TSearchRec;
     re,im:single;

begin

  fs:=TFileStream.Create(fn,fmOpenRead);

  FindFirst(fn,faAnyFile,sr);
  c:=sr.FindData.nFileSizeLow;
  n:=c div((2*integer(AL))*sizeof(integer) + 0);

  setLength(ri,2*AL+256);

  for k := 0 to n - 1 do begin
    Add; // FIDǉ
    FID[length_2D-1].AL:=AL;
    FID[length_2D-1].DW:=DW;

    fs.Read(ri[0], (2*FAL)*sizeof(integer) + 0);

    m:=0; q:=0;
    repeat
      re:=ri[q];
      FID[length_2D-1].real.sig[m]:=re;
      inc(q);
      im:=ri[q];
      FID[length_2D-1].imag.sig[m]:=im;
      inc(q);
      FID[length_2D-1].abs.sig[m]:=sqrt(re*re+im*im);
      inc(m);
    until m=integer(al)-1;

  end;  //for k

  fs.Free;

end;


{----------------------------------------------------------}
procedure TFID_2D.SaveToFile(fn:string);
var  sm2dFile, reFile, imFile :string;
     fs:TFileStream;
     i,j:integer;
     st:TStrings;
//     re,im:single;
     ri:array of single;
     str: String;

begin

  if saveReImSeparately then begin
    setLength(ri,AL);
    reFile:=ChangeFileExt(fn,'.bre');
    fs:=TFileStream.Create(reFile,fmCreate);
      for i:=0 to length_2D-1 do begin
        for j:=0 to AL-1 do ri[j]:=FID[i].real.sig[j];
        fs.Write(ri[0],al*sizeof(single));
      end; // i
    fs.Free;

    imFile:=ChangeFileExt(fn,'.bim');
    fs:=TFileStream.Create(imFile,fmCreate);
      for i:=0 to length_2D-1 do begin
        for j:=0 to AL-1 do ri[j]:=FID[i].imag.sig[j];
          fs.Write(ri[0],al*sizeof(single));
      end; // i
    fs.Free;
  end;


  setLength(ri,2*AL);

  sm2dFile:=ChangeFileExt(fn,'.sm2d');
  fs:=TFileStream.Create(sm2dFile,fmCreate);
    for i:=0 to length_2D-1 do begin
      for j:=0 to AL-1 do begin
         ri[2*j]:=FID[i].real.sig[j];
         ri[2*j+1]:=FID[i].imag.sig[j];

//        re:=FID[i].real.sig[j];  // double -> single conversion (|mlqƌ݊ۂ߁EEE)
//        im:=FID[i].imag.sig[j];  // double -> single conversion
//        fs.Write(re, sizeof(single));
//        fs.Write(im, sizeof(single));
      end; // j
      fs.Write(ri[0],2*al*sizeof(single));
    end; // i
  fs.Free;


//  if saveMode=smAbs then begin
//  st:=TStringList.Create;

//    for i:=0 to length_2D-1 do begin
//      str:='';
//      for j:=0 to AL-1 do begin
//        str:=str+FLoatToStr(FID[i].abs.sig[j])+'  ';
//      end; // j
//      st.add(str);
//    end; // i
//    st.SaveToFile(ChangeFileExt(fn,'.abs'));
//   st.Free;
//  end;  //smabs

  if saveMode=smBoth then begin     // Text data
  st:=TStringList.Create;

    for i:=0 to length_2D-1 do begin
      for j:=0 to AL-1 do begin
        st.Add(           FloatToStr(FID[i].real.sig[j])
                + '   ' + FloatToStr(FID[i].imag.sig[j])
                + '   ' + FLoatToStr(FID[i].abs.sig[j])
              );
      end; // j
    end; // i
    st.SaveToFile(ChangeFileExt(fn,'.dat'));

    st.Clear;

    if saveReImSeparately then begin
      for i:=0 to length_2D-1 do begin
        str:='';
        for j:=0 to AL-1 do begin
          str:=str+FLoatToStr(FID[i].real.sig[j])+'  ';
        end; // j
        st.add(str);
      end; // i
      st.SaveToFile(ChangeFileExt(fn,'.real'));
      st.Clear;

      for i:=0 to length_2D-1 do begin
        str:='';
        for j:=0 to AL-1 do begin
          str:=str+FLoatToStr(FID[i].imag.sig[j])+'  ';
        end; // j
        st.add(str);
      end; // i
      st.SaveToFile(ChangeFileExt(fn,'.imag'));
      st.Clear;

      for i:=0 to length_2D-1 do begin
        str:='';
        for j:=0 to AL-1 do begin
          str:=str+FLoatToStr(FID[i].abs.sig[j])+'  ';
        end; // j
        st.add(str);
      end; // i
      st.SaveToFile(ChangeFileExt(fn,'.abs'));
      st.Clear;

    end;


    st.Free;
  end;  // smBoth

 // st.Free;

end;
//------------------------------------------------------------------------------
procedure TFID_2D.setVoffset(v: Single);
begin
  if v>0.5 then FvOffset:=0.5 else if v<-0.5 then FVOffset:=-0.5
  else FVOffset:=v;

end;
//------------------------------------------------------------------------------
procedure TFID_2D.setXini(x:cardinal);
begin
  if x>Fal-2 then Fxini:=FAL-2 else if x>=Fxfin then Fxini:=Fxfin-1 else Fxini:=x;
  FXRange:=FXFin-FXini;
end;
//------------------------------------------------------------------------------
procedure TFID_2D.setXfin(x:cardinal);
begin
  if x>al-1 then Fxfin:=al-1 else if x<=FXini then Fxfin:=Fxini+1 else Fxfin:=x;
  FXRange:=FXFin-FXini;
end;
{----------------------------------------------------------}
procedure TFID_2D.drawCursor(ti:TPaintBox);
var re,im,ab,ph: single;
    newx:integer;
    unitString, valueString, deltaValueString:string;

begin

  if length_2D<1 then exit;
  if not (plotmode=pmNormal) then exit;
  if not cursorOn then exit;

  newx:=lMargin+
     Round((xPointLeft-xini)*(Cardinal(ti.Width)-lMargin-rMargin)/(xfin-xini));

  ti.Canvas.Pen.Mode:=pmCopy;

  ti.Canvas.Pen.Width:=cursorPenWidth;
  ti.Canvas.Pen.Color:=cursorColor;

  ti.Canvas.MoveTo(newx, tMargin);
  ti.Canvas.LineTo(newx,ti.Height-bMargin);
  oldx:=newx;
  cursorWasOff:=false;


  re:= FID[currentFID].real.sig[xPointLeft];
  im:= FID[currentFID].imag.sig[xPointLeft];
  ab:= FID[currentFID].abs.sig[xPointLeft];
  ph:=ArcTan2(im,re)*180.0/3.1415926535;

  ti.Canvas.Font.Name:=canvasFontName;
  ti.Canvas.Font.Size:=canvasFontSize;
  ti.Canvas.Font.Style:=canvasFontStyle;
  ti.Canvas.Font.Color:=canvasFontColor;

  ti.Canvas.TextOut(20 ,3, 'Cursor Position: ' + IntToStr(xPointLeft)+'      ');


  if FID[currentFID].domain=dmTime then begin
    unitString:=' ms        ';
    valueString:=FloatToStrF(0.001*dw*xPointLeft,ffGeneral,5,5);
    deltaValueString:=
            FloatToStrF(0.001*dw*integer(xPointLeft-latestClickedXPosition),
                        ffGeneral,5,5);
  end else begin

    unitstring:=' kHz        ';


    valueString:=FloatToStrF(500.0/dw-xPointLeft*1000.0/(dw*fal),
                            ffGeneral,5,5);
    deltaValueString:=FloatToStrF(integer(xPointLeft-latestClickedXPosition)*1000.0/(dw*fal),
                            ffGeneral,5,5);

  end;

  ti.Canvas.TextOut(20, 25, valueString+unitString);
  ti.Canvas.TextOut(150, 25, 'DeltaX = '+deltaValueString+unitString);


  ti.Canvas.Font.Color:=realColor;
  ti.Canvas.TextOut(300 ,3, 'Real: ' + FloatToStrF(re,ffGeneral,5,3)
                                     + '                ');
  ti.Canvas.Font.Color:=imagColor;
  ti.Canvas.TextOut(300 ,25, 'Imag: ' + FloatToStrF(im,ffGeneral,5,3)
                                     + '                ');
  ti.Canvas.Font.Color:=absColor;
  ti.Canvas.TextOut(450 ,3, 'Abs: ' + FloatToStrF(ab,ffGeneral,5,3)
                                     + '                ');
  ti.Canvas.Font.Color:=canvasFontColor;
  ti.Canvas.TextOut(450 ,25, 'Phase: ' + FloatToStrF(ph,ffGeneral,5,3)
                                     + '                ');

end;
{----------------------------------------------------------}
procedure TFID_2D.overWriteCursor(ti:TPaintBox);
var re,im,ab,ph: single;
    newx:integer;
    unitString, valueString, deltaValueString:string;

begin

  if length_2D<1 then exit;
  if not (plotmode=pmNormal) then exit;
  if not cursorOn then exit;


  ti.Canvas.Pen.Width:=cursorPenWidth;
  ti.Canvas.Pen.Color:=cursorColor;

  ti.Canvas.Pen.Mode:=pmNot;
  if not cursorWasOff then begin
    ti.Canvas.MoveTo(oldx,tMargin);
    ti.Canvas.LineTo(oldx,ti.Height-bMargin);
  end;

  newx:=lMargin+
     Round((xPointLeft-xini)*(Cardinal(ti.Width)-lMargin-rMargin)/(xfin-xini));


  if (oldx=newx) and (cursorOn=true) and (not cursorWasOff) then begin
    cursorOn:=false;
    cursorWasOff:=true;
    ti.Canvas.Pen.Mode:=pmCopy;
    exit;
  end;


  ti.Canvas.Pen.Width:=cursorPenWidth;
  ti.Canvas.Pen.Color:=cursorColor;

  ti.Canvas.MoveTo(newx, tMargin);
  ti.Canvas.LineTo(newx,ti.Height-bMargin);
  oldx:=newx;
  cursorWasOff:=false;

  ti.Canvas.Pen.Mode:=pmCopy;

  re:= FID[currentFID].real.sig[xPointLeft];
  im:= FID[currentFID].imag.sig[xPointLeft];
  ab:= FID[currentFID].abs.sig[xPointLeft];
  ph:=ArcTan2(im,re)*180.0/3.1415926535;

  ti.Canvas.Font.Name:=canvasFontName;
  ti.Canvas.Font.Size:=canvasFontSize;
  ti.Canvas.Font.Style:=canvasFontStyle;
  ti.Canvas.Font.Color:=canvasFontColor;

  ti.Canvas.TextOut(20 ,3, 'Cursor Position: ' + IntToStr(xPointLeft)+'      ');


  if FID[currentFID].domain=dmTime then begin
    unitString:=' ms        ';
    valueString:=FloatToStrF(0.001*dw*xPointLeft,ffGeneral,5,5);
    deltaValueString:=
            FloatToStrF(0.001*dw*integer(xPointLeft-latestClickedXPosition),
                        ffGeneral,5,5);
  end else begin

    unitstring:=' kHz        ';


    valueString:=FloatToStrF(500.0/dw-xPointLeft*1000.0/(dw*fal),
                            ffGeneral,5,5);
    deltaValueString:=FloatToStrF(integer(xPointLeft-latestClickedXPosition)*1000.0/(dw*fal),
                            ffGeneral,5,5);

  end;

  ti.Canvas.TextOut(20, 25, valueString+unitString);
  ti.Canvas.TextOut(150, 25, 'DeltaX = '+deltaValueString+unitString);


  ti.Canvas.Font.Color:=realColor;
  ti.Canvas.TextOut(300 ,3, 'Real: ' + FloatToStrF(re,ffGeneral,5,3)
                                     + '                ');
  ti.Canvas.Font.Color:=imagColor;
  ti.Canvas.TextOut(300 ,25, 'Imag: ' + FloatToStrF(im,ffGeneral,5,3)
                                     + '                ');
  ti.Canvas.Font.Color:=absColor;
  ti.Canvas.TextOut(450 ,3, 'Abs: ' + FloatToStrF(ab,ffGeneral,5,3)
                                     + '                ');
  ti.Canvas.Font.Color:=canvasFontColor;
  ti.Canvas.TextOut(450 ,25, 'Phase: ' + FloatToStrF(ph,ffGeneral,5,3)
                                     + '                ');

end;
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}




//------------------------------------------------------------------------------
constructor TFID.create(k:cardinal);
begin
  real:=THalfFID.create(k);
  imag:=THalfFID.create(k);
  abs:=THalfFID.create(k);
  FAL:=k;

  domain:=dmTime;   //  ŏ͎ԃf[^BFFTg(dmFrequency)ɁB

  phase0:=0.0;
  actualNA:=0;

  initialize;
end;
//------------------------------------------------------------------------------
destructor TFID.free;
begin
  real.free;
  imag.free;
  abs.free;
end;
//------------------------------------------------------------------------------
procedure TFID.initialize;
begin
  real.initialize;
  imag.initialize;
  abs.initialize;
end;
//------------------------------------------------------------------------------
function TFID.phase(k: Integer):double;
begin
  result:=ArcTan2(imag.sig[k],real.sig[k]);
end;
//------------------------------------------------------------------------------
function TFID.averagePhase:double;
begin
  result:=averagePhase(0,al-1);
end;
//------------------------------------------------------------------------------
function TFID.averagePhase(kini,kfin:integer):double;
var k:integer;
    ap:double;
begin
  ap:=0.0;
  if (kini<0) or (kfin>integer(al)-1) or (kini>kfin) then result:=0.0
  else begin
    for k := kini to kfin do ap:=ap+phase(k);
    result:=ap/(kfin-kini+1);
  end;
end;
//------------------------------------------------------------------------------
function TFID.phaseStandardDeviation:double;
begin
  result:=phaseStandardDeviation(0,al-1);
end;
//------------------------------------------------------------------------------
function TFID.phaseStandardDeviation(kini,kfin: Integer):double;
var k:integer;
    asd,sd:double;
begin
  sd:=0.0;
  if (kini<0) or (kfin>integer(al)-1) or (kini>kfin) then result:=0.0
  else begin
    asd:=averagePhase(kini,kfin);
    for k := kini to kfin do sd:=sd+(phase(k)-asd)*(phase(k)-asd);
    sd:=sd/(kfin-kini+1);
    sd:=sqrt(sd);
    result:=sd;
  end;
end;
//------------------------------------------------------------------------------
function TFID.averageIntensity:double;
begin
  result:=averageIntensity(0,al-1);
end;
//------------------------------------------------------------------------------
function TFID.averageIntensity(kini,kfin: Integer):double;
var k:integer;
    ai:double;
begin
  ai:=0.0;
  if (kini<0) or (kfin>integer(al)-1) or (kini>kfin) then result:=0.0
  else begin
    for k := kini to kfin do ai:=ai+abs.sig[k];
    result:=ai/(kfin-kini+1);
  end;
end;
//------------------------------------------------------------------------------
function TFID.intensityStandardDeviation:double;
begin
  result:=intensityStandardDeviation(0,al-1);
end;
//------------------------------------------------------------------------------
function TFID.intensityStandardDeviation(kini,kfin: Integer):double;
var k:integer;
    asd,sd:double;
begin
  sd:=0.0;
  if (kini<0) or (kfin>integer(al)-1) or (kini>kfin) then result:=0.0
  else begin
    asd:=averageintensity(kini,kfin);
    for k := kini to kfin do sd:=sd+(abs.sig[k]-asd)*(abs.sig[k]-asd);
    sd:=sd/(kfin-kini+1);
    sd:=sqrt(sd);
    result:=sd;
  end;
end;
//------------------------------------------------------------------------------
function TFID.intensityCount(i1,i2: double):integer;
begin
  result:=intensityCount(i1,i2,0,al-1);
end;
//------------------------------------------------------------------------------
function TFID.intensityCount(i1,i2:double; kini,kfin:integer):integer;
var k:integer;
    c:integer;
begin
  c:=0;
  if (kini<0) or (kfin>integer(al)-1) or (kini>kfin) then result:=0
  else begin

    for k := kini to kfin do begin
      if (abs.sig[k]>=i1) and (abs.sig[k]<i2) then inc(c);
    end;

    result:=c;
  end;
end;
//------------------------------------------------------------------------------
procedure TFID.updateAbs;
var k:integer;
begin
  for k := 0 to integer(AL) - 1 do
    abs.sig[k]:=sqrt(real.sig[k]*real.sig[k]
                    + imag.sig[k]*imag.sig[k]);

end;
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
procedure TFID.rotate(phi_deg:double);
var
  phi,c,s: double;
  xr,xi: double;
  i: cardinal;
begin

  phi:=3.14159265358979*phi_deg/180.0;  // degree -> radian conversion
  c:=cos(phi);
  s:=sin(phi);

  for i:=0 to FAL-1 do begin
    xr:=real.sig[i];
    xi:=imag.sig[i];
    real.sig[i] := c*xr - s*xi;
    imag.sig[i] := c*xi + s*xr;
  end; // i

  phase0:=phase0+phi_deg;

end;
//------------------------------------------------------------------------------
procedure TFID.LT;
var
  n:cardinal;
  ex:double;
begin
  for n:=0 to FAL-1 do begin
    ex:=exp(-2.0*n/FAL);
    real.sig[n]:=real.sig[n]*ex;
    imag.sig[n]:=imag.sig[n]*ex;
  end;
  FFT;
end;
//------------------------------------------------------------------------------
procedure TFID.inverseLT;
var
  n:cardinal;
  ex: double;
begin
  inverseFFT;

  for n:=0 to FAL-1 do begin
    ex:=exp(2.0*n/FAL);
    real.sig[n]:=real.sig[n]*ex;
    imag.sig[n]:=imag.sig[n]*ex;
    abs.sig[n]:=abs.sig[n]*ex;
  end;

end;
//------------------------------------------------------------------------------
procedure TFID.inverseFFT;
var
  theta:double;
  yr,yi,xr,xi,wr,wi: double;
  i,j,k,m,mh,irev,n,nh: cardinal;
begin

  // FAL͂Q̉?
  n:=1;
  while n<FAL do begin
    n := n shl 1;
  end; // while

  // FALQnłȂ΁A[tB
  if FAL<n then begin
    k:=AL;
    FAL:=n;
    setLength(real.sig,AL);
    setLength(imag.sig,AL);
    setLength(abs.sig,AL);
    for m:=k to AL-1 do begin
      real.sig[m]:=0.0;
      imag.sig[m]:=0.0;
    end; // for
  end;

  theta:=6.28318530717958647692528676656/n;

  // -----------  inverse FFT -------------

  // STEP 1: NMRp̕בւ
  //    iS[ŁAvXgE}CiXgɂ̂߂Bj
  nh:=n div 2;
  for k:=0 to nh-1 do begin
    xr:=real.sig[k]; xi:=imag.sig[k];
    real.sig[k]:=real.sig[nh+k]; imag.sig[k]:=imag.sig[nh+k];
    real.sig[nh+k]:=xr; imag.sig[nh+k]:=xi;
  end;


  // STEP 2
  m:=n;
  mh:=m shr 1;

  while m>=2 do begin
    irev:=0;
    i:=0;
    while i<n do begin
      wr:=cos(theta * irev);
      wi:=sin(theta * irev);

      k:=n shr 2;
      irev:=irev xor k;
      while k>irev do begin
        k:=k shr 1;
        irev:=irev xor k;
      end; //while

      for j:=i to mh+i-1 do begin
        k:=j+mh;
        xr:=wr*real.sig[k]+wi*imag.sig[k];   // FFT̂Ƃ+-tI
        xi:=wr*imag.sig[k]-wi*real.sig[k];   // I
        yr:=real.sig[j]; yi:=imag.sig[j];
        real.sig[j]:=yr+xr;
        imag.sig[j]:=yi+xi;
        real.sig[k]:=yr-xr;
        imag.sig[k]:=yi-xi;;
      end; // for
    i:=i+m;
    end; // while
  m:=mh;
  mh:=mh shr 1;
  end; // while


  //STEP 3
  i:=0;
  for j:=1 to n-2 do begin
    k:=n shr 1;
    i:=i xor k;
    while k>i do begin
      k:=k shr 1;
      i:=i xor k;
    end; // while
    if j<i then begin
      xr:=real.sig[j];
      xi:=imag.sig[j];
      real.sig[j]:=real.sig[i];
      imag.sig[j]:=imag.sig[i];
      real.sig[i]:=xr;
      imag.sig[i]:=xi;
    end; // if(j<i)

  end; // for(j)


  // Abs ̌vZ
  for k:=0 to AL-1 do begin
    abs.sig[k]:=sqrt(real.sig[k]*real.sig[k] + imag.sig[k]*imag.sig[k]);
  end;

  if domain=dmTime then domain:=dmFrequency
  else if domain=dmFrequency then domain:=dmTime;

end;
//------------------------------------------------------------------------------
procedure TFID.FFT;
var
  theta:double;
  xr,xi,wr,wi: double;
  i,j,k,m,mh,irev,n,nh: cardinal;
begin

  // FAL͂Q̉?
  n:=1;
  while n<FAL do begin
    n := n shl 1;
  end; // while

  // FALQnłȂ΁A[tB
  if FAL<n then begin
    k:=AL;
    FAL:=n;
    setLength(real.sig,AL);
    setLength(imag.sig,AL);
    setLength(abs.sig,AL);
    for m:=k to AL-1 do begin
      real.sig[m]:=0.0;
      imag.sig[m]:=0.0;
    end; // for
  end;

  theta:=6.28318530717958647692528676656/n;

  // -----------  FFT -------------

  i:=0;
  for j:=1 to n-2 do begin
    k:=n shr 1;
    i:=i xor k;
    while k>i do begin
      k:=k shr 1;
      i:=i xor k;
    end; // while
    if j<i then begin
      xr:=real.sig[j];
      xi:=imag.sig[j];
      real.sig[j]:=real.sig[i];
      imag.sig[j]:=imag.sig[i];
      real.sig[i]:=xr;
      imag.sig[i]:=xi;
    end; // if(j<i)

  end; // for(j)


  mh:=1;
  m:=mh shl 1;
  while m<=n do begin
    irev:=0;
    i:=0;
    while i<n do begin
      wr:=cos(theta * irev);
      wi:=sin(theta * irev);

      k:=n shr 2;
      irev:=irev xor k;
      while k>irev do begin
        k:=k shr 1;
        irev:=irev xor k;
      end; //while

      for j:=i to mh+i-1 do begin
        k:=j+mh;
        xr:=real.sig[j]-real.sig[k];
        xi:=imag.sig[j]-imag.sig[k];
        real.sig[j]:=real.sig[j]+real.sig[k];
        imag.sig[j]:=imag.sig[j]+imag.sig[k];
        real.sig[k]:=wr*xr-wi*xi;
        imag.sig[k]:=wr*xi+wi*xr;
      end; // for
    i:=i+m;
    end; // while
  mh:=m;
  m:=mh shl 1;
  end; // while

  // NMRp̕בւiS[ŁAvXgE}CiXgj
  nh:=n div 2;
  for k:=0 to nh-1 do begin
    xr:=real.sig[k]; xi:=imag.sig[k];
    real.sig[k]:=real.sig[nh+k]; imag.sig[k]:=imag.sig[nh+k];
    real.sig[nh+k]:=xr; imag.sig[nh+k]:=xi;
  end;


  // Abs ̌vZ
  for k:=0 to AL-1 do begin
    abs.sig[k]:=sqrt(real.sig[k]*real.sig[k] + imag.sig[k]*imag.sig[k]);
  end;


  if domain=dmTime then domain:=dmFrequency
  else if domain=dmFrequency then domain:=dmTime;

end;
//------------------------------------------------------------------------------
function TFID.Max:double;
var q,r,i:double;
begin
  r:=real.Max;
  i:=imag.Max;
  if r>=i then q:=r else q:=i;
  result:=q;
end;

function TFID.Max(kini,kfin:integer):double;
var q,r,i:double;
begin
  r:=real.Max(kini,kfin);
  i:=imag.Max(kini,kfin);
  if r>=i then q:=r else q:=i;
  result:=q;
end;

//------------------------------------------------------------------------------

function TFID.Min:double;
var r,i:double;
begin
  r:=real.Min;
  i:=imag.Min;
  if r<=i then result:=r else result:=i;
end;

function TFID.Min(kini,kfin:integer):double;
var r,i:double;
begin
  r:=real.Min(kini,kfin);
  i:=imag.Min(kini,kfin);
  if r<=i then result:=r else result:=i;
end;


//------------------------------------------------------------------------------

function TFID.AbsMax:double;
var a,b:double;
begin
  a:=real.AbsMax; b:=imag.AbsMax;
  if a>=b then result:=a else result:=b;
end;

function TFID.AbsMax(kini,kfin:integer):double;
var a,b:double;
begin
  a:=real.AbsMax(kini,kfin);
  b:=imag.AbsMax(kini,kfin);
  if a>=b then result:=a else result:=b;
end;
//------------------------------------------------------------------------------
procedure TFID_2D.draw(ti:TPaintBox; k:cardinal);
begin
  if length_2D<1 then exit;

  if plotMode=pmNormal then cartesianDraw(ti,k) else polarDraw(ti,k);
 // cursorOn:=false;
 // if plotMode=pmNormal then drawCursor(ti, cursorX);

end;
//------------------------------------------------------------------------------
procedure TFID_2D.polarDraw(ti:TPaintBox; k:cardinal);
var hheight,q:integer;
    dx,x,y,scale:single;
//    am:single;
begin
  hheight:=Round(ti.Height/2);
  if ti.Width>ti.Height then dx:=(ti.Width-ti.Height)/2 else dx:=0.0;
  if autoScale then begin

    if FID[k].abs.Max=0 then scale:=1
    else scale:=(ti.Height)/(2.0*FID[k].Abs.Max);

    PlotScale:=scale;
  end else if receiverScale then begin
    scale:=(ti.height)/16384;
    PlotScale:=scale;
  end
  else scale:=PlotScale;


//  ti.Canvas.Brush.Color:=$00441111;
  ti.Canvas.Brush.Color:=backgroundColor;
  ti.Canvas.FillRect(Rect(0,0,ti.Width,ti.Height));

  ti.Canvas.Pen.Width:=axisWidth;
  ti.Canvas.Pen.Color:=axisColor;

  x:=hheight+dx;
  ti.Canvas.Ellipse(Round(x)-hheight,0,
                    Round(x)+hheight,ti.Height);

  // Y=0 line
//  ti.Canvas.Pen.Color:=clGray;
  y:=hheight;
  ti.Canvas.MoveTo(Round(dx), Round(y));
  ti.Canvas.LineTo(Round(dx+2*y-1), Round(y));

  // X=0 line
//  x:=hheight+dx;
  ti.Canvas.MoveTo(Round(x), 0);
  ti.Canvas.LineTo(Round(x), hheight*2);



  x:=hheight+scale*FID[k].real.sig[xini]+dx;
  y:=hheight-scale*FID[k].imag.sig[xini];
  ti.Canvas.Pen.Color:=$0000FF00;
  ti.Canvas.Brush.Color:=$0000FF00;
  ti.Canvas.Ellipse(Round(x-5),Round(y-5),
                    Round(x+5),Round(y+5));

  ti.Canvas.MoveTo(Round(x), Round(y));
  ti.Canvas.Pen.Color:=polarPlotColor;
  ti.Canvas.Brush.Color:=polarPlotColor;

  ti.Canvas.Pen.Width:=penWidth;
  for q:=xini+1 to xfin do begin
      x:=hheight+scale*FID[k].real.sig[q]+dx;
      y:=hheight-scale*FID[k].imag.sig[q];
      if x<=0 then x:=0 else if x>ti.Width then x:=ti.Width-1;
      if y<=0 then y:=0 else if y>ti.Height then y:=ti.Height-1;
//      ti.Canvas.LineTo(Round(x),Round(y));

   if penStyle=line then ti.canvas.lineTo(Round(x),Round(y))
   else ti.Canvas.Ellipse(Round(x-penWidth),Round(y-penWidth),
                           Round(x+penWidth),Round(y+penWidth));


  end; // k

  ti.Canvas.Pen.Color:=$000000FF;
  ti.Canvas.Brush.Color:=$000000FF;
  ti.Canvas.Ellipse(Round(x-5),Round(y-5),
                    Round(x+5),Round(y+5));


end;
//------------------------------------------------------------------------------
procedure TFID_2D.halfFIDDraw(hFID:THalfFID;ti:TPaintBox;color:TColor;
       dx:single; vCenter:word; scale:single);
var q,dq,qfin:integer;
    c,x,y:single;
begin
  x:=lMargin;
  y:=Round(vCenter-scale*hFID.sig[xini]);

  if y<=tMargin then y:=tMargin else if y>=(ti.Height-bMargin) then
                                        y:=ti.Height-bMargin-1;


  ti.Canvas.MoveTo(Round(x),Round(y));
  ti.Canvas.Pen.Color:=color;

  if AL>1 then q:=xini+1 else q:=xini;

  while q<=Integer(xfin) do begin
    c:=0.0;dq:=0;
    repeat
      c:=c+dx;
      inc(dq);
//    until trunc(c)>=1;
    until c>=0.4;
    if dq=1 then begin
      x:=x+dx;
      y:=Round(vCenter-scale*(hFID.sig[q]+PlotOffset));
      if y<=tMargin then y:=tMargin
      else if y>=(ti.Height-bMargin) then
                                         y:=ti.Height-bMargin-1;
      ti.Canvas.LineTo(Round(x),Round(y));
    end else begin
      x:=x+dx*dq;
      if q+dq>integer(xfin) then qfin:=xfin else qfin:=q+dq;
      y:=Round(vCenter-scale*(hFID.Min(q,qfin)+PlotOffset));
      if y<=tMargin then y:=tMargin+1 else if y>=(ti.Height-bMargin) then
                                         y:=ti.Height-bMargin-1;
      ti.Canvas.LineTo(Round(x),Round(y));

      y:=Round(vCenter-scale*(hFID.Max(q,qfin)+PlotOffset));
      if y<=tMargin then y:=tMargin+1 else if y>=(ti.Height-bMargin) then
                                         y:=ti.Height-bMargin-1;
      ti.Canvas.LineTo(Round(x),Round(y));
    end;
    q:=q+dq;
  end; // while

end;
//------------------------------------------------------------------------------
procedure TFID_2D.cartesianDraw(ti:TPaintBox; k:cardinal);
var hheight:integer;
    dx,scale:single;
    y:single;
    plotWidth,plotHeight:word;
    vCenter:word;
    t, unitString:string;
    fini,ffin:single;
    q:single;
    qq,c:integer;
  //  c:single;
  //  dq,qfin:integer;
begin

  plotWidth:=ti.Width-lMargin-rMargin;
  plotHeight:=ti.Height-tMargin-bMargin;

  dx:=plotWidth/(xfin-xini);

  hheight:=Round(plotHeight/2);

  vCenter:=tMargin+hheight+Round(plotHeight*Fvoffset);

  if autoScale then begin

    if not (FIDAbs in plotOption) then begin
      if FID[k].AbsMax>0 then scale:=(plotHeight)/(2.0*FID[k].AbsMax)
      else scale:=1;
    end else begin
      if FID[k].abs.Max=0 then scale:=1 else
                scale:=(plotHeight)/(2.0*FID[k].Abs.Max);
    end;

    PlotScale:=scale;
  end else if receiverScale then begin
    if (FID[k].actualNA=0) or (repeatScan) then
       scale:=plotHeight/16384
    else scale:=plotHeight/(16384.0*FID[k].actualNA);
    //  16384.0  (single! Not integer (16384) !!)
    //  Otherwise, overflow results in division by zero!
    Plotscale:=scale;
  end else begin // manual scale;
    scale:=PlotScale;
  end;


  ti.Canvas.Brush.Color:=backgroundColor;
  ti.Canvas.FillRect(Rect(0,0,ti.Width,ti.Height));

  ti.Canvas.Pen.Width:=axisWidth;
  ti.Canvas.Pen.Color:=axisColor;
  ti.Canvas.Rectangle(Rect(lMargin,tMargin,ti.Width-rMargin,ti.Height-bMargin));

  ti.Canvas.MoveTo(lMargin+(plotWidth div 2), ti.Height-bMargin-10);
  ti.Canvas.LineTo(lMargin+(plotWidth div 2), ti.Height-bMargin);
  ti.Canvas.MoveTo(lMargin+(plotWidth div 4), ti.Height-bMargin-10);
  ti.Canvas.LineTo(lMargin+(plotWidth div 4), ti.Height-bMargin);
  ti.Canvas.MoveTo(lMargin+(3*plotWidth div 4), ti.Height-bMargin-10);
  ti.Canvas.LineTo(lMargin+(3*plotWidth div 4), ti.Height-bMargin);

  ti.Canvas.Font.Name:=axisFontName;
  ti.Canvas.Font.Style:=axisFontStyle;
  ti.Canvas.Font.Size:=axisFontSize;
  ti.Canvas.Font.Color:=axisFontColor;








  if FID[k].domain=dmTime then begin
    unitString:='ms';

    ti.Canvas.TextOut(lMargin+(plotWidth div 2)-axisFontSize,
                      ti.Height-15-axisFontSize, unitString);
    if xini=0 then t:='0' else t:= FloatToStrF(xini*dw/1000,ffGeneral,5,3) ;
    ti.Canvas.TextOut(lMargin-5,ti.Height-bMargin, t);
    t:=FloatToStrF(xfin*dw/1000,ffGeneral,5,3);
    ti.Canvas.TextOut(ti.Width-rMargin-20,ti.Height-bMargin, t);
    t:=FloatToStrF(0.5*(xfin+xini)*dw/1000,ffGeneral,5,3);
    ti.Canvas.TextOut(lMargin+(plotWidth div 2)-15,ti.Height-bMargin, t);
    t:=FloatToStrF((0.25*xfin+0.75*xini)*dw/1000,ffGeneral,5,3);
    ti.Canvas.TextOut(lMargin+(plotWidth div 4)-15,ti.Height-bMargin, t);
    t:=FloatToStrF((0.75*xfin+0.25*xini)*dw/1000,ffGeneral,5,3);
    ti.Canvas.TextOut(lMargin+(3*plotWidth div 4)-15,ti.Height-bMargin, t);

//    ti.Canvas.TextOut(25,3, 'Max: ' + FloatToStrF(FID[k].abs.Max,ffExponent,5,3)
//                                  + '    ');

    // Y=0 line
    ti.Canvas.Pen.Color:=axisColor;
    ti.Canvas.Pen.Width:=axisWidth;
    y:=vCenter-scale*(plotOffset);
    ti.Canvas.MoveTo(lMargin,Round(y));
    ti.Canvas.LineTo(ti.Width-rMargin-1, Round(y));


    // bit lines

    ti.Canvas.font.size:=9;
    ti.canvas.font.color:=axisColor;

    q:=(vcenter-tmargin)/scale;
    qq:=Trunc(Log2(q));
    y:=vCenter-scale*(power(2,qq));

    ti.Canvas.MoveTo(lMargin,Round(y));
    ti.Canvas.LineTo(ti.Width-rMargin-1, Round(y));

    t:=IntToStr(qq+1);
    ti.Canvas.TextOut(ti.Width-rmargin-50, Round(y-14), t+' bit');

    c:=1;

    while (qq>0) and (c<=3) do begin
      inc(c);
      qq:=qq-1;
      y:=vCenter-scale*(power(2,qq));

      ti.Canvas.MoveTo(lMargin,Round(y));
      ti.Canvas.LineTo(ti.Width-rMargin-1, Round(y));

      t:=IntToStr(qq+1);
      ti.Canvas.TextOut(ti.Width-rmargin-50, Round(y-14), t+' bit');

    end;

    y:=vCenter+scale*(power(2,qq));
    while (y<ti.Height-bmargin) do begin
      ti.Canvas.MoveTo(lMargin,Round(y));
      ti.Canvas.LineTo(ti.Width-rMargin-1, Round(y));
      t:=IntToStr(qq+1);
      ti.Canvas.TextOut(ti.Width-rmargin-50, Round(y-14), t+' bit');

      qq:=qq+1;
      y:=vCenter+scale*(power(2,qq));

    end;



  end  // dmTime
  else begin // dmFreq
    ti.Canvas.Font.Name:=axisFontName;
    ti.Canvas.Font.Style:=axisFontStyle;
    ti.Canvas.Font.Size:=axisFontSize;
    ti.Canvas.Font.Color:=axisFontColor;

    unitString:='kHz';
    ti.Canvas.TextOut(lMargin+(plotWidth div 2)-Round(1.5*axisFontSize),
                      ti.Height-15-axisFontSize, unitString);

    if xini=0 then fini:= 500/dw
              else fini:= 500/dw-xini*1000/(dw*fal);
    t:=FloatToStrF(fini,ffGeneral,4,3);
    ti.Canvas.TextOut(lMargin-5,ti.Height-bMargin, t);
    if xfin=FAL then ffin:=-500/dw // in kHz
                else ffin:=-500/dw+(fal-1-xfin)*1000/(dw*fal);
    t:=FloatToStrF(ffin,ffGeneral,4,3);
    ti.Canvas.TextOut(ti.Width-rMargin-20,ti.Height-bMargin, t);

    t:=FloatToStrF((ffin+fini)/2,ffGeneral,4,3);
    ti.Canvas.TextOut(lMargin+(plotWidth div 2)-15,ti.Height-bMargin, t);

    t:=FloatToStrF((0.25*ffin+0.75*fini),ffGeneral,4,3);
    ti.Canvas.TextOut(lMargin+(plotWidth div 4)-15,ti.Height-bMargin, t);
    t:=FloatToStrF((0.75*ffin+0.25*fini),ffGeneral,4,3);
    ti.Canvas.TextOut(lMargin+(3*plotWidth div 4)-15,ti.Height-bMargin, t);

  end; //dmFreq


  ti.canvas.Pen.Width:=penWidth;
  // REAL
  if FIDReal in PlotOption then
        halfFIDDraw(FID[k].real,ti,realColor,dx,vCenter,scale);
  // IMAG
  if FIDImag in PlotOption then
        halfFIDDraw(FID[k].imag,ti,imagColor,dx,vCenter,scale);
  // ABS
  if FIDAbs in PlotOption then
        halfFIDDraw(FID[k].abs,ti,absColor,dx,vCenter,scale);


end;
//------------------------------------------------------------------------------
procedure TFID.setAL(a:cardinal);
begin
  FAL:=a;
  real.AL:=a;
  imag.AL:=a;
  abs.AL:=a;

end;

//------------------------------------------------------------------------------
procedure TFID.setDW(r: single);
begin
  FDW:=r;
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
constructor THalfFID.create(k:cardinal);
begin
  FAL:=k; // default al
  setLength(sig,FAL);
end;
//------------------------------------------------------------------------------
destructor THalfFID.free;
begin

end;
//------------------------------------------------------------------------------
procedure THalfFID.DCOffsetCut;
var k:integer;
    av: double;
begin
  av:=average;
  for k:=0 to al-1 do begin
    sig[k]:=sig[k]-av;
  end;
end;
//------------------------------------------------------------------------------
function THalfFID.Average: double;
begin
  result:=sum/al;
end;

//------------------------------------------------------------------------------
function THalfFID.Sum(kini: Integer; kfin: Integer):double;
var mini,mfin,k:integer;
    n: double;
begin
  if kini<0 then mini:=0 else mini:=kini;
  if kfin>integer(al)-1 then mfin:=integer(al)-1 else mfin:=kfin;
  n:=0;
  for k:=mini to mfin do begin
    n:=n+sig[k];
  end;
  result:=n;
end;
//------------------------------------------------------------------------------
function THalfFID.Sum: double;
var k: integer;
    r: double;
begin
  r:=0;
  for k:=0 to al-1 do begin;
    r:=r+sig[k];
  end;
  result:=r;
end;

//------------------------------------------------------------------------------
function THalfFID.AbsMaxPoint(kini,kfin:integer):integer;
var mini,mfin,k,p:integer;
    m,n: double;
begin
  if kini<0 then mini:=0 else mini:=kini;
  if kfin>integer(al)-1 then mfin:=integer(al)-1 else mfin:=kfin;

  m:=Abs(sig[mini]);
  p:=mini;

  for k:=mini+1 to mfin do begin
    n:=Abs(sig[k]);
    if m<n then begin m:=n; p:=k; end;
  end;

  result:=p;

end;
//------------------------------------------------------------------------------
function THalfFID.AbsMax: double;
begin
  result:=AbsMax(0,al-1);
end;
function THalfFID.AbsMax(kini,kfin:integer):double;
var mini,mfin,k:integer;
    m,n: double;
begin
  if kini<0 then mini:=0 else mini:=kini;
  if kfin>integer(al)-1 then mfin:=integer(al)-1 else mfin:=kfin;
  m:=Abs(sig[mini]);
  for k:=mini+1 to mfin do begin
    n:=Abs(sig[k]);
    if m<n then m:=n;
  end;
  result:=m;
end;
//------------------------------------------------------------------------------

function THalfFID.Average(kini,kfin:integer):double;
var mini,mfin,k:integer;
    n: double;
begin
  if kini<0 then mini:=0 else mini:=kini;
  if kfin>integer(al)-1 then mfin:=integer(al)-1 else mfin:=kfin;
  n:=0;
  for k:=mini to mfin do begin
    n:=n+sig[k];
  end;
  result:=n/(mfin-mini+1);
end;
//------------------------------------------------------------------------------
function THalfFID.standardDeviation:double;
begin
  result:=standardDeviation(0,al-1);
end;
//------------------------------------------------------------------------------
function THalfFID.standardDeviation(kini,kfin:integer):double;
var k:integer;
    av,sigma:double;
begin

  sigma:=0.0;
  if (kini<0) or (kfin>integer(al)-1) or (kini>kfin) then result:=0.0
  else begin
    av:=average(kini,kfin);
    for k := kini to kfin do sigma:=sigma+(sig[k]-av)*(sig[k]-av);
    sigma:=sigma/(kfin-kini+1);
    sigma:=sqrt(sigma);
    result:=sigma;
  end;

end;
//------------------------------------------------------------------------------
function THalfFID.Max: double;
begin
  result:=Max(0,al-1);
end;

function ThalfFID.Max(kini,kfin:integer): double;
var k: integer;
    m: double;
begin
  m:=sig[kini];
  for k:=kini+1 to kfin do begin
    if m<sig[k] then m:=sig[k]
  end;
  result:=m;
end;
//------------------------------------------------------------------------------
function THalfFID.Min: double;
begin
  result:=Min(0,al-1);
end;

function THalfFID.Min(kini,kfin:integer): double;
var k: integer;
    m: double;
begin
  m:=sig[kini];
  for k:=kini+1 to kfin do begin
    if m>sig[k] then m:=sig[k]
  end;
  result:=m;
end;

//------------------------------------------------------------------------------
procedure THalfFID.initialize;
var k:cardinal;
begin
  for k:=0 to al-1 do begin
    sig[k]:=0;
  end;
end;
//------------------------------------------------------------------------------
procedure THalfFID.setAL(a:cardinal);
begin
  FAL:=a; // default al
  setLength(sig,FAL);
end;
//------------------------------------------------------------------------------
procedure THalfFID.squareRoot;
var k:cardinal;
begin
  for k := 0 to al - 1 do sig[k]:=sqrt(abs(sig[k]));

end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
end.
