unit digitalFilter;

interface

uses classes, windows, extctrls, stdctrls, sysutils, graphics, math, StrUtils;

type TWindow = (hamming);

{-------------------------------------------------}
type TDigitalFilter=class(TObject)
  private
    FTaps: word;
    FCutoff: single;
    FBitLength: word;


   function h(k:integer):single;
   function HammingWindow(k:integer):single;

  public

    constructor Create; overload;
    constructor Create(n:cardinal); overload;
    destructor Free;

    function coeff(k:word):single;
    function truncatedCoeff(k:word):single;
    function HexCoeff(k:word): string;

    function HexCoeffArray: string;
    function CoeffArray: string;

    function frequencyResponse(wt:single):single;
    function truncatedFrequencyResponse(wt:single):single;

    procedure setTaps(k:word);
    procedure setCutOff(r:single);
    procedure setBitLength(k:word);

    procedure plotCoeff(paintbox:TPaintBox);
    procedure plotResponse(paintbox:TPaintBox);


    property Taps: word read FTaps write setTaps;
    property CutOff: single read FCutoff write setCutOff;
    property BitLength: word read FBitLength write setBitLength;

end;
{-------------------------------------------------}



implementation
{-------------------------------------------------}
{-------------------------------------------------}

//-------------------------------------------------
procedure TdigitalFilter.plotResponse(paintBox:TPaintBox);
var
  scale,x,y,dwt: single;
  hheight,offset,q:integer;
  plotWidth, hMargin, vMargin: integer;

begin
  hMargin:=25;  vMargin:=20;
  plotWidth:=paintBox.Width-(2*hMargin);
  dwt:= 3.141592/plotWidth;

  hheight:=Round(paintBox.Height/2);

  offset:=hheight div 4;
  scale:=hheight/100;  // Ƃ肠

  with PaintBox do begin
    Canvas.Brush.Color:=$00000000;
    Canvas.FillRect(Rect(0,0,Width,Height));
    Canvas.Pen.Width:=1;
    Canvas.Pen.Color:=$00555555;
    Canvas.Rectangle(Rect(hMargin,vMargin,Width-hMargin,Height-vMargin));

    for q:=-4 to 10 do begin;
      Canvas.Moveto(hMargin,Round(hheight-offset+scale*q*10));
      Canvas.Lineto(Width-hMargin,Round(hheight-offset+scale*q*10));
    end;  // q

    for q:=1 to 9 do begin
      Canvas.Moveto(Round(hMargin+q*PlotWidth/10), vMargin);
      Canvas.Lineto(Round(hMargin+q*PlotWidth/10), height-vMargin);
    end;

  end;

  with PaintBox do begin
    Canvas.Font.Color:=clLime;
    Canvas.TextOut(Width-hMargin-10, height-15,'2 M');
    Canvas.TextOut(hMargin+(plotWidth div 2)-10, height-15,'1 M');
    Canvas.TextOut(hMargin-5, height-15,'0');
    Canvas.TextOut(5, 15 ,'dB');
    Canvas.TextOut(5, Round(hheight-offset-scale*40-5) ,'+40');
    Canvas.TextOut(5, Round(hheight-offset-scale*20-5) ,'+20');
    Canvas.TextOut(10, hheight-offset-5 ,'0');
    Canvas.TextOut(5, Round(hheight-offset+scale*20-5) ,'-20');
    Canvas.TextOut(5, Round(hheight-offset+scale*40-5) ,'-40');
    Canvas.TextOut(5, Round(hheight-offset+scale*60-5) ,'-60');
    Canvas.TextOut(5, Round(hheight-offset+scale*80-5) ,'-80');
    Canvas.TextOut(5, Round(hheight-offset+scale*100-5) ,'-100');

    Canvas.Font.Color:=$005555EE;
    Canvas.TextOut(Width-hMargin-60,45,'Designed');
    Canvas.Font.Color:=$00FFFF00;
    Canvas.TextOut(Width-hMargin-60,30,'Actual');
  end; // with paintbox

  if Taps=0 then exit;  // no digital filter -> ڐ肾`悵ďI

  y:=scale*frequencyResponse(0);
  paintBox.Canvas.MoveTo(hMargin,Round(hheight-y - offset));
  paintBox.Canvas.Pen.Color:=$005555EE;

  x:=dwt;
  for q:=hMargin to paintBox.Width-hMargin do begin
    y:=scale*frequencyResponse(x);
    paintBox.Canvas.LineTo(q,Round(hheight-y - offset));
    x:=x+dwt;
  end; // q;


  y:=scale*truncatedFrequencyResponse(0);
  paintBox.Canvas.MoveTo(hMargin,Round(hheight-y - offset));
  paintBox.Canvas.Pen.Color:=$00FFFF00;

  x:=dwt;
  for q:=hMargin to paintBox.Width-hMargin do begin
    y:=scale*truncatedFrequencyResponse(x);
    paintBox.Canvas.LineTo(q,Round(hheight-y - offset));
    x:=x+dwt;
  end; // q;


end;


//-------------------------------------------------
procedure TdigitalFilter.plotCoeff(paintBox:TPaintBox);
var
  scale,x,y,yval: single;
  hheight,offset,q:integer;
  plotWidth, hMargin, vMargin: integer;
begin

  if Taps=0 then exit;

  hMargin:=25;  vMargin:=20;
  plotWidth:=paintBox.Width-2*hMargin;

//  if Taps>1 then dx:=plotWidth/(Taps-1) else dx:=paintBox.Width;
  hheight:=Round(paintBox.Height/2);

  offset:=hheight div 2;
  scale:=hheight/coeff(Round(((Taps-1)/2)));  // Ƃ肠

  with paintBox do begin
    Canvas.Brush.Color:=$00000000;
    Canvas.FillRect(Rect(0,0,Width,Height));
    Canvas.Pen.Width:=1;
    Canvas.Pen.Color:=$00555555;
    Canvas.Rectangle(Rect(hMargin,vMargin,hMargin+plotWidth,Height-vMargin));


    Canvas.Moveto(hMargin,Round(hheight+offset));
    Canvas.Lineto(hMargin+plotWidth,Round(hheight+offset));



    for q:=1 to 9 do begin
      Canvas.Moveto(Round(hMargin+q*PlotWidth/10), vMargin);
      Canvas.Lineto(Round(hMargin+q*PlotWidth/10), height-vMargin);
    end;

    Canvas.Font.Color:=clLime;
    Canvas.TextOut(Width-hMargin-5, height-15, IntToStr(Taps));
    Canvas.TextOut(hMargin+(plotWidth div 2)-5, height-15, IntToStr((Taps-1) div 2));
    Canvas.TextOut(hMargin-5, height-15,'1');

    Canvas.TextOut(10, Round(hheight+offset-5) ,'0');
    yval:=coeff((Taps-1) div 2);
    y:=scale*yval;
    Canvas.TextOut(5, Round(hheight-y) + offset -5,
           FloatToStrF(yval, ffGeneral, 5, 5));

    Canvas.Moveto(hMargin,Round(hheight+offset-y));
    Canvas.Lineto(hMargin+plotWidth,Round(hheight+offset-y));

    yval:=yval/2.0;
    y:=scale*yval;
    Canvas.TextOut(5, Round(hheight-y) + offset -5,
           FloatToStrF(yval, ffGeneral, 5, 5));
    Canvas.Moveto(hMargin,Round(hheight+offset-y));
    Canvas.Lineto(hMargin+plotWidth,Round(hheight+offset-y));


//    Canvas.TextOut(5, Round(hheight-offset+scale*20-5) ,'-20');
//    Canvas.TextOut(5, Round(hheight-offset+scale*40-5) ,'-40');
//    Canvas.TextOut(5, Round(hheight-offset+scale*60-5) ,'-60');
//    Canvas.TextOut(5, Round(hheight-offset+scale*80-5) ,'-80');
//    Canvas.TextOut(5, Round(hheight-offset+scale*100-5) ,'-100');

//    Canvas.Font.Color:=$005555EE;
//    Canvas.TextOut(Width-hMargin-60,45,'Designed');
 //   Canvas.Font.Color:=$00FFFF00;
 //   Canvas.TextOut(Width-hMargin-60,30,'Actual');
  end; // with paintbox


//  x:=hMargin+1;
//  paintBox.Canvas.MoveTo(Round(x), Round(hheight-scale*coeff(0)) + offset);
//  paintBox.Canvas.Pen.Color:=$005555EE;

//  for q:=1 to Taps-1 do begin
//    x:=hMargin + q*plotWidth/(Taps-1);
//    y:=scale*coeff(q);
//    paintBox.Canvas.LineTo(Round(x),Round(hheight-y) + offset);
//  end; // q;


  x:=hMargin+1;
//  y:=coeff(0) * (power(2,FBitLength-1)-1);
  y:=scale*truncatedCoeff(0);

  paintBox.Canvas.MoveTo(Round(x),Round(hheight-y) + offset);
  paintBox.Canvas.Pen.Color:=$00ffff00;

  for q:=1 to Taps-1 do begin
    x:=hMargin + q*plotWidth/(Taps-1);
    y:=scale*truncatedCoeff(q);
    paintBox.Canvas.LineTo(Round(x),Round(hheight-y)+ offset);
  end; // q;


end;
//-------------------------------------------------
procedure TdigitalFilter.setBitLength(k:word);
begin
  FBitLength:=k;
end;
//-------------------------------------------------
procedure TdigitalFilter.setCutOff(r:single);
begin
  FCutOff:=r;
end;
//-------------------------------------------------
procedure TdigitalFilter.setTaps(k:word);
begin
  FTaps:=k;
end;
//-------------------------------------------------
function TdigitalFilter.HexCoeffArray:string;
var
  i:integer;
  s:string;
begin
  s:='';

  for i:=0 to FTaps-1 do begin
    s:=s+HexCoeff(i);
  end; // i

  HexCoeffArray:=s;

end;
//-------------------------------------------------
function TdigitalFilter.CoeffArray:string;
var
  i:integer;
  s:string;
begin
  s:='';

  for i:=0 to FTaps-1 do begin
    s:=s+FloatToStr(Coeff(i));
    if i<FTaps-1 then s:=s+','
  end; // i

  CoeffArray:=s;

end;
//-------------------------------------------------
function TDigitalFilter.HexCoeff(k:word):string;
var p,digits:integer;
    s: string;
begin
  if (FBitLength mod 4)=0 then p:=0 else p:=1;
  digits:=  (FBitLength div 4) + p;
  s:=IntToHex(Round(coeff(k) * (power(2,FBitLength-1)-1)),digits);
  HexCoeff:=RightStr(s,digits);
end;
//-------------------------------------------------
function TDigitalFilter.coeff(k:word):single;
var shift:integer;
begin
  shift:=(FTaps-1) div 2;

  // NOTE:
  //    this is tentative ! (2014.1.9 Takeda)
  //
  //  result:=HammingWindow(k-shift)*h(k-shift)/(2*CutOff);




  result:=HammingWindow(k-shift)*h(k-shift);


end;
//-------------------------------------------------
function TDigitalFilter.truncatedCoeff(k:word):single;
begin
  truncatedCoeff:=
   Round(coeff(k) * (power(2,FBitLength-1)-1) ) / (power(2,FBitLength-1)-1);
end;
//-------------------------------------------------
function TDigitalFilter.h(k:integer):single;
var th,hh:single;
begin
  if k=0 then hh:=2.0*FCutOff
//  if k=0 then hh:=1.0
  else begin
    th:=6.283184*FCutoff*k;
//    hh:= sin(th)/th;
    hh:= 2.0*FCutOff*sin(th)/th;
  end;
  result:=hh;
end;
//-------------------------------------------------
function TDigitalFilter.HammingWindow(k:integer):single;
var w:single;
begin
  if Ftaps=1 then w:=1.0
  else w:=0.54 + 0.46*cos(6.283184*k/(Ftaps-1));
  Result:=w;
end;
//-------------------------------------------------
function TDigitalFilter.frequencyResponse(wt:single):single;
var r:single;
    k,center:word;

begin
  center:=(Ftaps-1) div 2;

  r:=coeff(center);
  for k:=1 to center do begin
    r:=r + 2 * coeff(center-k) * cos(k*wt);
  end; // k

  frequencyResponse:=20*log10(abs(r));
end;

//-------------------------------------------------
function TDigitalFilter.truncatedFrequencyResponse(wt:single):single;
var r:single;
    k,center:word;

begin
  center:=(Ftaps-1) div 2;

  r:=truncatedCoeff(center);
  for k:=1 to center do begin
    r:=r + 2 * truncatedCoeff(center-k) * cos(k*wt);
  end; // k

  if r=0 then truncatedFrequencyResponse := -60 else
  truncatedFrequencyResponse:=20*log10(abs(r));
end;

//-------------------------------------------------
//-------------------------------------------------
//-------------------------------------------------
//-------------------------------------------------
constructor TDigitalFilter.Create;
begin
    FTaps:=135;
    FCutoff:=0.5;
    FBitLength:=10;
end;

constructor TDigitalFilter.Create(n:cardinal);
begin
  FTaps:=n;
  FCutoff:=0.5;
  FBitLength:=10;
end;


destructor TDigitalFilter.Free;
begin

end;

{-------------------------------------------------}
end.
