unit gates;

//----------------------------------------------------------
interface

uses math,classes,inifiles,sysutils;

 // const maxBitLength=32;

  type TlineBits = 0..63;

  type TKindOfGate = (amplitude, phase, logic, logic_vector,
                     AD9858, phaselist, shape, int);

  type TlineBitsSet = set of TlineBits;

  type TKindOfShape = (shAmplitude, shPhase, shAP);

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TGateControlWord = class
    private
      FLength: word;
      FBits: array of TlineBits;
      FWord: TlineBitsSet;
      FKind: TKindOfGate;
      FChannel: word;

      procedure setBitLength(k:word);
      procedure setChannel(k:word);

    public
      name: string;
      caption: string;
      used: boolean;  {sŕgĂȂ`FbN邽}

      Offset: double; {AD9858݂̂Ɏg}
      property Length: word read FLength write setBitLength;
      property Kind: TKindOfGate read FKind;
      property channel: word read FChannel write setChannel;

      procedure gateAssign(m:word; l:TlineBits);
      function getAssignment(m:word): TlineBits;


      function output(r:single):TlineBitsSet; overload;virtual;abstract;
      function output(k:longInt):TlineBitsSet; overload;virtual;abstract;
      function output:TlineBitsSet; overload;virtual;abstract;
      function output(PS:ShortInt;f:double;k:integer):TlineBitsSet; overload;virtual;abstract;


    end;

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TshapeWord=class(TGateControlWord)

    order:word;
    KindOfShape:TKindOfShape;
    function output:TlineBitsSet; override;
    constructor Create(k:word; koShape:TKindOfShape); Overload;

    end;

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TphaseListWord=class(TGateControlWord)

    order:word;

    function output:TlineBitsSet; override;
    constructor Create(k:word); Overload;

    end;

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TampWord=class(TGateControlWord)

    function output(r:single):TlineBitsSet; override;
    constructor Create(k:word); Overload;

    end;

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TphaseWord=class(TGateControlWord)

    function output(r:single):TlineBitsSet; override;
    constructor Create(k:word); Overload;

    end;

{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TLogicWord=class(TGateControlWord)

    function output:TlineBitsSet; override;
    constructor Create; Overload;

    end;
{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TLogicVectorWord=class(TGateControlWord)

    function output(k:longInt):TlineBitsSet; override;
    constructor Create(k:word); Overload;

    end;
{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TIntWord=class(TGateControlWord)

    function output(k:integer):TlineBitsSet; override;
    constructor Create(k:word); Overload;

    end;
{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  type TAD9858Word=class(TGateControlWord)
    Offset: double;
    function output(PS:ShortInt;f:double;k:integer):TlineBitsSet; override;
    constructor Create(f:double); Overload;

    end;
{++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  type TGates = class
    private
      FLength: word;

    public
      gate: array of TGateControlWord;
      error: boolean;
      messages,info: TStrings;
      Definition: TStrings;

      property Count: word read FLength;

      procedure add(s,kind:string; len:word);overload;
      procedure add(s,kind:string);overload;
      procedure add(s,kind:string;f:double);overload;

      function readFromFile(fn:string):boolean;

      procedure delete(k:integer);
      function index(s:string):integer;

      constructor create;
      destructor free;

    end;
//----------------------------------------------------------
  function extractTLineBitsSet(s:string):TLineBitsSet;
//----------------------------------------------------------

implementation

//------------------------------------------------------------------------------
function extractTLineBitsSet(s:string):TLineBitsSet;
var k,m: integer;
    tm: set of 0..3;
    ls:TLineBitsSet;
begin
  ls:=[];
  for k:=1 to 16 do begin
    case s[k] of
      '0': tm:=[];
      '1': tm:=[0];
      '2': tm:=[1];
      '3': tm:=[0,1];
      '4': tm:=[2];
      '5': tm:=[0,2];
      '6': tm:=[1,2];
      '7': tm:=[0,1,2];
      '8': tm:=[3];
      '9': tm:=[0,3];
      'a','A': tm:=[1,3];
      'b','B': tm:=[0,1,3];
      'c','C': tm:=[2,3];
      'd','D': tm:=[0,2,3];
      'e','E': tm:=[1,2,3];
      'f','F': tm:=[0,1,2,3];
      else tm:=[];
    end; // case

    for m:=0 to 3 do begin
      if m in tm then ls:=ls+[m+(16-k)*4];
    end; // m
  end; // k
  result:=ls;
end;

{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
function TGates.index(s:string):integer;
var k,q:integer;
begin
  q:=-1;
  for k:=0 to Count-1 do begin
    if gate[k].name=s then q:=k;
  end;
  result:=q;
end;
{----------------------------------------------------------}
procedure TGates.add(s, kind:string; f:double);
var i:integer;
    ki:string;
begin
  i:=index(s);
  ki:=UpperCase(kind);
  if (i=-1) and (ki='AD9858') then begin
    inc(FLength);
    SetLength(gate,Count);
    gate[Count-1]:=TAD9858Word.Create(f);
    gate[Count-1].name:=s;

    // ǂŃItZbgȂƂȂ悤B
    // RXgN^ŏĂȂ̂ǁBBB
    // ( 2006.6.12)
    gate[Count-1].Offset:=f;

  end else begin
    error:=true;
    if i>-1 then
      messages.Add('Multiple declaration of the gate variable ' + s);

    if (ki<>'AD9858') then
       messages.Add('Only AD9858 '
             + 'is allowed.');

  end; // if

end;



procedure TGates.add(s, kind:string);
var i:integer;
    ki:string;
begin
  i:=index(s);
  ki:=UpperCase(kind);
  if (i=-1) and (ki='LOGIC') then begin
    inc(FLength);
    SetLength(gate,Count);
    gate[Count-1]:=TlogicWord.create;
    gate[Count-1].name:=s;

  end else begin
    error:=true;
    if i>-1 then
      messages.Add('Multiple declaration of the gate variable ' + s);
    if (ki<>'LOGIC') then
       messages.Add('Only LOGIC is allowed.');
  end;
end;

procedure TGates.add(s, kind:string; len:word);
var i:integer;
    ki:string;
begin
  i:=index(s);
  ki:=UpperCase(kind);
  if (i=-1) and ((ki='AMPLITUDE') or (ki='PHASE')
                  or (ki='LOGIC_VECTOR') or (ki='PHASELIST')
                  or (ki='AMPSHAPE') or (ki='PHASESHAPE')
                  or (ki='APSHAPE') or (ki='INTEGER')
                  )
  then begin
          {gp̖OȂOK else G[}
          {logiclenvȂ}
    inc(FLength);
    SetLength(gate,Count);
     if ki='AMPLITUDE' then gate[Count-1]:=TampWord.create(len)
     else if ki='PHASE' then gate[Count-1]:=TphaseWord.create(len)
     else if ki='LOGIC_VECTOR' then
               gate[Count-1]:=TlogicVectorWord.Create(len)
     else if ki='PHASELIST' then
               gate[Count-1]:=TPhaseListWord.Create(len)
     else if ki='AMPSHAPE' then
               gate[Count-1]:=TShapeWord.Create(len,shAmplitude)
     else if ki='PHASESHAPE' then
               gate[Count-1]:=TShapeWord.Create(len,shPhase)
     else if ki='APSHAPE' then
               gate[Count-1]:=TShapeWord.Create(len,shAP)
     else if ki='INTEGER' then
               gate[Count-1]:=TIntWord.Create(len);

     gate[Count-1].name:=s;
  end else begin
    error:=true;
    if i>-1 then
      messages.Add('Multiple declaration of the gate variable ' + s);
    if ki='LOGIC' then messages.Add('A logic word does not require length.');
    if (ki<>'AMPLITUDE') and (ki<>'PHASE')
       and (ki<>'LOGIC') and (ki<>'LOGIC_VECTOR')
       and (ki<>'PHASELIST') and (ki<>'AMPSHAPE')
       and (ki<>'PHASESHAPE') and (ki<>'APSHAPE')
       and (ki<>'INTEGER') then
       messages.Add('Only AMPLITUDE, PHASE, LOGIC, LOGIC_VECTOR, '
             + 'INTEGER, AMPSHAPE, PHASESHAPE, and APSHAPE '
             + 'are allowed.');
  end; // if
end;

{----------------------------------------------------------}
procedure TGates.delete(k:integer);
var q:integer;
begin
  for q:=k to Count-2 do begin
    gate[k]:=gate[k+1];
  end; // q
  gate[Count-1].Free;
  dec(FLength);
end;
{----------------------------------------------------------}
destructor Tgates.free;
begin
  messages.Free; info.Free;
  definition.Free;
  while FLength>0 do begin
    delete(FLength-1);
  end;
end;
{----------------------------------------------------------}
constructor TGates.create;
begin
  error:=false;
  messages:=TStringList.Create;
  info:=TStringList.Create;
  definition:=TStringList.Create;
  SetLength(gate,0);
  FLength:=0;
end;
{----------------------------------------------------------}
constructor TAD9858Word.Create(f:double);
begin
  FKind:=AD9858;
  Offset:=f;
  SetBitLength(17);
end;
{----------------------------------------------------------}
function TAD9858Word.output(PS:ShortInt;f:double;k:integer):TlineBitsSet;
var dat,add,wr,fud:TlineBitsSet;
    datOffset, addOffset:word;
    a:word;
    j,m: integer;
    FTW0: cardinal; //  Unsigned!!!
    DFTW: integer;  // Singed!!!  for frequency-sweep mode operation
begin
  dat:=[];
  add:=[];
  wr:=[];
  fud:=[];

  addOffset:=8;

  if (k=1) or (k=2) or (k=3) or (k=4) then wr:=[getAssignment(14)];
  if k=5 then fud:=[getAssignment(15)];

  if (k>0) and (k<5) then begin
    case k of
      1: a:= 10 + (PS*6); {AD9858݃AhX}
                {10=0x0A}
      2: a:= 10 + (PS*6) + 1;
      3: a:= 10 + (PS*6) + 2;
      4: a:= 10 + (PS*6) + 3;
      else a:= 0;
    end; {case k}

    for j:=0 to 5 do begin {AD9858̃AhX0..5}
      m:=(a shr j) and $1;
      if m=1 then add:=add+[getAssignment(j+addOffset)];
    end; {j}
    {kԖڂ̏݃AhXݒ肳ꂽ}

  FTW0:=round(f*power(2,32)/power(10,9)); //  cardinal, unsigned!!
  DFTW:=round(f*power(2,32)/power(10,9)); // integer, signed!!
//  DFTW:=round(f*power(2,31)/power(10,9));
//    ƎvĂǁIHApower(2,32)̂悤IH2011.6.20

  case k of
    1: datOffset:=0;
    2: datOffset:=8;
    3: datOffset:=16;
    4: datOffset:=24;
    else datOffset:=0;
  end; {case k of}

  case PS of
    -1: begin    // sweep mode
                 //  tinteger^f[^
          for j:=datOffset to datOffset+7 do begin
            if (DFTW shr j) and 1 = 1 then
            dat:=dat+[getAssignment(j-datOffset)];
          end;
        end;
    0..3: begin
            for j:=datOffset to datOffset+7 do begin
              if (FTW0 shr j) and 1 = 1 then
              dat:=dat+[getAssignment(j-datOffset)];
             end;
          end;
    end;
  end; {if (k>0) or (k<5)}

  result:= dat+add+wr+fud;

end;
{----------------------------------------------------------}
constructor TIntWord.Create(k:word);
begin
  FKind:=Int;
  SetBitLength(k);
end;
{----------------------------------------------------------}
function TIntWord.output(k:integer):TLineBitsSet;
var
  res: TLineBitsSet;
  j,q:integer;
begin
  res:=[];

  for j:= 0 to Length-1 do begin
    q := (k shr j) and $1;
    if q=1 then res := res + [getAssignment(j)];
  end;//j

  result:=res;
end;

{----------------------------------------------------------}
constructor TlogicVectorWord.Create(k:word);
begin
  FKind:=logic_vector;
  SetBitLength(k);
end;
{----------------------------------------------------------}
function TlogicVectorWord.output(k:longInt):TlineBitsSet;
var
  res: TlineBitsSet;
  j,q: cardinal;

begin
  res:=[];

  for j:= 0 to Length-1 do begin
    q := (k shr j) and $1;
    if q=1 then res := res + [getAssignment(j)];
  end;//j

  result:=res;
end;

{***********************************************************}

constructor TLogicWord.Create;
begin
  FKind:=logic;
  SetBitLength(1);
end;
{----------------------------------------------------------}
function TlogicWord.output:TlineBitsSet;
begin
  result:=[getAssignment(0)];
end;
{***********************************************************}
constructor TphaseWord.Create(k:word);
begin
  FKind:=phase;
  SetBitLength(k);
end;
{----------------------------------------------------------}
function TphaseWord.output(r:single):TlineBitsSet;
var
  res: TlineBitsSet;
  a,j,k,m: cardinal;
  q: single;

begin
  res:=[];
  m:=round(power(2,Length))-1;

  q:=r;
  while q>=360 do q:=q-360; //
  while q<0 do q:=q+360;    //
                            //q0ȏ360ɂȂ
  a:=round(m*q/360);

  for j:= 0 to Length-1 do begin
    k := (a shr j) and $1;
    if k=1 then res := res + [getAssignment(j)];
  end;//j

  result:=res;

end;
{***********************************************************}
constructor TampWord.Create(k:word);
begin
  FKind:=amplitude;
  SetBitLength(k);
end;
{----------------------------------------------------------}
function TampWord.output(r:single):TlineBitsSet;
var
  res: TlineBitsSet;
  a,j,k: cardinal;
  m:single;

begin
  res:=[];
  m:=power(2,Length)-1;

  if r>100.0 then begin
    r:=100.0;
  end;
  a:=Round(r*(m/100.0));

  for j:= 0 to Length-1 do begin
    k := (a shr j) and $1;
    if k=1 then res := res + [getAssignment(j)];
  end;//j

  result:=res;

end;
{***********************************************************}
{----------------------------------------------------------}
constructor TshapeWord.Create(k:word; koShape:TKindOfShape);
begin
  FKind:=shape;
  KindOfShape:=koShape;
  SetBitLength(1);  // ֌WȂAƂ肠PƂĂ
  order:=k; // Y`l̉ԖڂshapeH
end;
{----------------------------------------------------------}
function TShapeWord.output:TLineBitsSet;
begin
  result:=[];   //ł悢
end;
{----------------------------------------------------------}
{----------------------------------------------------------}
{----------------------------------------------------------}
function TPhaseListWord.output:TLineBitsSet;
begin
  result:=[];   //ł悢
end;
{----------------------------------------------------------}

constructor TPhaseListWord.Create(k:word);
begin
  FKind:=phaselist;
  SetBitLength(2);
  order:=k; // Y`l̉Ԗڂ̃XgH(1..7̂ǂꂩ)
end;

{***********************************************************}
{----------------------------------------------------------}
function TGates.readFromFile(fn:string):boolean;
var gateFile:TIniFile;
    s,t: string;
    i,k,p:integer;
    len:word;
    ok:boolean;
    f: double;
    sections: TStrings;
    name: string;
begin
  len:=1;
  if not FileExists(fn) then begin
    messages.Add('Gate definition file "' + fn +
                  '" is not found.'
                );
    ok:=false;
  end else begin
  gateFile := TIniFile.Create(fn);
  ok:=true;
  sections:=TStringList.Create;

  with gateFile do begin

    ReadSections(sections);

    for i:=0 to sections.Count-1 do begin

    name:=UpperCase(sections[i]);

    t := ReadString(name, 'BitLength', '');
    if t='' then begin
      messages.Add('In the gate definition file "' + fn +
                 '", the key "BitLength" is missing in the section [' +
                   name + ']');
      ok:=false;
    end else begin
      Val(t,f,p);
      if (trunc(f)-f<>0.0) or (f<1) then begin
        ok:=false;
        messages.Add('Invalid assignment of '+
                     '"BitLength" for the section [' + name + '] '+
                     'in the gate definition file "' + fn + '".');
      end else begin
        len:=StrToInt(t);
      end;
    end;

    s := UpperCase(ReadString(name, 'Kind', ''));
    if s<>'' then begin
      if s='AMPLITUDE' then add(name,'AMPLITUDE',len)
      else if s='PHASE' then add(name,'PHASE',len)
      else if s='LOGIC' then add(name,'LOGIC')
      else if s='LOGIC_VECTOR' then add(name,'LOGIC_VECTOR',len)
      else if s='INTEGER' then add(name,'INTEGER',len)
      else if s='AD9858' then begin
        t:=ReadString(name,'Offset','0');
        Val(t,f,p);
        add(name, 'AD9858',f);
      end
      else ok:=false;
    end else begin
      ok:=false;
      messages.Add('In the gate difinition file "' +
                  fn +
                  '", the key "Kind" is missing, or syntax is invalid.');
    end;



    if (s='AD9858') and (len<>16) then begin
      ok:=false;
      messages.Add('The bitLength has to be 16 for AD9858 gate parameter "'
                    + name + '".' +
                    ' (Length has been changed from 17 to 16 in Ver 0.958 and later)');
    end;

    if (s='logic') and (len<>1) then begin
      ok:=false;
      messages.Add('The bitLength has to be 1 for logic gate parameter "'
                     + name + '".');
    end;

    s:=ReadString(name,'channel','');
    if s='' then begin
      ok:=false;
      messages.Add('In the gate difinition file "' +
                  fn +
                  '", the key "Channel" is missing for the section ['
                  + name + '].');
    end else begin
      Val(s,f,p);
      if (trunc(f)-f<>0.0) or (f<1) then begin
        ok:=false;
        messages.Add('Invalid channel assignment for the section ['
                     + name + '] in the gate definition file "'
                     + fn + '".');
      end else begin
        gate[Count-1].FChannel:=trunc(f);
      end;
    end;

    for k:=0 to len-1 do begin
      s:=ReadString(name, name+'_'+IntToStr(k), '');
      if s<>'' then begin
        gate[Count-1].gateAssign(k, TlineBits(StrToInt(s)));
      end else begin
        ok:=false;
        messages.Add('Failed to assign the ' + IntToStr(k) + '-th bit '
                     + ' for the gate [' + name + ']');
      end;
    end;  // k

    gate[Count-1].caption:=ReadString(name,'caption','no caption');

    case gate[Count-1].Kind of
      amplitude: t:='an amplitude';
      phase: t:='a phase';
      logic: t:='a logic';
      logic_vector: t:='a logic_vector';
      int: t:='an integer';
      AD9858: t:='an AD9858'
                  + ' (Offset: '
                  + FloatToStr(gate[Count-1].offset)
                  + ' MHz) ';
    end; // case
      info.Add(' [Info] : ' + gate[Count-1].name + ' ('
               + gate[Count-1].caption + ') '
               + 'has been defined as ' + t + ' gate having '
               + IntToStr(gate[Count-1].Length)
               + ' bit length.' );

    end; // i

  end; // with gateFile
  gateFile.Free;
  sections.Free;
  end; // if fileExists

  if error=true then ok:=false;
  result:=ok;

end;
{----------------------------------------------------------}
procedure TGateControlWord.setChannel(k:word);
begin
  FChannel:=k;
end;
{----------------------------------------------------------}
procedure TGateControlWord.setBitLength(k:word);
begin
  SetLength(FBits,k);
  Flength:=k;
end;
{----------------------------------------------------------}
procedure TGateControlWord.gateAssign(m:word; l:TlineBits);
begin
  if m<FLength then FBits[m]:=l;
  if not (l in FWord) then FWord := FWord + [l];
end;
{----------------------------------------------------------}
function TGateControlWord.getAssignment(m:word): TlineBits;
begin
  result:=FBits[m];
end;

//----------------------------------------------------------
end.
