unit ppgCompiler;

interface

uses classes, sysutils, GATES, math;

type TUnitOfTime = (nanoSeconds,microSeconds,milliSeconds,Seconds,noUnit);
type TKindOfArray = (Fixed, Linear, Random);
type TType = (TimeVariable, LoopVariable, NAVariable,
              ampVariable, phaseVariable, intVariable,
              freqVariable);
type TAsyncMode=(asContinue, asStandby);
type lp_type=1..2;
{-------------------------------------------------}
type TVariable=class
  private
    FNOfChannels: integer;
    FNOfWhereIs: array of integer;
    FWhereIs: array of array of integer;
    FPSAD9858: array of array of shortInt;

    FGateIndex: array of array of integer;

    //TODO: FArrayDimension

    FarrayLength: integer;
    FarrayQ: boolean;
    FCurrentArrayIndex: integer;
    FMin, FMax: double;
    FDeclarationLine: integer;
   // function setLinearArray(curr:integer):string;
   // function setRandomArray(curr:integer):string;

  public
    Name: string;
    Caption: string;
    Value: double;      // ӖFValue=Numeric*unitOfTime

    kindOfArray:TKindOfArray;
    ArrayedValues: array of double;
    typeOfVar: TType;
    Numeric: double;
    unitOfTime: TUnitOfTime;
    unitOfTimeString: string;

    declarationPosition:integer; // sڂɐ錾Ă邩H
    updated:boolean;

    initialValue:double; // for linear array
    increment: double;   // for linear array

    constructor Create(nch:cardinal); OverLoad;
    constructor Create(nch:cardinal; s:string); Overload;

    function whereIs(k,m:integer):integer;
    function gateIndex(k,m:integer):integer;
    function psAD9858(k,m:integer):shortInt;

    function Count(k:integer):integer;
    procedure setarrayLength(k:integer);
    procedure setArrayQ(b:boolean);

    property channels:Integer read FNOfChannels;

    property arrayQ: boolean read FArrayQ write setArrayQ;
    property arrayLength: integer read FarrayLength write setArrayLength;
    property Min:double read FMin;
    property Max:double read FMax;
    property DeclarationLine:integer read FDeclarationLine;

 //   function setLinearArray(aini,ainc:double; an:cardinal);

end;

{-------------  under development: 2013Apr-... ---------------------}


type TArrayCounterElement=class
  private
    FArrayLength: integer;
    FNOfArrayedVariables: integer;

  public
    arrayedVariableIndex: array of integer;
    currentPosition: integer;

    constructor Create;
    destructor Free;

    procedure addArrayedVariable(varIndex:integer);
    procedure deleteArrayedVariable(varIndex:integer);overload;
    procedure deleteArrayedVariable;overload;
    function getIndex(varIndex:integer):integer;
    procedure setArrayLength(n:integer);

    property arrayLength: integer read FArrayLength write setArrayLength;
    property NOfArrayedVariables: integer read FNOfArrayedVariables;



end;


type TArrayCounter=class
  private
    FDimension:integer;

  public
    arrayCounterElement: array of TArrayCounterElement;
    finished: boolean;

    constructor Create;
    destructor Free;

    function isDefined(varIndex:integer):boolean;

    procedure reset;
    procedure toggle;
    procedure addDimension;
    procedure deleteDimension(k:integer);
    property dimension:integer read FDImension;

end;

{-------------------------------------------------}
type TVariables=class
  private
    FCount: integer;
    FNOfChannels: cardinal;

    FNOfArray: integer;
    FArrayLength: integer;



    procedure add(s:string);overload;

  public
    arrayCounter: TarrayCounter;
    Variable: array of TVariable;
    arrayedVariableIndex: array of Integer;
    constructor Create(nch:cardinal);
    destructor Free;

    procedure add(s:string; t:TType); overload;
    function defined(s:string):boolean;
    function index(s:string):integer;
    procedure delete(k:integer); overload;
    procedure delete; overload;

    procedure setNOfArray(k:integer);
    procedure setArrayLength(k: integer);

    property count:integer read FCount;
    property NOfArray: integer read FNOfArray write setNOfArray;
    property arrayLength: integer read FArrayLength write setArrayLength;

end;
{-------------------------------------------------}
type auxParameters=class
  private
    FAL: cardinal;
    FDW: double;
    FACQTIME: double;

  public
    PD: double;
    definePD: boolean;
    NA: cardinal;
    defineNA: boolean;
    defineAL: boolean;
    defineDW: boolean;

    constructor Create;

    procedure setAL(a:cardinal);
    procedure setDW(d:double);

    property AL: cardinal read FAL write setAL;
    property DW: double read FDW write setDW;
    property ACQTIME: double read FACQTIME;

end;

{-------------------------------------------------}
type TShapedPulse=class
protected
  FKindOfShape: TKindOfShape;
  FChannel: word;
  FLength: word;
  FName: string;
  FMaxLength: cardinal;
  FOrder: word;  // `lŉԖڂshape
public
  amp: array of single;
  phase: array of single;
  address: word;
  updateCount: word;

  constructor Create(name:string; channel:word; order:word); overload;
  constructor Create(name:string; channel:word; order:word; koShape:TkindOfShape); overload;

  destructor Free;
  procedure Add(ra:single; rp:single);
  function Delete(k:word):ShortInt; overload;
  procedure Delete; overload;

  property Channel: word read FChannel;
  property Name:string read FName;
  property Count:word read FLength;
  property Order:word read FOrder;
  property KindOfShape:TKindOfShape read FKindOfShape;

  function nOfLines: integer;  // f[^]ɉsKvH
  function genBin(k:word):string;
end;
{-------------------------------------------------}
type TShapedPulseArray=class
   private
     FCount: word;
     FNOfChannels: word;
   public
     Shape: array of TShapedPulse;
     source: TStrings;
     err:boolean;
     msg: string;
     constructor Create(n:word);
     destructor Free;
     property count: word read FCount;
     function NOfLists(channel:word):word;
     function Add(s:string; kos:TKindOfShape):integer;

     function index(s:string):integer; overload;
     function index(ch, order:word):integer; overload;

     procedure delete(k:word); overload;
     procedure delete; overload;


//     procedure gen(channel:word);overload;
//     procedure gen; overload;

end;
{-------------------------------------------------}
type TPhaseList=class
  private
    FChannel: word;
    FLength: word;
    FName: string;
    FMaxLength: cardinal;
    FOrder: word;  // `lŉԖڂphaseList(1..7)

  public
    Phase: array of single;
    Position: word; // 1..7

    constructor Create(name:string; channel:word; order:word);
    destructor Free;
    function Add(r:single):ShortInt;
    function Delete(k:word):ShortInt; overload;
    procedure Delete; overload;

    property Channel: word read FChannel;
    property Name:string read FName;
    property Count:word read FLength;
    property Order:word read FOrder;


end;
{-------------------------------------------------}
type TPhaseListArray=class
   private
     FCount: word;
     FNOfChannels: word;
   public
     phaseList: array of TPhaseList;
     source: TStrings;
     err:boolean;
     msg: string;
     constructor Create(n:word);
     destructor Free;
     property count: word read FCount;
     function NOfLists(channel:word):word;
     function Add(s:string):integer;

     function index(s:string):integer; overload;
     function index(ch, order:word):integer; overload;

     procedure delete(k:word); overload;
     procedure delete; overload;

     procedure gen(channel:word);overload;
     procedure gen; overload;

end;
{-------------------------------------------------}
type TAcqPhaseList=class
   private
     FCount:word;
     FMaxLength: cardinal;
   public
     source: TStrings;
     msg: string;
     Phase: array of word;

     constructor Create;
     destructor Free;
     function Add(w:word):integer;
     function ReadFromString(s:string):integer;
     property count: word read FCount;

     procedure gen;

end;
{-------------------------------------------------}
type TAsyncPPG=class
  private
    FNOfWhereIs: integer;
    FWhereIs: array of integer;
  public
    name:string;
    channel:word;
    sequence:TStrings;
    address: integer;
    used: boolean;
    asyncMode: TasyncMode;
    constructor Create;
    destructor Free;
    function whereIs(k:integer):integer;
    property Count:integer read FNOfWhereIs;
end;
{-------------------------------------------------}
type TAsyncPPGList=class
  public
    length: integer;
    asyncPPG: array of TAsyncPPG;
    itemIndex: integer;
    function index(name:string):integer;
    procedure add(a:TAsyncPPG);
    procedure delete;
    constructor Create;
    destructor Free;
end;
{-------------------------------------------------}
type TnmrPPG=class
  private
    lb:string;
    cnum,cnump,lnum,lnump: integer;  { CURRENT LINE, CURRENT COLUMN}
    lnum_bak:integer;
    ch: char;
    EndOfFile:boolean;
    err_str: string;
    Yes_Loop: boolean;
    Yes_AD9858: boolean; {舵ĂQ[gAD9858ǂ̔tO}
    AD9858Buffer: array[0..2] of array[0..4] of TLineBitsSet;
    lp:array[0..2] of boolean;
    FNOfChannels: word;
    currentCH: word;
    asyncQ: boolean;
    asyncFlag: boolean;
    lineOffset,phListArrayOffset: integer;
        {G[݂s̊ZɎg}
    shapeQ: boolean;
    shapedPulseIndex:integer;
   // TLQ: TType;

    imported: boolean;

    Uline,Lline:cardinal; {pulse, waitR}hŎgBoutput port̕ێp}

    procedure error(N:integer);
 //   procedure Check_N(ch:char; var i:integer);
    procedure Nextln;
    procedure getch;
    procedure Yomitobashi;
    function checkCH(c:char):Boolean;
    function Search(s:string):Boolean;
    procedure getFromLoops(var N:double);
    procedure getFromTimes(var N:double);
    procedure Number(var X:double);
    procedure Number_Time(var X,Num: double;
                          var unitOfTime:TUnitOfTime);
    procedure Jikann_Hennkann(X:double; var Uwidth:word; var Lwidth:cardinal);
    procedure duration(var Uwidth:word; var Lwidth:cardinal);

    procedure Get_S1S2(var X:double; var S1,S2:string);

    procedure SetShape(kos:TKindOfShape);
    procedure SetPhaseList;
    procedure SetAcqPhase;
    procedure ampVariables;
    procedure intVariables;
    procedure phaseVariables;
    procedure loopVariables;
    procedure timeVariables;
    procedure freqVariables;
//    procedure gateDefine;
    procedure setAuxParams;
    procedure auxCom(var ok:boolean);

    procedure initial;

    procedure Examine_Gates;
      //--- procedures called from Examine_Gates ---
      // i: ID number of the "gate" in the "gates"
      procedure setAmplitudeGate(i:integer);
      procedure setPhaseGate(i:integer);
      procedure setPhaseListGate(i:integer);
      procedure setIntGate(i:integer);
      procedure setLogicGate(i:integer);
      procedure setLogicVectorGate(i:integer);
      procedure setAD9858Gate(i:integer);
      //--------------------------------------------

    procedure init;
    procedure import;
    procedure pulse;
    procedure burst;
    procedure delay;
    procedure wait_hl(hl:byte);
    procedure getLoopUnit(var N:lp_type);
    procedure loop;
    procedure endLoop;
    procedure async;
    procedure sync;

   // procedure accum;

    procedure relax;

    procedure symbol;
    procedure assemble;


    procedure updateTimeVariable(q:integer; d: double; ut:TunitOfTime);
    procedure updateLoopVariable(q:integer; d: integer);
    procedure updateIntVariable(q:integer; d: integer);
    procedure updateAmpVariable(q:integer; d: double);
    procedure updatePhaseVariable(q:integer; d: double);
    procedure updateFreqVariable(q:integer; d: double);

  public
    CLK: double;  // CLK frequency in MHz  (new in 2008 Aug)
    lineset: TlineBitsSet;
    workingLines: TStrings;
    importedLines: TStrings;

    ppgLines: TStrings; {Ăяo͂ɃpXvO}
    preamble: TStrings; {vAui[}
    mainPPG: TStrings; {ppgLinesasynchronousCi[}
    asyncPPGList: TasyncPPGList;
    compiledPPG: array of TStrings;
                       {ɃRpCς݃pXvOi[}
                       {channelƂɗpӁB}

    NOfLinesPerCommand: array of TStrings;
         {AD9858䖽ߓApulsesKvȏꍇ̏L^}
    mnemonic: array of TStrings;    {Ƃi[B}
    Messages, info, Warnings: TStrings;
    reservedWords: TStrings;
    Variables: TVariables;
    updateCommand: TStrings; // ϐ̕ύXFPGAɑR}hi[

    phListArray: TPhaseListArray;
    acqPhaseList: TAcqPhaseList;

    shapedPulseArray: TShapedPulseArray;

    gates:TGates;
    auxParams: auxParameters;
    erra,compiled:boolean;

    ppgFileName:string; // ppg̃t@C
    ppgFilePath:string; // ppgt@ĈfBNg̃pX

    gateDefFileName: string; {Q[g`t@C}
    gateDefFileDir: string;
         //  New!  2009.2.9
         //  exet@CƓfBNgɒuƂɃ[ύX(2009.2.9)
         //

    pathDefined: boolean;  {Qd``FbNpBcompile̍ۂfalseɏB}

    importFileNames: TStrings; // import file list
    currentImportedFileIndex:integer;

    acqPhaseDefined: boolean; {acqphaseZbgĂ邩`FbN邽߂̃tO}

    constructor Create(NOfChannels: word);overload;
    constructor Create(NOFChannels: word; fn:string);overload;

    destructor Free;
    procedure compile;


    function loadFromFile(fn:string):integer;

    procedure updateVariable(variableName:string; num:double; uni:TUnitOfTime);overload;
    procedure updateVariable(variableName:string; num:double);overload;
 //   procedure updateVariable(variableName:string; num:integer);overload;

    property NOFChannels: word read FNOfChannels;
end;

//function Hex(n:longint; d:integer):string; overload;
function Hex(n:cardinal; d:integer):string;


{-------------------------------------------------}

implementation
{-------------------------------------------------}

constructor TarrayCounterElement.Create;
begin
  FarrayLength:=0;
  FnOfArrayedVariables:=0;
  currentPosition:=0;
  setLength(arrayedVariableIndex,0);
end;

destructor TarrayCounterElement.Free;
begin
  setLength(arrayedVariableIndex,0);
end;

//------------------------------------------------------------
procedure TarrayCounterElement.setArrayLength(n: Integer);
begin
  FArrayLength:=n;
end;
//------------------------------------------------------------
//
function TarrayCounterElement.getIndex(varIndex:integer):integer;
var k,q:integer;
begin
  q:=-1;
  for k:=0 to FNOfArrayedVariables-1 do begin
    if varIndex=arrayedVariableIndex[k] then begin
      q:=k;
      break;
    end;
  end;
  result:=q;
end;
//
//------------------------------------------------------------
//
//  This does not anything if the variable has already been set.
//
//
procedure TarrayCounterElement.addArrayedVariable(varIndex:integer);
begin
  if getIndex(varIndex)=-1 then begin
    Inc(FNOfArrayedVariables);
    setLength(arrayedVariableIndex,FNOfarrayedVariables);
    arrayedVariableIndex[FNOfArrayedVariables-1]:=varIndex;
  end;
end;
//------------------------------------------------------------
//
//
procedure TArrayCounterElement.deleteArrayedVariable(varIndex: Integer);
var k,q:integer;
begin
  q:=getIndex(varIndex);
  if q<>-1 then begin
    for k := q to NOfArrayedVariables-2 do
      arrayedVariableIndex[k]:=arrayedVariableIndex[k+1];
    dec(FNOfArrayedVariables);
    setLength(arrayedVariableIndex,NOfArrayedVariables);

  end;
end;


procedure TArrayCounterElement.deleteArrayedVariable;
begin
  FNOfArrayedVariables:=0;
  setLength(arrayedVariableIndex,0);
end;


constructor TArrayCounter.Create;
begin
  FDimension:=0;
  finished:=false;
  setLength(arrayCounterElement,0);
end;

destructor TarrayCounter.Free;
begin
  while FDimension>1 do deleteDimension(FDimension-1);
end;

function TArrayCounter.isDefined(varIndex: Integer):boolean;
var k:integer;
    ans:boolean;
begin
  ans:=false;
  for k := 0 to FDimension-1 do begin
    if arrayCounterElement[k].getIndex(varIndex)<>-1 then begin
      ans:=true;
      break;
    end;
  end;

  result:=ans;

end;

procedure TArrayCounter.addDimension;
begin
  Inc(FDimension);
  setLength(arrayCounterElement,FDimension);
  arrayCounterElement[FDimension-1].Create;
end;


procedure TArrayCounter.deleteDimension(k: Integer);
var i:integer;
begin
  for i := k to FDimension-2 do
    arrayCounterElement[i]:=arrayCounterElement[i+1];
  arrayCounterElement[FDimension-1].Free;
  Dec(FDimension);
  setLength(arrayCounterElement,FDimension);
end;


procedure TArrayCounter.reset;
var cdim:integer;
begin
  finished:=false;
  for cdim := 0 to FDimension-1 do
    arrayCounterElement[cdim].currentPosition:=0;
end;

procedure TArrayCounter.toggle;
var cdim:integer;
begin
  for cdim := 0 to FDimension-1 do begin
    if arrayCounterElement[cdim].currentPosition
       =arrayCounterElement[cdim].arrayLength-1 then begin
          arrayCounterElement[cdim].currentPosition:=0;
          if cdim=FDimension-1 then finished:=true;

    end else begin

      arrayCounterElement[cdim].currentPosition
        :=arrayCounterElement[cdim].currentPosition+1;

      break;

    end;

  end;

end;

{-------------------------------------------------}
constructor TAcqPhaseList.Create;
begin
  source:=TStringList.Create;
  FCount:=0;
  FMaxLength:=128;
  SetLength(Phase,FCount);
end;
{-------------------------------------------------}
destructor TAcqPhaseList.Free;
begin
  source.Free;
end;
{-------------------------------------------------}
function TAcqPhaseList.Add(w:word): integer;
begin
  if Count=FMaxLength then Add:=-1  // ȏ㑫ȂI
  else begin
    Inc(FCount);
    SetLength(Phase,FCount);
    Phase[FCount-1]:=w;
    Add:=0;
  end;
end;
{-------------------------------------------------}
function TAcqPhaseList.ReadFromString(s:string):integer;
// :  ACQ_PHASE=x,-Y,+Y,-x;
var
  t:string;
  k:word;
  found, finished:boolean;
  ph:word;
begin
  k:=0; ph:=0;
  finished:=false;
  repeat
    t:=''; found:=false;
    repeat
      inc(k);
      if s[k]<>' ' then begin
        if (s[k]=',') or (s[k]=';') then found:=true
        else t:=t+UpperCase(s[k]);
      end;
    until found or (k=Length(s));

    if t='' then begin  // ۂ
      msg:='Acq Phase value expected before ' + s[k] + '.';
      Result:=-1;
      exit;
    end;

    if (t='+X') or (t='X') then ph:=0
    else if (t='+Y') or (t='Y') then ph:=1
    else if (t='-X') then ph:=2
    else if (t='-Y') then ph:=3
    else begin // l
      msg:='Invalid acquisition phase value (should be either X,Y,-X, or -Y).';
      Result:=-1;
      exit;
    end;

    if Add(ph)<>0 then begin
      msg:='Phase-list length exceeds the maximum value of 32.';
      Result:=-1;
      exit;
    end;

    if s[k]=';' then finished:=true;

  until finished;

  Result:=0; // G[Ȃ̎A[ԂB

end;
{-------------------------------------------------}
procedure TAcqPhaseList.gen;
var q:word;
begin

  source.Clear;
  if Count=0 then exit;

  source.Add('pulse(1u;                 ACQ_PHASE_CS, F1PhaseMode(2), '
                                        + 'F1PhaseListLength('
                                        + IntToStr(Count) + '))');
  source.Add('pulse(1u; F1PhaseCommand, ACQ_PHASE_CS, F1PhaseMode(2), '
                                        + 'F1PhaseListLength('
                                        + IntToStr(Count) + '))');
  source.Add('pulse(1u; F1PhasePointer_Init, ACQ_PHASE_CS, F1PhaseMode(1), '
                                        + 'ACQ_PHASE('
                                        + IntToStr(Phase[0]) +'))');
  source.Add('pulse(1u; F1PhaseCommand, ACQ_PHASE_CS, F1PhaseMode(1), '
                                        + 'ACQ_PHASE('
                                        + IntToStr(Phase[0]) +'))');

  for q:=1 to Count-1 do begin
    source.Add('pulse(1u; F1TogglePhase, ACQ_PHASE_CS, F1PhaseMode(1), '
                                        + 'ACQ_PHASE('
                                        + IntToStr(Phase[q]) +'))');
    source.Add('pulse(1u; F1PhaseCommand, ACQ_PHASE_CS, F1PhaseMode(1), '
                                        + 'ACQ_PHASE('
                                        + IntToStr(Phase[q]) +'))');
  end; // q

end;

//------------------------------------------------------------------------------
constructor TShapedPulseArray.Create(n:word);
begin
  FNOfChannels:=n; // channel
  FCount:=0;
  SetLength(Shape,0);
  err:=false;
  msg:='';
  Source:=TStringList.Create;
  Source.Clear;
end;
//------------------------------------------------------------------------------
destructor TShapedPulseArray.Free;
begin
  Delete;
  Source.Free;
end;
//------------------------------------------------------------------------------
function TShapedPulseArray.NOfLists(channel:word):word;
var k,c:integer;
begin
  c:=0;
  for k:=0 to Count-1 do begin
    if Shape[k].Channel=channel then inc(c);
  end;
  result:=c;
end;
//------------------------------------------------------------------------------
function TShapedPulseArray.Add(s:string; kos:TKindOfShape):integer;
// s̗:  shape1=(ch1;100.0,90.0,85.3,23.5,.....)
var
  name:string;
  n,channel:word;
  t:string;
  j,k,q:integer;
  found, finished:boolean;
  amp,ph:single;
  P:integer;
  address: cardinal;
begin
  amp:=0.0; ph:=0.0;
  // O𒊏onameɊi[
  k:=0; found:=false;
  repeat
    inc(k); t:=s[k];
    if s[k]='=' then found:=true;
  until found or (k=Length(s));

  if not found then begin {error}
    msg:='"=" expected.';
    Add:=-1;
    exit;
  end;

  name:=''; q:=0;
  repeat
    inc(q);
    if s[q]<>' ' then name:=name+UpperCase(s[q]); //Xy[XȂ
  until q=k-1;

  if index(name)<>-1 then begin //OdĂI
    msg:='Duplicate identifier (' + name + ').';
    Add:=-1;
    exit;
  end;

  if (Ord(name[1]) in [Ord('0')..Ord('9')]) then begin //Ŏn܂ĂH
    msg:='Shaped pulse identifier must not start with a number.';
    Add:=-1;
    exit;
  end;


  // `lNo𒊏ochannelɊi[
  repeat
    inc(k);
  until (s[k]<>' ') or (k=Length(s));
  if (s[k]<>'(') or (k=Length(s)) then begin
    msg:='"(" expected.';
    Add:=-1;
    exit;
  end;

  t:=''; found:=false;
  repeat
    inc(k);
    if s[k]<>' ' then t:=t+UpperCase(s[k]);
    if t='CH' then found:=true;
  until found or (k=Length(s));

  if not found then begin {error}
    msg:='"CH" expected.';
    Add:=-1;
    exit;
  end;

  t:=''; found:=false;
  repeat
    inc(k);
    if s[k]<>' ' then begin
      if s[k]=';' then found:=true
      else if not(Ord(s[k]) in [Ord('0')..Ord('9')]) then begin
        msg:='Channel No. expected.';
        Add:=-1;
        exit;
      end else begin
        t:=t+s[k];
      end;
    end;
  until found or (k=Length(s));
  if not found then begin {error}
    msg:='";" expected after channel identifier.';
    Add:=-1;
    exit;
  end;

  channel:=StrToInt(t);
  if channel>FNOfChannels then begin
    msg:='The specified channel no. '
         + t + ' exceeds the number of available channels.';
    Add:=-1;
    exit;
  end;

  n:=NOfLists(channel);
//  if n=7 then begin // Y`lɗpł
                                    // g؂Ă
//    msg:='Channel '
//         + IntToStr(channel)
//         + ' has already had the maximum definabe phase lists.';
//    Add:=-1;
//    exit;
//  end;


  // starting address
  address:=0;
  j:=0;
  while j<FCount do begin
    if shape[j].Channel=channel then address:=address + shape[j].count;
    inc(j);
  end; // while

  // ̂ƂG[ȂAǉI
  Inc(FCount);
  SetLength(Shape,FCount);
  Shape[FCount-1]:=TShapedPulse.Create(name,channel,n,kos);
       // Y`lŁANOfLists(channel)Ԗڂshaped pulseƂȂ
       // ŏ0ŁA1,2,3...ƂÂ
  finished:=false;
  repeat


    if (kos=shAmplitude) or (kos=shAP) then begin

        t:=''; found:=false;
        repeat
          inc(k);
          if s[k]<>' ' then begin
            if (s[k]=',') or (s[k]=')') then found:=true
            else t:=t+UpperCase(s[k]);
          end;
        until found or (k=Length(s));

        if t='' then begin  // ۂ
          msg:='amplitude value expected before ' + s[k] + '.';
          Add:=-1;
          Shape[FCount-1].Free;  // Œǉshaped pulseB
          Dec(FCount);
          SetLength(shape,FCount);
          exit;
        end;


        Val(t,amp,P);
        if P<>0 then begin
          msg:='Invalid value.';
          shape[FCount-1].Free;
          Dec(FCount);
          SetLength(shape,FCount);
          Add:=-1;
          exit;
        end;

    end; // if (kos=shAmplitude) or (kos=shAP)

    if (kos=shPhase) or (kos=shAP) then begin
        t:=''; found:=false;
        repeat
          if k=Length(s) then break;
          inc(k);
          if s[k]<>' ' then begin
            if (s[k]=',') or (s[k]=')') then found:=true
              else t:=t+UpperCase(s[k]);
            end;

        until found or (k=Length(s));

        if t='' then begin  // ۂ
          msg:='phase value expected before ' + s[k] + ' .';
          Add:=-1;
          Shape[FCount-1].Free;  // Œǉshaped pulseB
          Dec(FCount);
          SetLength(shape,FCount);
          exit;
        end;


        Val(t,ph,P);
        if P<>0 then begin
          msg:='Invalid value.';
          shape[FCount-1].Free;
          Dec(FCount);
          SetLength(shape,FCount);
          Add:=-1;
          exit;
        end;


    end; // if (kos=shPhase) or (kos=shAP)

    case kos of
      shAmplitude: shape[FCount-1].Add(amp,0.0);
      shPhase: shape[FCount-1].Add(0.0,ph);
      shAP: shape[FCount-1].Add(amp,ph);
    end; // case;


    if s[k]=')' then finished:=true;
  until finished;

  while k<Length(s) do begin
    inc(k);
    if s[k]<>' ' then begin // Xy[XȊO͎cĂȂ͂B΃G[B
      msg:='Invalid expression: ' + s[k] + '.';
        shape[FCount-1].Free;
        Dec(FCount);
        SetLength(shape,FCount);
        Add:=-1;
        exit;
    end;
  end; // while;

  Shape[FCount-1].address:=address;
  Shape[FCount-1].updateCount:=1;  // Ƃ肠1ƂB
                          // pulseR}hɕύX邱ƂɂȂB

  Add:=0; // G[Ȃ̎A[ԂB
end;
//------------------------------------------------------------------------------
function TShapedPulseArray.index(s:string):integer;
var k,m:integer;
begin
  m:=-1;
  for k:=0 to Count-1 do begin
    if Shape[k].Name=s then begin m:=k; break; end;
  end;
  result:=m;
end;
//------------------------------------------------------------------------------
function TShapedPulseArray.index(ch, order:word):integer;
var q,m:integer;
begin
  m:=-1;
  for q:=0 to Count-1 do begin
    if (Shape[q].Channel=ch) and (Shape[q].Order=order) then begin
      m:=q; break;
    end;
  end;
  result:=m;
end;
//------------------------------------------------------------------------------
procedure TShapedPulseArray.delete(k:word);
var m:integer; // ͕łȂĂ͂ȂȂI
               // ȂƉfor q:=0 to -1 do
               // ƂŎsĂقȂꍇsĂ܂I
begin
  if k<FCount then begin
    for m:=k to FCount-2 do begin
      Shape[k]:=Shape[k+1];
    end;
    Shape[FCount-1].Free;
    Dec(FCount);
    SetLength(Shape,FCount);
  end;
end;
//------------------------------------------------------------------------------
procedure TShapedPulseArray.delete;
begin
  while FCount>0 do delete(FCount-1);
end;
//------------------------------------------------------------------------------

//     procedure TShapedPulseArray.gen(channel:word);overload;
//------------------------------------------------------------------------------
//     procedure TShapedPulseArray.gen; overload;

//------------------------------------------------------------------------------

constructor TPhaseListArray.Create(n:word);
begin
  FNOfChannels:=n; // channel
  FCount:=0;
  SetLength(phaseList,0);
  err:=false;
  msg:='';
  Source:=TStringList.Create;
  Source.Clear;
end;
{-------------------------------------------------}
destructor TPhaseListArray.Free;
begin
  Delete;
  Source.Free;
end;
{-------------------------------------------------}
procedure TPhaseListArray.delete;
begin
  while FCount>0 do delete(FCount-1);
end;
{-------------------------------------------------}
procedure TPhaseListArray.delete(k:word);
var m:integer; // ͕łȂĂ͂ȂȂI
               // ȂƉfor q:=0 to -1 do
               // ƂŎsĂقȂꍇsĂ܂I
begin
  if k<FCount then begin
    for m:=k to FCount-2 do begin
      phaseList[k]:=phaseList[k+1];
    end;
    phaseList[FCount-1].Free;
    Dec(FCount);
    SetLength(phaseList,FCount);
  end;
end;
{-------------------------------------------------}
procedure TPhaseListArray.gen;
var i,j:integer;
    sl: TStrings;
begin
  sl:=TStringList.Create;
  sl.Clear;
  for i:=1 to FNOfChannels do begin
    gen(i);
    for j:=0 to Source.Count-1 do begin
      sl.Add(Source.Strings[j]);
    end; // j
  end; // i

  Source.Assign(sl);
  sl.Free;
end;
{-------------------------------------------------}
procedure TPhaseListArray.gen(channel:word);
{
  pulse(dt; F1PhaseCycle_init)
  pulse(dt;                 F1PhaseList(1), F1PhaseMode(2), F1PhaseListLength(4))
  pulse(dt; F1PhaseCommand, F1PhaseList(1), F1PhaseMode(2), F1PhaseListLength(4))
  pulse(dt; F1PhasePointer_Init, F1PhaseList(1), F1PhaseMode(1), F1Phase(0.0))
  pulse(dt; F1PhaseCommand,      F1PhaseList(1), F1PhaseMode(1), F1Phase(0.0))
  pulse(dt; F1TogglePhase,   F1PhaseList(1), F1PhaseMode(1), F1Phase(90.0))
  pulse(dt; F1PhaseCommand,  F1PhaseList(1), F1PhaseMode(1), F1Phase(90.0))
  pulse(dt; F1TogglePhase,   F1PhaseList(1), F1PhaseMode(1), F1Phase(180.0))
  pulse(dt; F1PhaseCommand,  F1PhaseList(1), F1PhaseMode(1), F1Phase(180.0))
  pulse(dt; F1TogglePhase,   F1PhaseList(1), F1PhaseMode(1), F1Phase(270.0))
  pulse(dt; F1PhaseCommand,  F1PhaseList(1), F1PhaseMode(1), F1Phase(270.0))
}
var ch,dt:string;
     k,p,q:integer;
begin
  dt:='1u';
  ch:=IntToStr(channel);
  Source.Clear;
  if NOfLists(channel)>0 then
     Source.Add('pulse(' + dt + '; ' + 'F'+ch+'PhaseCycle_Init)');
  for k:=1 to NOfLists(channel) do begin
    p:=index(channel,k);
    Source.Add('pulse(' + dt + '; ' + 'F'+ch+'PhaseList('+ IntToStr(k) +'), '
                 + 'F'+ch+'PhaseMode(2), '
                 + 'F'+ch+'PhaseListLength('+IntToStr(phaseList[p].Count)+'))'
              );
    Source.Add('pulse(' + dt + '; '+ 'F'+ch+'PhaseCommand, '
                                   + 'F'+ch+'PhaseList('+ IntToStr(k) +'), '
                 + 'F'+ch+'PhaseMode(2), '
                 + 'F'+ch+'PhaseListLength('+IntToStr(phaseList[p].Count)+'))'
              );
    Source.Add('pulse(' + dt + '; '+ 'F'+ch+'PhasePointer_Init, '
                                   + 'F'+ch+'PhaseList('+ IntToStr(k) +'), '
                 + 'F'+ch+'PhaseMode(1), '
                 + 'F'+ch+'Phase('+FloatToStr(phaseList[p].Phase[0])+'))'
              );
    Source.Add('pulse(' + dt + '; '+ 'F'+ch+'PhaseCommand, '
                                   + 'F'+ch+'PhaseList('+ IntToStr(k) +'), '
                 + 'F'+ch+'PhaseMode(1), '
                 + 'F'+ch+'Phase('+FloatToStr(phaseList[p].Phase[0])+'))'
              );
    for q:=1 to phaseList[p].Count-1 do begin
      Source.Add('pulse(' + dt + '; '+ 'F'+ch+'TogglePhase, '
                                     + 'F'+ch+'PhaseList('+ IntToStr(k) +'), '
                   + 'F'+ch+'PhaseMode(1), '
                   + 'F'+ch+'Phase('+FloatToStr(phaseList[p].Phase[q])+'))'
                );
      Source.Add('pulse(' + dt + '; '+ 'F'+ch+'PhaseCommand, '
                                     + 'F'+ch+'PhaseList('+ IntToStr(k) +'), '
                   + 'F'+ch+'PhaseMode(1), '
                   + 'F'+ch+'Phase('+FloatToStr(phaseList[p].Phase[q])+'))'
                );
    end; // q


  end; // k

end;
{-------------------------------------------------}
function TPhaseListArray.Add(s:string):integer;
// s̗:  HP0=(ch1;x,-Y,+Y,23.5,-x)
var
  name:string;
  n,channel:word;
  t:string;
  k,q:word;
  found, finished:boolean;
  ph:single;
  P:integer;
begin

  // O𒊏onameɊi[
  k:=0; found:=false;
  repeat
    inc(k); t:=s[k];
    if s[k]='=' then found:=true;
  until found or (k=Length(s));

  if not found then begin {error}
    msg:='"=" expected.';
    Add:=-1;
    exit;
  end;

  name:=''; q:=0;
  repeat
    inc(q);
    if s[q]<>' ' then name:=name+UpperCase(s[q]); //Xy[XȂ
  until q=k-1;

  if index(name)<>-1 then begin //OdĂI
    msg:='Duplicate identifier (' + name + ').';
    Add:=-1;
    exit;
  end;

  if (Ord(name[1]) in [Ord('0')..Ord('9')]) then begin //Ŏn܂ĂH
    msg:='Phase-list identifier must not start with a number.';
    Add:=-1;
    exit;
  end;


  // `lNo𒊏ochannelɊi[
  repeat
    inc(k);
  until (s[k]<>' ') or (k=Length(s));
  if (s[k]<>'(') or (k=Length(s)) then begin
    msg:='"(" expected.';
    Add:=-1;
    exit;
  end;

  t:=''; found:=false;
  repeat
    inc(k);
    if s[k]<>' ' then t:=t+UpperCase(s[k]);
    if t='CH' then found:=true;
  until found or (k=Length(s));

  if not found then begin {error}
    msg:='"CH" expected.';
    Add:=-1;
    exit;
  end;

  t:=''; found:=false;
  repeat
    inc(k);
    if s[k]<>' ' then begin
      if s[k]=';' then found:=true
      else if not(Ord(s[k]) in [Ord('0')..Ord('9')]) then begin
        msg:='Channel No. expected.';
        Add:=-1;
        exit;
      end else begin
        t:=t+s[k];
      end;
    end;
  until found or (k=Length(s));
  if not found then begin {error}
    msg:='";" expected after channel identifier.';
    Add:=-1;
    exit;
  end;

  channel:=StrToInt(t);
  if channel>FNOfChannels then begin
    msg:='The specified channel no. '
         + t + ' exceeds the number of available channels.';
    Add:=-1;
    exit;
  end;

  n:=NOfLists(channel);
  if n=7 then begin // Y`lɗpł
                                    // phaseListg؂Ă
    msg:='Channel '
         + IntToStr(channel)
         + ' has already had the maximum definabe phase lists.';
    Add:=-1;
    exit;
  end;


  // ̂ƂG[ȂAǉI
  Inc(FCount);
  SetLength(phaseList,FCount);
  phaseList[FCount-1]:=TPhaseList.Create(name,channel,n+1);
       // Y`lŁANOfLists(channel)+1ԖڂphaseListƂȂI(1..7)
  finished:=false;
  repeat

    t:=''; found:=false;
    repeat
      inc(k);
      if s[k]<>' ' then begin
        if (s[k]=',') or (s[k]=')') then found:=true
        else t:=t+UpperCase(s[k]);
      end;
    until found or (k=Length(s));

    if t='' then begin  // ۂ
      msg:='Phase value expected before ' + s[k] + '.';
      Add:=-1;
      phaseList[FCount-1].Free;  // ŒǉphaseListB
      Dec(FCount);
      SetLength(phaseList,FCount);
      exit;
    end;

    if (t='+X') or (t='X') then ph:=0.0
    else if (t='+Y') or (t='Y') then ph:=90.0
    else if (t='-X') then ph:=180.0
    else if (t='-Y') then ph:=270.0
    else begin // l

      Val(t,ph,P);
      if P<>0 then begin
        msg:='Invalid phase value.';
        phaseList[FCount-1].Free;
        Dec(FCount);
        SetLength(phaseList,FCount);
        Add:=-1;
        exit;
      end;

    end;

    if phaseList[FCount-1].Add(ph)<>0 then begin
      msg:='Phase-list length exceeds the maximum value of 32.';
      phaseList[FCount-1].Free;
      Dec(FCount);
      SetLength(phaseList,FCount);
      Add:=-1;
      exit;
    end;


    if s[k]=')' then finished:=true;
  until finished;

  while k<Length(s) do begin
    inc(k);
    if s[k]<>' ' then begin // Xy[XȊO͎cĂȂ͂B΃G[B
      msg:='Invalid expression: ' + s[k] + '.';
        phaseList[FCount-1].Free;
        Dec(FCount);
        SetLength(phaseList,FCount);
        Add:=-1;
        exit;
    end;
  end; // while;


  Add:=0; // G[Ȃ̎A[ԂB
end;
{-------------------------------------------------}
function TPhaseListArray.index(s:string):integer;
var k,m:integer;
begin
  m:=-1;
  for k:=0 to Count-1 do begin
    if phaseList[k].Name=s then begin m:=k; break; end;
  end;
  result:=m;
end;
{-------------------------------------------------}
function TPhaseListArray.index(ch,order:word):integer;
var q,m:integer;
begin
  m:=-1;
  for q:=0 to Count-1 do begin
    if (phaseList[q].Channel=ch) and (phaseList[q].Order=order) then begin
      m:=q; break;
    end;
  end;
  result:=m;
end;
{-------------------------------------------------}
function TPhaseListArray.NOfLists(channel:word):word;
var k,c:integer;
begin
  c:=0;
  for k:=0 to Count-1 do begin
    if phaseList[k].Channel=channel then inc(c);
  end;
  result:=c;
end;
{-------------------------------------------------}
constructor TShapedPulse.Create(name:string;channel:word;order:word);
begin
  FName:=name;
  FChannel:=channel;
  FOrder:=order;
  FLength:=0;
  SetLength(amp,FLength);
end;
{-------------------------------------------------}
constructor TShapedPulse.Create(name:string; channel:word; order:word; koShape:TkindOfShape);
begin
  Create(name,channel,order);
  FKindOfShape:=koShape;
end;
{-------------------------------------------------}
procedure TShapedPulse.Add(ra:single; rp:single);
begin
  Inc(FLength);
  SetLength(amp,FLength);
  amp[FLength-1]:=ra;
  SetLength(phase,FLength);
  phase[FLength-1]:=rp;

end;
{-------------------------------------------------}
function TShapedPulse.Delete(k:word):ShortInt;
var q:integer; // ͕łȂĂ͂ȂȂI
               // ȂƉfor q:=0 to -1 do
               // ƂŎsĂقȂꍇsĂ܂I
begin
  if k>FLength-1 then Delete:=-1  // ͈̓G[
  else begin
    for q:=k to FLength-2 do begin amp[q]:=amp[q+1]; phase[q]:=phase[q+1]; end;
    Dec(FLength);
    SetLength(amp,FLength);
    SetLength(phase,FLength);
    Delete:=0;
  end;
end;
{-------------------------------------------------}
procedure TShapedPulse.Delete;
begin
  While FLength>0 do Delete(FLength-1);
end;
{-------------------------------------------------}
destructor TShapedPulse.Free;
begin
  Delete;
end;
{-------------------------------------------------}
function TShapedPulse.nOfLines:integer;
begin
  if (count mod 5) = 0 then result:=count div 5
  else result:=(Count div 5) +1;
end;
{-------------------------------------------------}
function TShapedPulse.genBin(k:word):string;
var c,j:integer;
    i:integer;
    t:string;
    ph:real;
    p: array[0..4] of integer;
begin

  j:=k * 5;

  c:=0;
  p[0]:=0; p[1]:=0; p[2]:=0; p[3]:=0; p[4]:=0;


  repeat
    i := Round(amp[j]); if i<-511 then i:=-511; if i>511 then i:=511;
    p[c] := i and $3FF;

    ph:=phase[j]; while ph<0 do ph:=ph+360.0; while ph>360 do ph:=ph-360.0;
    i:=Round(1023*ph/360);
    i:=i and $3FF;
    i:=i shl 10;
    p[c]:= p[c] or i;

    inc(j); inc(c);

  until (j=Count) or (c=5);

  t:='0'+Hex(p[4],5)+Hex(p[3],5)+Hex(p[2],5)+Hex(p[1],5)+Hex(p[0],5);

  Insert(' ',t,11);
  Insert(' ',t,20);
  result:=t;


end;
{-------------------------------------------------}
{-------------------------------------------------}
constructor TPhaseList.Create(name:string; channel:word; order:word);
begin
  FName:=name;
  FChannel:=channel;
  FOrder:=order;
  FMaxLength:=128;
  FLength:=0;
  SetLength(Phase,FLength);
end;
{-------------------------------------------------}
destructor TPhaseList.Free;
begin
  Delete;
end;
{-------------------------------------------------}
function TPhaseList.Add(r:single):ShortInt;
begin
  if FLength=FMaxLength then Add:=-1 //ȏǉłȂI
  else begin
     Inc(FLength);
     SetLength(Phase,FLength);
     Phase[FLength-1]:=r;
    Add:=0;
  end;
end;
{-------------------------------------------------}
function TPhaseList.Delete(k:word):ShortInt;
var q:integer; // ͕łȂĂ͂ȂȂI
               // ȂƉfor q:=0 to -1 do
               // ƂŎsĂقȂꍇsĂ܂I
begin
  if k>FLength-1 then Delete:=-1  // ͈̓G[
  else begin
    for q:=k to (FLength-2) do Phase[q]:=Phase[q+1];
    Dec(FLength);
    SetLength(Phase,FLength);
    Delete:=0;
  end;
end;
{-------------------------------------------------}
procedure TPhaseList.Delete;
begin
  while FLength>0 do delete(FLength-1);
end;
{-------------------------------------------------}
constructor auxParameters.Create;
begin
  {Defaultl̐ݒ}
  PD:=1.0E6; {1s=10^6us}
  definePD:=false;
  NA:=1;
  defineNA:=false;
  FAL:=1024;
  defineAL:=false;
  FDW:=10.0;   {10 us}
  defineDW:=false;
  FACQTIME:=10240.0;  {10.24 ms}
end;
{-------------------------------------------------}
procedure auxParameters.setAL(a:cardinal);
begin
  if a<2 then FAL:=2
  else if a>32768 then FAL:=32768
  else FAL:=a;
  FACQTIME:=FAL*FDW;
end;
{-------------------------------------------------}
procedure auxParameters.setDW(d:double);
begin
  if d<0.025 then FDW:=0.025
// else if d>1000 then FDW:=1000
  else FDW:=d;
  FACQTIME:=FAL*FDW;
end;
{-------------------------------------------------}
function Hex(n:cardinal; d:integer):string;
const h:array[0..15] of char='0123456789ABCDEF';
var i:integer;
    s:string;
begin
  s:='';
  for i:=1 to d do begin
    s:=h[n and $f] + s;
    if not (i=d) then n:=n shr 4;
  end; // i
  hex:=s;
end;
{-------------------------------------------------}
procedure TnmrPPG.error(N:integer);
var t,comment:string;
begin
  if not imported then
    t:='[ERROR ('+IntToStr(N)+ ') AT LINE '+IntToStr(lnum+lineOffset-1)+'] : '
  else
    t:='[ERROR ('+IntToStr(N)+ ') AT LINE '+IntToStr(lnum)
       + ' OF ' + importFileNames.Strings[currentImportedFileIndex]
       + '] : ';

  //  + ' COLUMN='+IntToStr(cnum-1) +'] : ';
  CASE N OF
   {0-9 : @֌W}
    0: comment:='Cannot find '+err_str+'.';
    1: comment:='Must be integer expression!';
    2: comment:='Must be number expression!';
    3: comment:= err_str; //'Invalid number!';
    4: comment:='Duplicated ID!';
    5: comment:='Out of range!';
    6: comment:='Invalid character: ' + ch;
    7: comment:='Compiled ppg has ' + err_str +
                ' lines, which exceeds the maximum available (2048) lines.';
    8: comment:='"," cannot be followed by ")".';
   {10-19 : ^Cv~X֌W}
   10: comment:='";" or ")" expected.';

   {20-25: loop֌W}
   21: comment:='Available loop units are 1 or 2!';
   22: comment:='Loop counter ' + err_str + ' is not closed.';
   23: comment:='You are trying either to close loop '
                 + err_str
                 + ' that does not exist,'
                 + ' or to cross loop-endlp and async-sync cycles.';
   24: comment:='Loop-endlp and async-sync cycles cannot be crossed with each other.';

   {26-29: Ԃ̊֌W}
   26: comment:='Time constant not defined!';
   27: comment:='Time duration must not be shorter than 4 clock counts.';
   28: comment:='Time too long!';

   {30-: ʑ֌W}
   30: comment:='Phase variable must be one of X,-X,Y,-Y, or Numbers!';
   31: comment:='Phase variable definition error!';
 //  32: comment:='Phase resolution is 0.5 degree!';
 //  33: comment:='The maximum phases is '+IntToStr(max_phase);
   35: comment:='Phase-list definition error: ' + err_str;
   36: comment:='A mandatory phaselist gate "'+err_str+'" is not defined in the'
                 + ' current gate definition file "' +gateDefFileName+ '".';
   37: comment:='Multiple use of ' + err_str + 'within the same line.';


   38: comment:='Error in acquisition-phase setting: '+ err_str;
   39: comment:='"=" expected after "ACQPHASE"';
   40: comment:='";" expected at the end of ACQPHASE setting.';

   41: comment:='PPG empty ("start" missing?).';


   44: comment:='A mandatory gate "'+err_str+'" is not defined in the'
                 + ' current gate definition file "' +gateDefFileName+ '".';
   50: comment:='Invalid character after "start".';

   66: comment:='Invalid command: ' + err_str;

   70: comment:='Amplitude variable ' + err_str + ' is out of range. '
                 + 'It should be between 0.0 and 100.0.';

   71: comment:='Integer variable ' + err_str + ' is out of range. ';

   72: comment:='Loop variable ' + err_str + ' is out of range. '
                 + 'It should be between 1 and 65535.';

   74: comment:='time variable ' + err_str + ' is out of range. '
                 + 'It should be between 0.04 us and 10000 s.';

   80: comment:='Shaped pulse definition error: ' + err_str;

   100: comment:='Multiple declaration of gate name '+err_str + '.';
   101: comment:='Failed to read the gate-definition file '+err_str + '.';
   102: comment:='Invalid gate identifier' + err_str + '.';
   103: comment:='Gate identifier "' + err_str + '" is not defined.';
   104: comment:='The gate ' + err_str + ' is defined as a logic gate' +
                 ' and should not be followed by a parenthesis.';

   105: comment:='The gate ' + err_str + ' is defined as an amplitude gate' +
                 ' and should have a parameter.';
   106: comment:='Number can no longer be used for the gates. ' +
                 'Please define a gate variable.';
   107: comment:='Multiple usage of gate variable "'+ err_str + '".';
   108: comment:='The gate "' + err_str + '" is defined as a logic_vector gate'
                 + ' and should have a parameter.';
   109: comment:='The gate "' + err_str + '" is defined as an AD9858 gate'
                 + ' and should have a parameter.';
//   110: comment:='Invalid usage of AD9858 gate parameter "'
//                 + err_str
//                 + '". Usage: '
//                 + err_str
//                 + '(PS[0..3];freq:double)';
//   111: comment:='Only 0, 1, 2, or 3 is allowed for the first parameter '
//                 + '(Profile Select: PS) '
//                 + 'of AD9858 gate "' + err_str + '".';
   112: comment:='The second parameter of AD9858 gate "'
                 + err_str + '" must be a number expression '
                 + '(frequency in MHz).';
   113: comment:='Frequency is out of range.';

   114: comment:='The gate ' + err_str + ' is defined as a phaselist gate' +
                 ' and should not be followed by a parenthesis.';

   115: comment:='Frequency cannot be set in the wait_H/wait_L commands.';

   116: comment:='The gate "' + err_str + '" is defined as an integer gate'
                 + ' and should have a parameter.';

   120: comment:='Unknown parameter %' + err_str + '. The candidates are' +
                 ' %NA, %AL, %PD, and %DW.';
   121: comment:='PD (Pulse Delay) is too short! it should at least be 1 ms.';
   122: comment:='PD (Pulse Delay) is too long! it should not exceed 10000 s.';
   123: comment:='AL (Acquisition Length) should at least be 2.';
   124: comment:='AL (Acquisition Length) should not exceed 32768.';
   125: comment:='AL (Acquisition Length) should be an integer.';
   126: comment:='NA (Number of Acquisitions) should be an integer.';
   127: comment:='NA (Number of Acquisitions) should at least be 1, and at most be 4294967295';
   128: comment:='DW (DWell time) should at least be 25 ns.';
   129: comment:='DW (DWell time) should not exceed 1 ms.';
   130: comment:='You are trying to define a gate parameter named "'
                  + err_str
                  + '", but it is a reserved word.';
   131: comment:='You are trying to define a time variable named "'
                  + err_str
                  + '", but it is a reserved word.';
   132: comment:='You are trying to define a variable "'
                  + err_str
                  + '", but it is a reserved word.';
   133: comment:='';

   140: comment:= '"=" expected after "uses".';
   141: comment:= '";" is missing in the "uses" command.';
   142: comment:= 'Multiple usage of a "uses" command.';

   143: comment:= '"=" expected after "CLK".';
   144: comment:= '";" is missing in the "CLK" command.';
   145: comment:= 'Invalid number in the "CLK" command.';

   150: comment:='Unit of time (n, u, m, or s) is missing!';
   197: comment:='"Init" command missing';
   198: comment:='Cannot find the start of the main part of the pulse program '+
                 '(PPG empty or "Start" command missing?).';

   199: comment:='Cannot find the end of the pulse program '+
                 '(PPG empty or "Relax" command missing?).';
   200: comment:='Syntax error at line ' + err_str;
   201: comment:='Parenthesis "(" missing in "async" declaration at line '
                  + err_str + '.';
   202: comment:='Parenthesis ")" missing in "async" declaration at line '
                  + err_str + '.';
   203: comment:='Invalid character ' + err_str + '.';
   204: comment:='Invalid usage of ASYNC declaration at line '
                  + err_str + '. Usage: async(ch;name).';
   205: comment:='At line ' + err_str +
                 ', only 2 or 3 is allowed for an asynchronous channel.';
   206: comment:='Cannot find the end ("continue" command) of the asynchronous sequence '
                  + err_str + '.';
   207: comment:='Multiple definition of an asynchronous sequence identifier '
                  + err_str + '.';
   208: comment:='The first character of an asynchronous sequence identifier '
                  + 'should be an alphabetical character.';
   209: comment:='Invalid character for an asynchronous sequence identifier '
                  + 'at line ' + err_str + '.';
   210: comment:='Asynchronous sequence channel must be an integer number.';
   211: comment:= err_str;

   220: comment:='The command "init" is not allowed in an asynchronous sequence.';
   221: comment:='The command "relax" is not allowed in an asynchronous sequence';
   222: comment:='The command "async" is not allowed in an asynchronous sequence';
   223: comment:='The command "sync" is not allowed in an asynchronous sequence';
   224: comment:='The asynchronous sequence "'
                  + err_str
                  + '" does not exist!';
   225: comment:='Asynchronous operation has not been ended up '
                  + 'with re-synchronization. Plaease use a "sync" command '
                  + 'before the "relax" command.';

   226: comment:='Cannot find the other parenthesis ")" in the "async" command';
   227: comment:='Cannot find the other parenthesis ")" in the "sync" command';

   228: comment:='A parenthesis "(" is missing in the "async" command.';
   229: comment:='A parenthesis "(" is missing in the "sync" command.';
   230: comment:='Number expected for the parameter of "sync".';
   231: comment:='Invalid parameter for the "sync" command. '
                 + 'Only 2 or 3 is allowed.';

   232: comment:='Cannot use an asynchronous sequence ' + err_str;
   233: comment:='You are trying to finish an asynchronous sequence on channel '
                 + err_str
                 + ' that has not been started.';
   234: comment:='Multiple usage of "' + err_str + '".';

   250: comment:='The 1st argument of the wait_H/wait_L command should be ' +
                 'TRIG_E, TRIG_R, TRIG_0, TRIG_1, TRIG_2, or TRIG_3';
   251: comment:='In the wait_H/wait_L command, ";" or ")" expected after TRIG_?';
   252: comment:='Shaped pulse cannot be used in the wait_H/wait_L commands.';

   300: comment:= 'The parameter for amplitude gate ' + err_str
                + ' is out of range (0 to 100).';
   301: comment:= '"' + err_str + '"' + ' is not an amplitude variable.';
   302: comment:= '"' + err_str + '"' + ' is not a phase variable.';
   303: comment:= '"' + err_str + '"' + ' is not an integer variable.';
   304: comment:= '"' + err_str + '"' + ' is not a frequency variable.';

   400: comment:= 'Frequency setting on AD9858 requires >50/CLK.';


   500: comment:= 'File "'+err_str+'" does not exist.';

   600: comment:= '"(" expected after "burst".';
   601: comment:= 'Cannot include AD9858 control in the "burst" command.';

   800: comment:= 'Cannot import from the imported file.';
   801: comment:= 'Import: "(" expected.';
   802: comment:= 'Import: ")" expected.';
   803: comment:= 'Cannot open the input file "'+err_str+'"';

   999: comment:='This is a test message.   : ' + err_str;
 end;
  Messages.Add(t+comment);
  erra:=true;
  Compiled:=False;
//  EndOfFile:=True;
END;

{-------------------------------------------------}
function TnmrPPG.loadFromFile(fn:string):integer;
begin
  if not fileExists(fn) then begin
    err_str:=fn;
    error(500);
    result:=-1;
  end else begin
    ppgLines.LoadFromFile(fn);
    result:=0;
  end;
end;

{-------------------------------------------------}
constructor TnmrPPG.Create(NOfChannels: Word; fn:string);
begin
  create(nOfChannels);
  ppgFilename:=fn;
  ppgFilePath:=extractFilePath(ppgFileName);
//  gateDefFileDir:=ExtractFilePath(ppgFileName);
end;

constructor TnmrPPG.Create(NOfChannels: word);
var k: word;
begin
  ppgLines:=TStringList.Create;
  {TStringspʃIuWFNgTStringListCreate}
  workingLines:=TStringList.Create;
  importedLines:=TStringList.Create;
  preamble:=TStringList.Create;
  mainPPG:=TStringList.Create;
  asyncPPGList:=TasyncPPGList.Create;

  FNOfChannels:=NOfChannels;
  SetLength(compiledPPG,NOfChannels);
  SetLength(NOfLinesPerCommand,NOfChannels);
  SetLength(mnemonic,NOfChannels);
  for k:=0 to NOfChannels-1 do begin
    compiledPPG[k]:=TStringList.Create;
    NOfLinesPerCommand[k]:=TStringList.Create;
    mnemonic[k]:=TStringList.Create;
  end; // k;

  Messages:=TStringList.Create;
  info:=TStringList.Create;
  Warnings:=TStringList.Create;

  phListArray:=TPhaseListArray.Create(FNOfChannels);
  acqPhaseList:=TAcqPhaseList.Create;

  shapedPulseArray:=TShapedPulseArray.Create(FNOfChannels);

  Variables:= TVariables.Create(FNOfChannels);


  gates:=TGates.create;
  auxParams:=auxParameters.Create;
  reservedWords:=TStringList.Create;
  reservedWords.Add('ACQTIME');
  reservedWords.Add('NA');
  reservedWords.Add('AL');
  reservedWords.Add('PD');
  reservedWords.Add('DW');
 // reservedWords.Add('F1');
 // reservedWords.Add('F2');
 // reservedWords.Add('F3');
  reservedWords.Add('TIME');
  reservedWords.Add('LOOP');
  reservedWords.Add('AUX');
  reservedWords.Add('ACQ');

  // used for exp condition file
  reservedWords.Add('FILE');
  reservedWords.Add('GENERAL');
  reservedWords.Add('ACQUISITION');
  reservedWords.Add('TRANSFER');
  reservedWords.Add('DIGITALFILTER');


  gateDefFileName:=''; // gate definition file name
  gateDefFileDir:='';

  compiled:=false;

  acqPhaseDefined:=false;

  updateCommand:=TStringList.Create;

  importFileNames:=TStringList.Create;

  CLK:=160.0;  // default CLK: 160 MHz

end;
{-------------------------------------------------}
destructor TnmrPPG.Free;
var k: word;
begin
  ppgLines.Free;
  workingLines.Free;
  importedLines.Free;
  preamble.Free;
  mainPPG.Free;
  asyncPPGList.Free;

  for k:=0 to NOfChannels-1 do begin
    compiledPPG[k].Free;
    NOfLinesPerCommand[k].Free;
    mnemonic[k].Free;
  end;

  phListArray.Free;
  acqPhaseList.Free;

  shapedPulseArray.Free;

  Messages.Free;
  info.Free;
  ReservedWords.Free;
  Warnings.Free;
  Variables.Free;
  Gates.free;
  auxParams.Free;

  updateCommand.Free;
  importFileNames.Free;
end;
{-------------------------------------------------}
procedure TnmrPPG.Nextln; {* Nextln *}
begin

  if imported=false then begin
   {̈sobt@[(lb)ɓāCŏ̕Ă}
    inc(lnum); EndOfFile:=(lnum>workingLines.Count);
    if not EndOfFile then begin
      lb:=workingLines[lnum-1];
      CNUM:=1;
      if lb='' then ch:=' ' else ch:=upcase(LB[1]);
    end;
  end else begin    //imported
    inc(lnum);
    if lnum<=importedLines.Count then begin
      lb:=importedLines[lnum-1];
      CNUM:=1;
      if lb='' then ch:=' ' else ch:=upcase(LB[1]);

    end else begin   // end of imported lines!

      imported:=false;
      lnum:=lnum_bak;  //  ƂƂ̍sɂǂ

      inc(lnum); EndOfFile:=(lnum>workingLines.Count);
      if not EndOfFile then begin
        lb:=workingLines[lnum-1];
        CNUM:=1;
        if lb='' then ch:=' ' else ch:=upcase(LB[1]);
      end;
    end;

  end;

 end;
{-------------------------------------------------}
procedure TnmrPPG.getch; {* GETCH *}
begin  {ch:charƂĂ}
  inc(CNUM);
  if LENGTH(LB) < CNUM then Nextln
                       else CH := UPcase(LB[CNUM]);
  { if ch='{' then nextln;} {c̓RgI}
                   {ꂾƁA̍sĂуRgꍇɑΉłȂ̂
                   ȉ̂悤ɂB2006.2.21 Takeda}
  if ch='{' then begin
    repeat
      nextln;
    until ch<>'{';
  end;
end;
{-------------------------------------------------}
procedure TnmrPPG.Yomitobashi;
begin {󔒂ǂݔ΂}
 while ((ch=' ') or (ch='@') or (ch=#9)) and (not EndOfFile) do getch;
end;
{-------------------------------------------------}
function TnmrPPG.checkCH(c:char):Boolean;
begin  {chcǂ`FbN}
 Yomitobashi;
 Result:=c=ch;
 if (not Result) then
    begin err_str:=c; error(0); end;
end;
{-------------------------------------------------}
function TnmrPPG.Search(s:string):Boolean;
{sɊ܂܂镶܂
 t@C̍Ō܂łTŁ[}
var i:integer; Find:boolean;
begin
 Find:=False;
 Yomitobashi;
 while (not find) and (not EndOfFile) do begin
 i:=0;
  repeat
   inc(i);
   find:=s[i]=ch;
  until find or (i=length(s));
  if (not find) and (not EndOfFile) then getch;
 end;

 Result:=Find;

 if (not Find) and (not EndOfFile) then
    begin err_str:=s; error(0); end;
end;

{-------------------------------------------------}
procedure TnmrPPG.getFromLoops(var N:double);
var
  s:string;
  i:integer;
begin
  S:='';
  while ch<>')' do begin
    if (ch<>' ') and (ch<>#9) then s:=s+ch;
    getch;
  end;

  with Variables do begin
    i:=index(s);

    if defined(s) then begin
      if not asyncFlag then begin
        inc(Variable[i].FNOfWhereIs[currentCH-1]);
        SetLength(Variable[i].FWhereIs[currentCH-1],Variable[i].FNOfWhereIs[currentCH-1]);
        Variable[i].FWhereIs[currentCH-1][Variable[i].count(currentCH-1)-1]
          :=compiledPPG[currentCH-1].Count;

        SetLength(Variable[i].FGateIndex[currentCH-1],Variable[i].FNOfWhereIs[currentCH-1]);
        Variable[i].FGateIndex[currentCH-1][Variable[i].Count(currentCH-1)-1]
                       := -1;
         // [vϐ̓Q[gƖ֌WȂ̂ŁAƂ肠_~[Ƃ-1ƂĂB

      end; {if not asyncFlag}

      N:=Variable[i].Value;

    end else begin
      err_str:=s;
      error(0);
    end;
  end; // with
end;

{-------------------------------------------------}
procedure TnmrPPG.getFromTimes(var N:double);
var
  s:string;
  i:integer;
begin
  s:='';
  while not (Ord(ch) in [Ord(';'),Ord(')')]) do begin
    if (ch<>' ') and (ch<>#9) then s:=s+ch;
    getch;
  end; // while


  with Variables do begin
    i:=index(s);
    if defined(s) then begin
      if not asyncFlag then begin
        inc(Variable[i].FNOfWhereIs[currentCH-1]);
        SetLength(Variable[i].FWhereIs[currentCH-1],Variable[i].FNOfWhereIs[currentCH-1]);
        Variable[i].FWhereIs[currentCH-1][Variable[i].count(currentCH-1)-1]
          :=compiledPPG[currentCH-1].Count;

        SetLength(Variable[i].FGateIndex[currentCH-1],Variable[i].FNOfWhereIs[currentCH-1]);
        Variable[i].FGateIndex[currentCH-1][Variable[i].Count(currentCH-1)-1]
                       := -1;
         // ԕϐ̓Q[gƖ֌WȂ̂ŁAƂ肠_~[Ƃ-1ƂĂB



      end; {if not asyncFlag}

      N:=Variable[i].Value;

    end else begin
      err_str:=s;
      error(0);
    end;
  end; // with

end;
{-------------------------------------------------}
procedure TnmrPPG.Number(var X:double);
var s:string; p:integer;
    N:double;

  procedure Get_String;
  begin
    if not(Ord(ch) in [Ord('0')..Ord('9'),Ord('-'),Ord('.'),Ord('e'),Ord('E')]) then begin error(2);exit end;
    while Ord(ch) in [Ord('0')..Ord('9'),Ord('-'),Ord('.'),Ord('e'),Ord('E')] do begin
      s:=s+ch;
      getch;
    end; // while
  end;

begin // Number main
  Yomitobashi;

  if not(Ord(ch) in [Ord('0')..Ord('9'),Ord('-')]) then begin
    if Yes_Loop then getFromLoops(N) else getFromTimes(N);
    if erra then exit;
  end else begin{̏ꍇ}
//    if ch='-' then begin s:='-'; getch; end else s:='';
    Get_String;
//    if ch='.' then begin
//      s:=s+ch; getch;
//      Get_String;
//    end;
    val(S,N,P);
    if P<>0 then begin
      err_str:='Invalid expression: ' + S ;
      error(3);
      exit;
    end;

  end;{else}

  X:=N;
  {ch;)̂͂}
end;
{-------------------------------------------------}
procedure TnmrPPG.Number_Time(var X,Num:double;
                             var unitOfTime:TUnitOfTime);
var s:string; p:integer;
    unitQ:boolean;
    N:double;

  procedure Get_String;
  begin
    if not(Ord(ch) in [Ord('0')..Ord('9')]) then begin error(2);exit end;
    while Ord(ch) in [Ord('0')..Ord('9')] do begin
      s:=s+ch;
      getch;
    end; // while
  end;

begin // Number main
  N:=0;
  Yomitobashi;

  if not(Ord(ch) in [Ord('0')..Ord('9')]) then begin {ԕϐ̏ꍇ}
    getFromTimes(X);
    if erra then exit;
  end else begin                      {l̏ꍇ}
    s:='';
    Get_String;
    if ch='.' then begin
      s:=s+ch; getch;
      Get_String;
    end;
    val(S,N,P);
    if P<>0 then begin
      err_str:='Invalid expression: ' + S;
      error(3);
      exit;
    end;

    unitQ:=false;
    case ch of
     'N' : begin X:=N/1000.0; unitOfTime:=nanoSeconds; unitQ:=true; getch; end;
     'U' : begin X:=N;        unitOfTime:=microSeconds; unitQ:=true; getch; end;
     'M' : begin X:=N*1000.0; unitOfTime:=milliSeconds; unitQ:=true; getch; end;
     'S' : begin X:=N*1000000.0; unitOfTime:=Seconds; unitQ:=true; getch; end;
    end;{Case}

    if unitQ=false then begin {Pʂ͕Kw肷悤}
      error(150);
      exit
    end;
  end;{else}

  Num:=N;  {l}


  {ch;)̂͂}
end;
{-------------------------------------------------}
procedure TnmrPPG.Jikann_Hennkann(X:double; var Uwidth:word; var Lwidth:cardinal);
             {shortInt8bitAlongInt32bitA킹40bit}
var a,b:double;
begin
  if X<(4/CLK) then begin X:=(4/CLK); error(27);end;
     {4 clocks counts}
  if X>(1.099E12/CLK) {} then begin X:=(1.099E12/CLK); error(28); end;

  a:=X*CLK;
  b:=Trunc(a);
  if a<>b then begin

// if (X*CLK)<>Trunc(X*CLK) then begin
//       ƌ쓮HH
//  ̂悤abɑĔrƑvA݂B
//   2009.9.29 K. Takeda
    Warnings.Add('Time duration (' + FloatToStr(X)
            +' us) was truncated to ' + FloatToStr(Trunc(X*CLK)/CLK)
            +' us');
  end;

  if (CLK*X-4) > 4294967295 {2^32-1} then begin
        {CLK[MHz] * X[us] --> number of count}
        {4 --> excess clocks}
    a:=Trunc((CLK*X-4)/4294967296);   // 2^32Ŋ
    b:=(CLK*X-4) - (a*4294967296);
  end else begin
    a:=0;
    b:=CLK*X-4;
  end;
    Uwidth:=Trunc(a);
    LWidth:=Trunc(b);
end;
{-------------------------------------------------}
procedure TnmrPPG.duration(var Uwidth:word; var Lwidth:cardinal);
var X,N:double;
    ut:TUnitOfTime;
begin
  if checkCH('(') then begin
    getch;
    number_time(X,N,ut); if erra then exit;
    Jikann_Hennkann(X,Uwidth,Lwidth);
    {ch;)̂͂}
    Yomitobashi;
  end; // if
end;
{-------------------------------------------------}
(****************************************************)
(************** ϐ̒`̉ ******************)
(****************************************************)
{-------------------------------------------------}
(***** =  (LvV) ; ƂĂ *****)
procedure TnmrPPG.Get_S1S2(var X:double; var S1,S2:string);
begin
 while ch <>'=' do begin
    if (CH<>' ') and (ch<>#9) then S1 := S1 + CH;
    GETCH;
   end;
 checkCH('='); {=܂œǂ}
 getch;

 Number(X);

 Search('(;');

 if ch='(' then begin
 getch;
 while ch <>')' do begin
    S2 := S2 + CH;
    GETCH;
 end;
 end else {(LvV)}
 begin
  //S2:='No caption';
  S2:='';
 end;

 Search(';');
end;
{-------------------------------------------------}
procedure TnmrPPG.setAuxParams;
var S1,S2:string;
    X,Num:double;
    ut:TUnitOfTime;
begin
  Yomitobashi;
  S1:=''; S2:='';
    while ch <>'=' do begin
      if (CH<>' ') and (ch<>#9) then S1 := S1 + CH;
      GETCH;
    end;
  checkCH('='); {=܂œǂ}

  if S1='PD' then begin
    getch;
    Number_Time(X,Num,ut);
    Search('(;');

    if ch='(' then begin
      getch;
      while ch <>')' do begin
        S2 := S2 + CH;
        GETCH;
      end;
    end else begin {(LvV)}
      S2:='No caption';
    end;

    Search(';'); getch;

 //   if (X<1.0E3) then begin error(121);exit end; {shorter than 1ms}
    if (X>1.0E10) then begin error(122);exit end; {longer than 10000s}
    auxParams.PD:=X;
    auxParams.definePD:=true;

    with Variables do begin
      Add('PD', TimeVariable);
      Variable[count-1].Value:=auxParams.PD;
      Variable[count-1].Numeric:=Num;
      Variable[count-1].unitOfTime:=ut;
      case ut of
        nanoSeconds: Variable[count-1].unitOfTimeString:='n';
        microSeconds: Variable[count-1].unitOfTimeString:='u';
        milliSeconds: Variable[count-1].unitOfTimeString:='m';
        Seconds: Variable[count-1].unitOfTimeString:='s';
      end; // case
      Variable[count-1].Caption:='Pulse Delay';
      Variable[count-1].declarationPosition:=lnum-2; // lum-1 -> lnum-2

    end; // with

  end else if S1='AL' then begin
    getch;
    Number(X);
    Search('(;');

    if ch='(' then begin
      getch;
      while ch <>')' do begin
        S2 := S2 + CH;
        GETCH;
      end;
    end else begin {(LvV)}
      S2:='No caption';
    end;

    Search(';'); getch;

    if Trunc(X)-X<>0 then begin error(125); exit end;  {not an integer!}
    if X<2 then begin error(123); exit end;
    if X>32768 then begin error(124); exit end;
    auxParams.AL:=Trunc(X);
    auxParams.defineAL:=true;

    with Variables do begin
      Add('AL', intVariable);
      Variable[count-1].Value:=auxParams.AL;
      Variable[count-1].Numeric:=auxParams.AL;
      Variable[count-1].Caption:='Acquisition Length';
      Variable[count-1].declarationPosition:=lnum-2; // lum-1 -> lnum-2;
    end; // with

  end else if S1='NA' then begin
    getch;
    Number(X);
    Search('(;');

    if ch='(' then begin
      getch;
      while ch <>')' do begin
        S2 := S2 + CH;
        GETCH;
      end;
    end else begin {(LvV)}
      S2:='No caption';
    end;

    Search(';'); getch;

    if Trunc(X)-X<>0 then begin error(126); exit end;  {not an integer!}

    auxParams.NA:=Trunc(X);
    auxParams.defineNA:=true;
    with Variables do begin
      Add('NA', NAVariable);
      if Trunc(X)<Variable[count-1].Min then begin error(127); exit end;
      if Trunc(X)>Variable[count-1].Max then begin error(127); exit end;
      Variable[count-1].Value:=auxParams.NA;
      Variable[count-1].Numeric:=auxParams.NA;
      Variable[count-1].Caption:='Number of Accum';
      Variable[count-1].declarationPosition:=lnum-2; // lum-1 -> lnum-2;

    end; // with

  end else if S1='DW' then begin
    getch;
    Number_Time(X,Num,ut);
    Search('(;');

    if ch='(' then begin
      getch;
      while ch <>')' do begin
        S2 := S2 + CH;
        GETCH;
      end;
    end else begin {(LvV)}
      S2:='No caption';
    end;

    Search(';'); getch;

    if (X<0.025) then begin error(128);exit end; {shorter than 25 ns}
    if (X>1.0E6) then begin error(129);exit end; {longer than 1ms}
    auxParams.DW:=X;
    auxParams.defineDW:=true;


    //Variables.variable[Variables.index('DW')].value;

    with Variables do begin
      Add('DW', TimeVariable);
      Variable[count-1].Value:=auxParams.DW;
      Variable[count-1].Numeric:=Num;
      Variable[count-1].unitOfTime:=ut;
      case ut of
        nanoSeconds: Variable[count-1].unitOfTimeString:='n';
        microSeconds: Variable[count-1].unitOfTimeString:='u';
        milliSeconds: Variable[count-1].unitOfTimeString:='m';
        Seconds: Variable[count-1].unitOfTimeString:='s';
      end; // case
      Variable[count-1].Caption:='Dwell time';
      Variable[count-1].declarationPosition:=lnum-2; // lum-1 -> lnum-2;

    end; // with

  end else begin
    err_str:=S1;
    error(120);
  end; // if(S1)

end;
{-------------------------------------------------}
procedure TnmrPPG.freqVariables;
var S1,S2:string;
    X: double;
begin
  Yomitobashi;
  S1:=''; S2:='';
  Get_S1S2(X,S1,S2);

  if reservedWords.IndexOf(S1)<>-1 then begin {\gĂI}
    err_str:=S1;
    error(132);
    exit;
  end;

  with Variables do begin
    if not defined(S1) then begin
      add(S1,freqVariable);
      Variable[count-1].Value:=X;
      Variable[count-1].Numeric:=X;
      Variable[count-1].Caption:=S2;
      Variable[count-1].declarationPosition:=lnum-1;
      getch;
    end else begin
      error(4); exit;
    end; // if
  end; // with

end;
{-------------------------------------------------}
procedure TnmrPPG.phaseVariables;
var S1,S2:string;
    X: double;
begin
  Yomitobashi;
  S1:=''; S2:='';
  Get_S1S2(X,S1,S2);

  if reservedWords.IndexOf(S1)<>-1 then begin {\gĂI}
    err_str:=S1;
    error(132);
    exit;
  end;

  with Variables do begin
    if not defined(S1) then begin
      add(S1,phaseVariable);
      Variable[count-1].Value:=X;
      Variable[count-1].Numeric:=X;
      Variable[count-1].Caption:=S2;
      Variable[count-1].declarationPosition:=lnum-1;
      getch;
    end else begin
      error(4); exit;
    end; // if
  end; // with

end;
{-------------------------------------------------}
procedure TnmrPPG.intVariables;
var S1,S2:string;
    X: double;

begin
  Yomitobashi;
  S1:=''; S2:='';
  Get_S1S2(X,S1,S2);

  if reservedWords.IndexOf(S1)<>-1 then begin {\gĂI}
    err_str:=S1;
    error(132);
    exit;
  end;

  if trunc(X)-X<>0 then begin error(1); exit; end; {}

  with Variables do begin
    if not defined(S1) then begin
      add(S1,intVariable);
      if (X<Variable[count-1].Min) or (X>Variable[count-1].Max) then begin
        err_str:=S1; error(71); exit;
      end;
      Variable[count-1].Value:=X;
      Variable[count-1].Numeric:=X;
      Variable[count-1].Caption:=S2;
      Variable[count-1].declarationPosition:=lnum-1;
      getch;
    end else begin
      error(4); exit;
    end; // if
  end; // with



end;
{-------------------------------------------------}
procedure TnmrPPG.ampVariables;
var S1,S2:string;
    X: double;
begin
  Yomitobashi;
  S1:=''; S2:='';
  Get_S1S2(X,S1,S2);

  if reservedWords.IndexOf(S1)<>-1 then begin {\gĂI}
    err_str:=S1;
    error(132);
    exit;
  end;

  with Variables do begin
    if not defined(S1) then begin
      add(S1,ampVariable);
      if (X<Variable[count-1].Min) or (X>Variable[count-1].Max) then begin
        err_str:=S1; error(70); exit;
      end;
      Variable[count-1].Value:=X;
      Variable[count-1].Numeric:=X;
      Variable[count-1].Caption:=S2;
      Variable[count-1].declarationPosition:=lnum-1;
      getch;
    end else begin
      error(4); exit;
    end; // if
  end; // with

end;
{-------------------------------------------------}
procedure TnmrPPG.loopVariables;
var S1,S2:string;
    X:double;
begin
  Yomitobashi;
  S1:=''; S2:='';
  Get_S1S2(X,S1,S2);
  if trunc(X)-X<>0 then begin error(1); exit; end; {}

  if reservedWords.IndexOf(S1)<>-1 then begin {\gĂI}
    err_str:=S1;
    error(132);
    exit;
  end;

  with Variables do begin
    if not defined(S1) then begin
      add(S1,loopVariable);
      if (X<Variable[count-1].Min) or (X>Variable[count-1].Max) then begin
        err_str:=S1; error(72); exit;
      end;
      Variable[count-1].Value:=X;
      Variable[count-1].Numeric:=X;
      Variable[count-1].Caption:=S2;
      Variable[count-1].declarationPosition:=lnum-1;
      getch;
    end else begin
      error(4); exit;
    end; // if
  end; // with
end;
{-------------------------------------------------}
procedure TnmrPPG.timeVariables;
var S1,S2:string;
    X,Num:double;
    ut:TUnitOfTime;
begin
  Yomitobashi;
  S1:=''; S2:='';

   while ch <>'=' do begin
    if (CH<>' ') and (ch<>#9) then S1 := S1 + CH;
    GETCH;
   end;
   checkCH('='); {=܂œǂ}
   getch;

   Number_Time(X,Num,ut);

   Search('(;');

   if ch='(' then begin
     getch;
     while ch <>')' do begin
       S2 := S2 + CH;
      GETCH;
     end;
   end else begin {(LvV)}
     S2:='No caption';
   end;

   Search(';');

   if reservedWords.IndexOf(S1)<>-1 then begin {\gĂI}
     err_str:=S1;
     error(131);
     exit;
   end;

  with Variables do begin
    if not defined(S1) then begin
      add(S1,TimeVariable);
      if (X<Variable[count-1].Min) or (X>Variable[count-1].Max) then begin
        err_str:=S1; error(74); exit;
      end;
      Variable[count-1].Value:=X;       {}CNb}
      Variable[count-1].Numeric:=Num;   {l}
      Variable[count-1].unitOfTime:=ut; {Ԃ̒P}
      Variable[count-1].Caption:=S2;
      Variable[count-1].declarationPosition:=lnum-1;
      case ut of
        nanoSeconds: Variable[count-1].unitOfTimeString:='n';
        microSeconds: Variable[count-1].unitOfTimeString:='u';
        milliSeconds: Variable[count-1].unitOfTimeString:='m';
        Seconds: Variable[count-1].unitOfTimeString:='s';
      end; // case
      getch;
    end else begin
      error(4); exit;
    end; // if
  end; // with
end;
{-------------------------------------------------}
procedure TnmrPPG.SetAcqPhase;
var S1:string;
begin

  while ch=' ' do getch;    {ch'='ɂȂ͂B}
  if ch<>'=' then begin
    err_str:='';
    error(39);
    exit;
  end else begin
    S1:='';
    repeat
      getch;
      if (ch<>' ') and (ch<>#9) then s1:=s1+ch;
      if EndOfFile then begin {Z~RtYH}
        error(40);
        exit;
      end;
    until ch=';';
    getch; {';'肷}

    if acqPhaseList.ReadFromString(S1)<>0 then begin
      err_str:=acqPhaseList.msg;
      error(38);
      exit;
    end;

    // [ACQ_Phase]Ƃڂ̃Q[g`ĂȂꍇ̓G[
    if gates.index('ACQ_PHASE')=-1 then begin
      err_str:= 'ACQ_PHASE';
      error(44);
      exit;
    end;

    acqPhaseDefined:=true;

  end;
end;
{-------------------------------------------------}
procedure TnmrPPG.SetPhaseList;
var S1:string;
begin
  S1:='';
  repeat
    getch;
    if (ch<>' ') and (ch<>#9) then S1:=S1+ch;
    if EndOfFIle then begin err_str:=')'; error(0); end;
  until ch=')';

  getch;
  while ch=' ' do getch;
  if ch<>';' then begin err_str:=';'; error(0); end;

  if phListArray.Add(S1)<>0 then begin
    err_str:=phListArray.msg; error(35); exit;
  end;

  getch;

  // [F#PhaseList]Ƃڂ̃Q[g`ĂȂꍇ̓G[
  if gates.index('F'+IntToStr(phListArray.phaseList[phListArray.Count-1].Channel)
                    +'PHASELIST')=-1 then begin
    err_str:= 'F'+IntToStr(phListArray.phaseList[phListArray.Count-1].Channel)
                    +'PHASELIST';
    error(36);
  end;

  gates.add(phListArray.phaseList[phListArray.Count-1].name,
            'PHASELIST',phListArray.phaseList[phListArray.Count-1].order);
  gates.gate[gates.Count-1].channel:=
        phListArray.phaseList[phListArray.Count-1].Channel; // set channel #


end;
{-------------------------------------------------}
procedure TnmrPPG.SetShape(kos:TKindOfShape);
var S1:string;
    ks:string;
begin
  S1:='';
  repeat
    getch;
    if (ch<>' ') and (ch<>#9) then S1:=S1+ch;
    if EndOfFIle then begin err_str:=')'; error(0); end;
  until ch=')';

  getch;
  while ch=' ' do getch;
  if ch<>';' then begin err_str:=';'; error(0); end;

  if shapedPulseArray.Add(S1,kos)<>0 then begin
    err_str:=shapedPulseArray.msg; error(80);
    exit;
  end;

  getch;

  // [F#PhaseList]Ƃڂ̃Q[g`ĂȂꍇ̓G[
//  if gates.index('F'+IntToStr(phListArray.phaseList[phListArray.Count-1].Channel)
//                    +'PHASELIST')=-1 then begin
//    err_str:= 'F'+IntToStr(phListArray.phaseList[phListArray.Count-1].Channel)
//                    +'PHASELIST';
//    error(36);
//  end;

  case kos of
    shAmplitude: ks:='AMPSHAPE';
    shPhase: ks:='PHASESHAPE';
    shAP: ks:='APSHAPE';
  end;

  with shapedPulseArray do begin
    gates.add(shape[Count-1].name, ks, shape[Count-1].order);
    gates.gate[gates.Count-1].channel:= shape[Count-1].Channel; // set channel #
  end;  // with

end;


(****************************************************)
(************* vAuR}h *****************)
(****************************************************)
procedure TnmrPPG.auxCom(var ok:boolean);
type sym=(USESSY, CLKSY, TIMESY, LOOPSY, AMPSY, INTSY,
          PHASESY, FREQSY, AUXSY, STARTSY, PHASELISTSY,
          ACQPHASESY, AMPSHAPESY, PHASESHAPESY, APSHAPESY, GUARD);
const key:array[USESSY..APSHAPESY] of string
          =('USES', 'CLK', 'TIME','LOOP','AMP', 'INT',
            'PHASE', 'FREQ', 'AUX', 'START', 'PHASELIST', 'ACQPHASE',
            'AMPSHAPE', 'PHASESHAPE', 'APSHAPE');
var s:string;
    i:sym; testln:integer;
    found: boolean;
    k, code:integer;
    f: double;
begin
  s:=''; testln:=lnum;
  while (Ord(ch) in [Ord('A')..Ord('Z'),Ord('.'),Ord('_')]) and (testln=lnum) do begin
    s:=s+ch;
    getch;
  end; // while

  i:=USESSY;
  while (i<=APSHAPESY) and (s<>key[i]) do inc(i);
  found:=(i<>GUARD);
  if FOUND and (not erra) then begin
    case i of
      USESSY: begin {ch'='' '̂͂B}
                if pathDefined then begin
                  error(142);
                  exit;
                end;
                Yomitobashi;
                if ch<>'=' then begin
                  err_str:='';
                  error(140);
                  exit;
                end else begin
                  getch; {'='肷}
                  s:='';
                  repeat
                    if (ch<>' ') and (ch<>#9) then s:=s+ch;
                    getch;
                    if EndOfFile then begin {Z~RtYH}
                      error(141);
                      exit;
                    end;
                  until ch=';';
                  getch; {';'肷}
                //  if s[Length(s)]<>'\' then s:=s+'\';


                If s=ExtractFilename(s) then begin
                  gateDefFileDir:=ppgFilePath;
                  gateDefFilename:=s
                end else begin
                  gateDefFileDir:=ExtractFilePath(s);
                  gateDefFilename:=ExtractFileName(s);

                end;


                 // fBNgpXexplicitɂĂ
                 //    tpXw肵ƔfB
                 // fBNgȂŃt@Ĉ݂Ȃ
                 //  ppgt@CƓfBNgɂƂ݂ȂB
                 //   2009.8.26

                 // gateDefFileName:=gateDefFileDir + s;
                       // Whats new in 2009.2.9
                       // Q[g`t@Cexet@CƓfBNg
                       //   -->ppgƓfBNgɕύXI2009.7.14
                  pathDefined:=true;
                end;

              if gates.readFromFile(gateDefFileDir+gateDefFileName)=false then begin
                Messages.Assign(gates.messages);
                err_str:=gateDefFileName;
                error(101);
                exit;
              end;

              for k:=0 to gates.Count-1 do begin {\gĂȂ`FbN}
                if reservedWords.IndexOf(gates.gate[k].name)<>-1 then begin
                  err_str:=gates.gate[k].name; {\gĂI}
                  error(130);
                  exit;
                end; // if
              end; // k

              info.Assign(gates.info);

              end;  // USES
      CLKSY: begin
                Yomitobashi;
                if ch<>'=' then begin
                  err_str:='';
                  error(143);
                  exit;
                end else begin
                  getch; {'='肷}
                  s:='';
                  repeat
                    if (ch<>' ') and (ch<>#9) then s:=s+ch;
                    getch;
                    if EndOfFile then begin {Z~RtYH}
                      error(144);
                      exit;
                    end;
                  until ch=';';
                  getch; {';'肷}

                  Val(s,f,code);
                  if code<>0 then begin error(145); exit; end;
                  CLK:=f;  // CLK frequency in MHz

                end;
             end; // CLK
      TIMESY: begin timeVariables; end;
      LOOPSY: begin loopVariables; end;
      AMPSY: begin ampVariables; end;
      INTSY: begin intVariables; end;
      PHASESY: begin phaseVariables; end;
      FREQSY: begin freqVariables; end;
      AUXSY: begin setAuxParams; end;
      STARTSY: begin
                 // startR}hƓsɂ̓RgȊOLqĂ
                 // ȂȂAƂ[ɂI
                 while cnum<Length(lb) do begin
                   inc(CNUM);
                   CH := UPcase(LB[CNUM]);
                   if (ch<>' ') and (ch<>#9) then begin
                     if ch<>'{' then begin error(50);exit; end
                     else break;
                   end;
                   ok:=true;
                 end; //while
                 ok:=true;
               end;
      PHASELISTSY: begin setPhaseList; end; // phase cycle
      ACQPHASESY: begin setAcqPhase; end; // acq phase cycle
      AMPSHAPESY: begin setShape(shAmplitude); end; // amplitude shape pulse
      PHASESHAPESY: begin setShape(shPhase); end; // phase shape pulse
      APSHAPESY: begin setShape(shAP); end; // amp&phase shape pulse

    end; // case
  end else begin
    err_str:=s;
    lnum:=testln;
    error(66);
    exit;
  end;
end;

{-------------------------------------------------}
(****************************************************)
(************ ݒƒ`̉ ****************)
(****************************************************)
procedure TnmrPPG.initial;
var OK:boolean;
begin
  Yes_Loop:=false;
  {lp[0]:=false;} lp[1]:=false; lp[2]:=false;
  erra:=false;
  OK:=false;

  nextln;
  while not OK do begin
    if EndOfFile then begin error(41); exit; end;
    if erra then exit;
    case ch of
      '{': nextln; {RgȂ̂œǂݔ΂}
      'A'..'Z','.','_': auxCom(ok);
      ' ',';': getch;
      else begin
        error(6);
        if erra then exit;
      end; // else
    end; // case
  end; // while


  if not acqPhaseDefined then begin // acqPhaseListݒ肳ĂȂꍇB
    acqPhaseList.ReadFromString('acqphase=x;');  //IxƂ
    Warnings.Add(' [Warning] : Acquisition phase is not explicitly specified ' +
       'in the current pulse program, and accordingly it has been fixed to +x.');
  end;


  if auxParams.definePD=false then begin
    Warnings.Add(' [Warning] : Pulse delay (PD) is not explicitly specified, and accrodingly ' +
             'PD has been set to the default value of 1 s.');
    with Variables do begin
      add('PD',TimeVariable);
      Variable[count-1].Value:=1000000;       {}CNb}
      Variable[count-1].Numeric:=1;           {l}
      Variable[count-1].unitOfTime:=Seconds;  {Ԃ̒P}
      Variable[count-1].unitOfTimeString:='s';
      Variable[count-1].Caption:='Pulse Delay';
    end; // with

    auxParams.PD:=1000000;
    auxParams.definePD:=true;

  end;

  if auxParams.defineNA=false then begin
    Warnings.Add(' [Warning] : Number of acquisition (NA) is not explicitly specified, and accrodingly ' +
             'NA has been set to the default value of 1.');
    with Variables do begin
      add('NA',NAVariable);
      Variable[count-1].Value:=1;
      Variable[count-1].Numeric:=1;
      Variable[count-1].Caption:='Number of Accum';
    end; // with

    auxParams.NA:=1;
    auxParams.defineNA:=true;

  end;

  if auxParams.defineAL=false then begin
    Warnings.Add(' [Warning] : ' +
             'Acquisition length (AL) is not explicitly specified, and accordingly '
             + 'AL has been set to the default value of 1024.');
    with Variables do begin
      add('AL',LoopVariable);
      Variable[count-1].Value:=1024;
      Variable[count-1].Numeric:=1024;
      Variable[count-1].Caption:='Acquisition Length';
    end; // with
    auxParams.AL:=1024;
    auxParams.defineAL:=true;

  end;

  if auxParams.defineDW=false then begin
    Warnings.Add(' [Warning] : ' +
             'Dwell time (DW) is not explicitly specified, and accordingly '
             + 'DW has been set to the default value of 10 us.');
    with Variables do begin
      add('DW',TimeVariable);
      Variable[count-1].Value:=10;       {}CNb}
      Variable[count-1].Numeric:=10;           {l}
      Variable[count-1].unitOfTime:=microSeconds;  {Ԃ̒P}
      Variable[count-1].unitOfTimeString:='u';
      Variable[count-1].Caption:='Dwell time';
    end; // with

    auxParams.DW:=10;
    auxParams.defineDW:=true;

  end;


  with Variables do begin
    Add('ACQTIME', TimeVariable);
    Variable[count-1].Value:=auxParams.ACQTIME;
    Variable[count-1].Numeric:=auxParams.ACQTIME/1000;
    Variable[count-1].unitOfTime:=milliSeconds;
    Variable[count-1].unitOfTimeString:='m';
    Variable[count-1].Caption:='DW x AL';
  end;
  info.Add(' [Info] : '
           + 'ACQTIME has automatically been set to '
           + FloatToStr(auxParams.ACQTIME)
           + ' us (DW x AL)');

  if erra then begin exit; end;


  cnump:=cnum;
  lnump:=lnum;
end;
(****************************************************)
(***************** {p[` *******************)
(****************************************************)


{-------------------------------------------------}
(*************************************************)
(**************** Q[g̐ݒ *******************)
(*************************************************)
//------------------------------------------------------------------
procedure TnmrPPG.setAmplitudeGate(i:integer);
                  // i: ID number of the "gate" in the "gates"
var
  t:string;
  p,q:integer;
  r:single;
begin
  if ch<>'(' then begin
   {KindvpeBamplitudeȂK{BʂȂ΃G[}
    err_str:=gates.gate[i].name;
    error(105);
    exit;
  end;
  t:=''; getch;
  repeat
    if (ch<>' ') and (ch<>#9) then t:=t+ch;
    getch;
  until ch=')';
   // ̎_tɊi[ꂽampϐl̂ǂ炩B
  getch; {')'߂B}

  {ĂQ[g̃`l}
  {currentCHɓ΁AQ[gǉ}
  if gates.gate[i].channel=currentCH then begin
    if Variables.index(t)>-1 then begin // ϐ
      if Variables.Variable[Variables.index(t)].typeOfVar
                  <>ampVariable then begin
                      // ampϐȂI
         err_str:=t;
         error(301);
         exit;
      end;
    //-------------  ϐ ------------------
      with Variables do begin
        q:=index(t);
        r:=Variable[q].Value;   //  rɒl
        inc(Variable[q].FNOfWhereIs[currentCH-1]);
        SetLength(Variable[q].FWhereIs[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                  Variable[q].FWhereIs[currentCH-1][Variable[q].count(currentCH-1)-1]
                       :=compiledPPG[currentCH-1].Count;
                     // gpAhX̊i[
        SetLength(Variable[q].FGateIndex[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                  Variable[q].FGateIndex[currentCH-1][Variable[q].Count(currentCH-1)-1]
                       := i; // gpQ[gio[̊i[

      end; //with
    //-----------------------------------------
    end // if Variables.index(t)>-1
    else begin // l̏ꍇ̏
      val(t,r,p);
      if p<>0 then begin
        err_str:='Invalid expression: '+ gates.gate[i].name
                                       + '(' + t + ')';
        error(3);
        exit;
      end;
      if (r>100) or (r<0) then begin
        err_str:=gates.gate[i].name;
        error(300);
        exit;
      end;
    end;

    lineset:=lineset+gates.gate[i].output(r);
    gates.gate[i].used:=true;

  end; // gates.gate[i].channel=currentCH

end;
//------------------------------------------------------------------
procedure TnmrPPg.setPhaseGate(i: Integer);
var
  t:string;
  p,q:integer;
  r:single;
begin
          if ch<>'(' then begin
            {KindvpeBphaseȂK{BʂȂ΃G[}
            err_str:=gates.gate[i].name;
            error(105);
            exit;
          end;
          t:=''; getch;
          repeat
            t:=t+ch;
            getch;
          until ch=')';
            // ̎_tɊi[ꂽphaseϐl̂ǂ炩B
          getch; {')'߂B}

          {ĂQ[g̃`l}
          {currentCHɓ΁AQ[gǉ}
          if gates.gate[i].channel=currentCH then begin

            if Variables.index(t)>-1 then begin // ϐ
              if Variables.Variable[Variables.index(t)].typeOfVar
                     <>phaseVariable then begin
                      // phaseϐȂI
                err_str:=t;
                error(302);
                exit;
              end;
              //---------------  ϐ -------------------
              with Variables do begin
                q:=index(t);
                r:=Variable[q].Value;   //  rɒl
                inc(Variable[q].FNOfWhereIs[currentCH-1]);
                SetLength(Variable[q].FWhereIs[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FWhereIs[currentCH-1][Variable[q].count(currentCH-1)-1]
                     :=compiledPPG[currentCH-1].Count;
                        // gpAhX̊i[
                SetLength(Variable[q].FGateIndex[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FGateIndex[currentCH-1][Variable[q].Count(currentCH-1)-1]
                      := i; // gpQ[gio[̊i[

              end; //with
              //---------------------------------------------
            end // if Variables.index(t)>-1
            else begin // l̏ꍇ̏
              val(t,r,p);
              if p<>0 then begin
                err_str:='Invalid expression: ' + t ;
                error(3);
                exit;
              end;
            end;

            lineset:=lineset+gates.gate[i].output(r);
            gates.gate[i].used:=true;

          end; // if gates.gate[i].channel=currentCH

end;
//------------------------------------------------------------------
procedure TnmrPPG.setPhaseListGate(i: Integer);
var q:integer;
begin
  {ĂQ[g̃`l}
  {currentCHɓ΁AQ[gǉ}
  if gates.gate[i].channel=currentCH then begin
        // Q[gʎq"F#PHASELIST"TB
    q:= gates.index('F'
         +IntToStr(phListArray.phaseList[phListArray.index(gates.gate[i].name)].Channel)
              +'PHASELIST');
                    // q=-1̏ꍇ̓G[o悤ɂĂ
                    // (procedure setPhaseListɂ)
    if gates.gate[q].used then begin // dgp -> G[
      err_str:='F'
                + IntToStr(phListArray.phaseList[phListArray.index(gates.gate[i].name)].Channel)
                + 'PHASELIST ' ;
      error(37);
    end;

    lineset:=lineset+
             gates.gate[q].output(phListArray.phaseList[
                     phListArray.index(gates.gate[i].name)].Order);

    gates.gate[i].used:=true;
    gates.gate[q].used:=true;
  end; // if
end;
//------------------------------------------------------------------
procedure TnmrPPG.setLogicVectorGate(i: Integer);
var t:string;
    k,p,q:integer;
begin
          if ch<>'(' then begin
            {KindvpeBlogic_vectorȂK{BʂȂ΃G[}
            err_str:=gates.gate[i].name;
            error(108);
            exit;
          end;
          t:=''; getch;
          repeat
            t:=t+ch;
            getch;
          until ch=')';
               // ̎_tɊi[ꂽintϐl̂ǂ炩B
          getch; {')'߂B}

          {ĂQ[g̃`l}
          {currentCHɓ΁AQ[gǉ}
          if gates.gate[i].channel=currentCH then begin

            if Variables.index(t)>-1 then begin // ϐ
              if Variables.Variable[Variables.index(t)].typeOfVar
                     <>intVariable then begin
                      // intϐȂI
                err_str:=t;
                error(303);
                exit;
              end;
              //---------------  ϐ -------------------
              with Variables do begin
                q:=index(t);
                k:=Trunc(Variable[q].Value);   //  kɒl
                inc(Variable[q].FNOfWhereIs[currentCH-1]);
                SetLength(Variable[q].FWhereIs[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FWhereIs[currentCH-1][Variable[q].count(currentCH-1)-1]
                     :=compiledPPG[currentCH-1].Count;
                        // gpAhX̊i[
                SetLength(Variable[q].FGateIndex[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FGateIndex[currentCH-1][Variable[q].Count(currentCH-1)-1]
                       := i; // gpQ[gio[̊i[

              end; //with
              //---------------------------------------------
            end // if Variables.index(t)>-1
            else begin // l̏ꍇ̏
              val(t,k,p); {logic_vector̈͐}
              if p<>0 then begin
                err_str:='Invalid expression: ' + t ;
                error(3);
                exit;
              end;
            end;

            lineset:=lineset+gates.gate[i].output(k);
            gates.gate[i].used:=true;

          end; // if gates.gate[i].channel=currentCH

end;
//------------------------------------------------------------------
procedure TnmrPPG.setIntGate(i: Integer);
var t:string;
    k,p,q:integer;
begin
          if ch<>'(' then begin
            {KindvpeBintegerȂK{BʂȂ΃G[}
            err_str:=gates.gate[i].name;
            error(116);
            exit;
          end;
          t:=''; getch;
          repeat
            t:=t+ch;
            getch;
          until ch=')';
               // ̎_tɊi[ꂽintϐl̂ǂ炩B
          getch; {')'߂B}

          {ĂQ[g̃`l}
          {currentCHɓ΁AQ[gǉ}
          if gates.gate[i].channel=currentCH then begin

            if Variables.index(t)>-1 then begin // ϐ
              if Variables.Variable[Variables.index(t)].typeOfVar
                     <>intVariable then begin
                      // intϐȂI
                err_str:=t;
                error(303);
                exit;
              end;
              //---------------  ϐ -------------------
              with Variables do begin
                q:=index(t);
                k:=Trunc(Variable[q].Value);   //  kɒl
                inc(Variable[q].FNOfWhereIs[currentCH-1]);
                SetLength(Variable[q].FWhereIs[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FWhereIs[currentCH-1][Variable[q].count(currentCH-1)-1]
                     :=compiledPPG[currentCH-1].Count;
                        // gpAhX̊i[
                SetLength(Variable[q].FGateIndex[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FGateIndex[currentCH-1][Variable[q].Count(currentCH-1)-1]
                       := i; // gpQ[gio[̊i[

              end; //with
              //---------------------------------------------
            end // if Variables.index(t)>-1
            else begin // l̏ꍇ̏
              val(t,k,p); {logic_vector̈͐}
              if p<>0 then begin
                err_str:='Invalid expression: ' + t ;
                error(3);
                exit;
              end;
            end;

            lineset:=lineset+gates.gate[i].output(k);
            gates.gate[i].used:=true;

          end; // if gates.gate[i].channel=currentCH

end;
//------------------------------------------------------------------
procedure TnmrPPG.setAD9858Gate(i: Integer);
var t:string;
    psAD9858: shortInt;
    d: double;
    m,p,q:integer;
    defaultProfile:boolean;
begin
  if ch<>'(' then begin // AD9858ȂΈK{BʂȂ΃G[
    err_str:=gates.gate[i].name;
    error(109);
    exit;
  end;
  t:=''; getch;
  repeat
 //   if ch=')' then begin  {܂̓Z~RȂ΃G[}
 //     err_str:=gates.gate[i].name;
 //     error(110);
 //     exit;
 //   end;
 //   if not (Ord(ch) in [Ord(' '), Ord('-'), Ord('.'),Ord('0')..Ord('9')]) then begin
 //     err_str:=gates.gate[i].name;
 //     error(111);
 //     exit;
 //   end;
    if (ch<>' ') and (ch<>#9) then t:=t+ch;
    getch;
  until (ch=';') or (ch=')');
  if ch=')' then defaultProfile:=true else defaultProfile:=false;

  getch; // ';' or ')' 肷

  if not DefaultProfile then begin // Profile select (AD9858)
    psAD9858:=StrToInt(t);
    if (psAD9858<-1) or (psAD9858>3) then begin
      err_str:=gates.gate[i].name;
      error(111);
      exit;
    end;
    // psAD9858PS(Profile Select)i[B
    // -1frequency sweepp
    t:='';
    repeat
      if (ch<>' ') and (ch<>#9) then t:=t+ch;
      getch;
    until ch=')';
       // ̎_tɊi[ꂽfreqϐl̂ǂ炩B
    getch; // ')'߂B

  end else begin  // If ';' is not used, Profile -> 0
    psAD9858:=0;
    // ̏ꍇAłɕϐtɂ͎gf[^i[Ă͂
  end;

  // ĂQ[g̃`l}
  // currentCHɓ΁AQ[gǉ}
  if gates.gate[i].channel=currentCH then begin
    if Variables.index(t)>-1 then begin // ϐ
      if Variables.Variable[Variables.index(t)].typeOfVar<>freqVariable then begin
                      // freqϐȂI
        err_str:=t;
        error(304);
        exit;
      end;
      //---------------  ϐ -------------------
      with Variables do begin
        q:=index(t);
        d:=Variable[q].Value;   //  dɒl
        inc(Variable[q].FNOfWhereIs[currentCH-1]);
        SetLength(Variable[q].FWhereIs[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
        Variable[q].FWhereIs[currentCH-1][Variable[q].count(currentCH-1)-1]
                    :=compiledPPG[currentCH-1].Count;
                        // gpAhX̊i[
        SetLength(Variable[q].FGateIndex[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
        Variable[q].FGateIndex[currentCH-1][Variable[q].Count(currentCH-1)-1]
                       := i; // gpQ[gio[̊i[

        setLength(Variable[q].FPSAD9858[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
                Variable[q].FPSAD9858[currentCH-1][Variable[q].count(currentCH-1)-1]
                     := psAD9858;

                       // ̍sŎgpĂPSlۑ
      end; //with
      //---------------------------------------------
    end // if Variables.index(t)>-1
    else begin // l̏ꍇ̏
      val(t,d,p); {dɎgf[^B}
      if p<>0 then begin
        err_str:=gates.gate[i].name;
        error(112);
        exit;
      end;
    end;

    if (d<0) and (psAD9858<>-1) then begin
      err_str:='';
      error(113);
      exit;
    end;

    if psAD9858=-1 then d:=d*1e6 //ps=-1Ȃsweep modeAItZbgȂMHzP
    else d:=(d + gates.gate[i].offset)*1e6; // ItZbgāAMHzHzɕϊ

    Yes_AD9858:=true;
    gates.gate[i].used:=true;
    for m:=1 to 5 do AD9858Buffer[currentCH-1][m-1]:=gates.gate[i].output(psAD9858,d,m);
  end; // if gates.gate[i].channel=currentCH
end;
//------------------------------------------------------------------
procedure TnmrPPG.setLogicGate(i: Integer);
begin
 {ĂQ[g̃`l}
 {currentCHɓ΁AQ[gǉ}
 if gates.gate[i].channel=currentCH then begin
   lineset:=lineset+gates.gate[i].output;
   gates.gate[i].used:=true;
 end; // if
end;
//------------------------------------------------------------------
procedure TnmrPPG.Examine_Gates;
var s:string;
    i:integer;
begin
  for i:=0 to gates.Count-1 do begin
    gates.gate[i].used:=false;
    //sŌJԂgĂȂ`FbN邽߂
    //usedvpeBgBfalseɏĂB
  end;

  Yes_AD9858:=false;
    //XAD9858pQ[ggĂ邩ǂ`FbN̂}
    //tOƂĂbooleanϐgĂBfalseɏB}

  Yomitobashi;
  if ch=';' then begin
      getch;
      Yomitobashi;
  end;

//  repeat
  while ch<>')' do begin

    Yomitobashi;
    s:='';
    if Ord(ch) in [Ord('0')..Ord('9')]  then
      { begin
         Set_Gates;
        end }
      begin  {ŒڃQ[gw肷͎̂~߂ɂB}
             {}``l̎A`IɎwłȂB}
             {gQ[g̓Q[gϐ`ĂˁB}
             {Q[gϐɂ͎gp`li[悤ɂȂĂ邵B}
             {2006 Feb Takeda}
        error(106);
        exit;
      end
    else if Ord(ch) in [Ord('A')..Ord('Z'),Ord('@')] then begin
      repeat
        if (ch<>' ') and (ch<>#9) then s:=s+ch;
        getch
      until Ord(ch) in [Ord('('),Ord(','),Ord(')')];
      i:=gates.index(s);

      {`Ă邩`FbN}
      if i=-1 then begin {the gate identifier not found}
        err_str:=s;
        error(103);
        exit;
      end;

      if asyncQ then begin
        {asynchronousV[PX̏Ȃ}
        {asyncPPGList.asyncPPG[k].channel=currentCH}
        {Ƃ̂ƂŃRRɗĂ͂}
        if gates.gate[i].channel<>currentCH then begin
          err_str:='Cannot use "' + gates.gate[i].name
                   +'" defined as a gate parameter for channel '
                   + IntToStr(gates.gate[i].channel)
                   + ' in an asynchronous sequnece "'
                   + asyncPPGList.asyncPPG[asyncPPGList.itemIndex].name
                   + '" defined for channel '
                   + IntToStr(currentCH)
                   + '.';
          error(211);
          exit;
        end;
      end; // asyncQ

      if asyncFlag then begin
        if gates.gate[i].channel=currentCH then begin
          Warnings.Add('[Warning] : Gate parameter "'
                       + gates.gate[i].name
                       + '" used at line '
                       + IntToStr(lnum+lineOffset)
                       + ' was ignored, because channel '
                       + IntToStr(currentCH)
                       + ' is under asynchronous operation.');

        end;
      end;

      {ch'('Ȃ΃vpeB'Kind'
       amplitudephaselogic_vector̂͂BlogicG[}
      if (ch='(') and (gates.gate[i].Kind=logic) then begin
        err_str:=gates.gate[i].name;
        error(104);
        exit;
      end;

      {ch'('Ȃ΃vpeB'Kind'
       amplitudephaselogic_vector̂͂BphaseListG[}
      if (ch='(') and (gates.gate[i].Kind=phaseList) then begin
        err_str:=gates.gate[i].name;
        error(114);
        exit;
      end;


      if gates.gate[i].used=true then begin
        err_str:=gates.gate[i].name;
        error(107);
      end;

      case gates.gate[i].Kind of
        amplitude: setAmplitudeGate(i);
        phase: setPhaseGate(i);
        phaselist: setPhaseListGate(i);

        shape: begin
          if gates.gate[i].channel=currentCH then begin
            shapeQ:=true;
            with ShapedPulseArray do begin
              shapedPulseIndex:=Shape[index(gates.gate[i].name)].Order;

            end;
          end; // if gates.gate[i].channel=currentCH
        end; // shape

        logic: setLogicGate(i);

        logic_vector: setLogicVectorGate(i);
        int: setIntGate(i);
        AD9858: setAD9858Gate(i);

      end; // case

      while ch=' ' do getch;

      if ch=',' then begin
        getch; while (ch=' ') or (ch=#9) do getch;
                 {R}̎ɂ󔒁A^uȊO̕Qbg}
        if ch=')' then begin ERROR(8); exit end;
                {R}̌Ɋʂ͕ς}
      end;

    end else begin
      err_str:=ch;
      error(102);
      exit;
    end;

  end;   //while
//  until ch=')';

  getch;
end;
{-------------------------------------------------}
procedure Tnmrppg.import;
var s,ifn:string;

begin
  if imported=true then begin  // import[h̎Aimport͕sI
    error(800);
    exit;
  end;

  Yomitobashi;
    if ch<>'(' then begin
      err_str:='';
      error(801);
      exit;
    end else begin
      getch; {'('肷}
      s:='';
      repeat
        if (ch<>' ') and (ch<>#9) then s:=s+ch;
        getch;
        if EndOfFile then begin {Z~RtYH}
          error(802);
          exit;
        end;
      until ch=')';
    end;

  If s=ExtractFilename(s) then ifn:=ppgFilePath + s
                          else ifn:=s;

  if FileExists(ifn) then begin
    if importFileNames.IndexOf(ifn)<0 then importFileNames.Add(ifn);
     // import file o^BdāB
    importedLines.LoadFromFile(ifn);
    imported:=true;
    lnum_bak:=lnum;
    lnum:=0;

    currentImportedFileIndex:=importFileNames.IndexOf(ifn);

    getch; {')'肷}
  end else begin
    err_str:=ifn;
    error(803);
    exit;
  end;

end;
{-------------------------------------------------}
procedure TnmrPPG.init;
var q:integer;
begin

  lnump:=lnum-1; cnump:=cnum;

//  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
//                                + ' = '
//                                + 'A1 00000186A0 00000000 00000000');
//                                   {10^5JEgB10^5=$186A0}
//  mnemonic[currentCH-1].Add('OUT (0xA1) for 100000=0x186A0 counts');

  with Variables do begin
    q:=index('NA');
    inc(Variable[q].FNOfWhereIs[currentCH-1]);
    SetLength(Variable[q].FWhereIs[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
    Variable[q].FWhereIs[currentCH-1][Variable[q].count(currentCH-1)-1]
            :=compiledPPG[currentCH-1].Count;

    SetLength(Variable[q].FGateIndex[currentCH-1],Variable[q].FNOfWhereIs[currentCH-1]);
    Variable[q].FGateIndex[currentCH-1][Variable[q].Count(currentCH-1)-1]
                       := -1;
         // [vϐ̓Q[gƖ֌WȂ̂ŁAƂ肠_~[Ƃ-1ƂĂB

  end; //with

  {LOOP0 (ώZpJE^)}
  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' = '
                                + '10 '
                                + '00'+Hex(auxParams.NA,8)
                                + ' 00000000 00000000');

  NOfLinesPerCommand[currentCH-1].Add('1');

  mnemonic[currentCH-1].Add('LOOP0 (0x10) with number of iterations set by %NA='
               + IntToStr(auxParams.NA)
               + ' (0x' + '00'+Hex(auxParams.NA,8) + ')');
                  {0x10: loop0}

  // 'ready' command added (2013.6.15 Takeda)
  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' FF '
                                + '0000000000'
                                + ' 00000000 00000000');

  NOfLinesPerCommand[currentCH-1].Add('1');

  mnemonic[currentCH-1].Add('READY (0xFF): Standby until all other rf channels '
                       +'have been ready for the next operation.'
                       );


  dec(lineOffset);
         // ŕt1s(pulse(1u;F#PhasePointer_Init))␳B
         // G[ꍇȀꏊisj𐳊mɎwł悤ɂ邽

end;
{-------------------------------------------------}
procedure TnmrPPG.wait_hl(hl:byte);
var
  a1,a2,s,t,w:string;
  c,k:integer;
begin
  if not checkCH('(') then exit;

  lineSet:=[];
  Uline:=0; Lline:=0;
  s:=''; t:='';

  getch;
  while (ch<>';') and (ch<>')') do begin
    if (ch<>' ') and (ch<>#9) then t:=t+ch;
    getch;
  end;

  if t='TRIG_E' then w:='0000000001'
  else if t='TRIG_R' then w:='0000000002'
  else if t='TRIG_0' then w:='0000000004'
  else if t='TRIG_1' then w:='0000000008'
  else if t='TRIG_2' then w:='0000000010'
  else if t='TRIG_3' then w:='0000000020'
  else error(250);


  {̎_ch";"")"Ȃ͂}
  if not ((ch=';') or (ch=')')) then begin error(251); end;
  lineset:=[];
  if ch=';' then begin
    shapeQ:=false;
    Examine_Gates;
  end else getch;

  if shapeQ then begin
    error(252);
    exit;
  end;

  if asyncFlag then exit; {asyncronousA֌WȂ̂łprocedureexit
                           A@`FbN͂Ă̂begin̒ɂ
                           ̏ꏊasyncFlag`FbN}

  if Yes_AD9858 then begin
    error(115);
    exit;
  end;

  for k:=0 to 31 do begin
    if k in lineset then Lline:=Lline or (1 shl k);
    if k+32 in lineset then Uline:=Uline or (1 shl k);
  end;

  s:=''; c:=0;
  for k:=0 to 63 do begin
    if k in lineset then begin s:=s+IntToStr(k)+','; inc(c); end;
  end;

  if c>0 then begin
    if (s[Length(s)]=',') then Delete(s,length(s),1);
  end;


  if hl=0 then begin // wait_L
    a1:=' A3 ';
    a2:='WAIT_LOW (0xA3)';
  end else begin
    a1:=' A2 ';
    a2:='WAIT_HIGH (0xA2)';
  end;

  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + a1
                                + w
                                + ' '
                                + Hex(ULine,8)
                                + ' '
                                + Hex(LLine,8) );

  NOfLinesPerCommand[currentCH-1].Add('1');

  if c>0 then begin
    if c=1 then w:= 'with an active gate ' + s
    else w:= 'with active gates ' + s;
  end else w:= 'with no active gates';

  mnemonic[currentCH-1].Add(a2+' for the '+ t +
                     ' flag ' + w );

end;
{-------------------------------------------------}
procedure TnmrPPG.pulse;
var
  Uwidth:word;
  Lwidth:cardinal;
  c,k,m:integer;
  s,t,com:string;
  q,r:double;
begin
  lineSet:=[];
  Uline:=0; Lline:=0;
  duration(Uwidth,Lwidth); {Ԑݒ}
  q:=(Uwidth*4294967296)+Lwidth+4; {4294967296=2^32}
                                   {4: excess count consumed by FPGA pulser}
  {qɂ̍s̃NbNJEgi[Ă}

  {̎_ch')' or ';'}

  if not ((ch=';') or (ch=')')) then begin error(10); end;

  if ch=';' then begin
    shapeQ:=false;
    Examine_Gates;
  end else getch;

  if not asyncFlag then begin {ASYNCSYNC̊Ԃȉ͔΂}

  if not Yes_AD9858 then begin
     {̍sAD9858g䂪gĂȂ}
     {ȉ ++++++  +++++܂łs}
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  for k:=0 to 31 do begin
    if k in lineset then Lline:=Lline or (1 shl k);
    if k+32 in lineset then Uline:=Uline or (1 shl k);
  end;

  s:=''; c:=0;
  for k:=0 to 63 do begin
    if k in lineset then begin s:=s+IntToStr(k)+','; inc(c); end;
  end;

  if c>0 then begin
    if (s[Length(s)]=',') then Delete(s,length(s),1);
  end;

  if shapeQ then com:='AF' else com:='A1';

  if shapeQ then begin // shaped pulsegĂI
    // ܂shape_Kreg̐ݒB
    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                              + ' ='
                              + ' 20 '
                              + '0000000' + Hex(shapedPulseIndex,3)
                              + ' '
                              + '00000000'
                              + ' '
                              + '00000000'
                               );
    NOfLinesPerCommand[currentCH-1].Add('1');
    mnemonic[currentCH-1].Add(shapedPulseArray.Shape[shapedPulseIndex].Name +
                    ': ' + IntToStr(shapedPulseIndex)
                    +'th shaped pulse in channel ' +
                    IntToStr(currentCH)
                    );


    shapedPulseArray.Shape[shapedPulseIndex].updateCount:=
               Trunc(q/shapedPulseArray.Shape[shapedPulseIndex].Count);
               // q ɂ̓JEgĂ

    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' 22 '
                                + '000000'
                                + Hex(shapedPulseArray.Shape[shapedPulseIndex].updateCount,4)
                              + ' '
                              + '00000000'
                              + ' '
                              + '00000000');

    NOfLinesPerCommand[currentCH-1].Add('1');

    mnemonic[currentCH-1].Add(shapedPulseArray.Shape[shapedPulseIndex].Name
                               + ': update count is '
                               + IntToStr(shapedPulseArray.Shape[shapedPulseIndex].updateCount)

                               );

    shapeQ:=false;

  end; // shapeQ


  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' ' + com + ' '   // ' A1 ' or ' AF '
                                + Hex(UWidth,2)
                                + Hex(LWidth,8)
                                + ' '
                                + Hex(ULine,8)
                                + ' '
                                + Hex(LLine,8) );

  NOfLinesPerCommand[currentCH-1].Add('1');

  if c>0 then begin
    if c=1 then t:= 'with an active gate ' + s
    else t:= 'with active gates ' + s;
  end else t:= 'with no active gates';

  mnemonic[currentCH-1].Add('OUT (0x'+com+') for '
             + FloatToStr(q/CLK) + ' us ('
             + FloatToStr(q) + '=0x'+ Hex(Uwidth,2)+Hex(LWidth,8)+'+4 counts) '
             + t);
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}
  end else begin {̍sAD9858g䂪gĂꍇ̏}

  if q<50 then begin {ԂȂI}
    err_str:=''; dec(lnum);
    error(400);
    exit;
  end;


  for m:=1 to 5 do begin
    Uline:=0; Lline:=0;
    for k:=0 to 31 do begin
      if k in lineset+AD9858Buffer[currentCH-1][m-1] then
                                  Lline:=Lline or (1 shl k);
      if k+32 in lineset+AD9858Buffer[currentCH-1][m-1] then
                                  Uline:=Uline or (1 shl k);
    end; {k}

    //
    // ԕBm=1-4Ȃ10/CLK us.  m=5ȂΎc(q-40)/CLK us
    // g[^qƂȂ悤ɂB
    // q̓JEg񐔁B100M CLKȂ0.01 usɑB
    //
    if m=5 then r:=(q-40.0)/CLK else r:=10.0/CLK;
    Jikann_Hennkann(r, UWidth,LWidth);


    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                  + ' ='
                                  + ' A1 '
                                  + Hex(UWidth,2)
                                  + Hex(LWidth,8)
                                  + ' '
                                  + Hex(ULine,8)
                                  + ' '
                                  + Hex(LLine,8) );


    if m=1 then NOfLinesPerCommand[currentCH-1].Add('5') else
                NOfLinesPerCommand[currentCH-1].Add('0');
      // 5s₵ĂẮAϐl̃Abvf[g̍ۂɕKvƂȂ


    s:=''; c:=0;
    for k:=0 to 63 do begin
      if k in lineset+AD9858Buffer[currentCH-1][m-1] then begin
        s:=s+IntToStr(k)+',';
        inc(c);
      end;
    end;

    if c>0 then begin
      if (s[Length(s)]=',') then Delete(s,length(s),1);
    end;

    if c>0 then begin
      if c=1 then t:= 'with an active gate ' + s
      else t:= 'with active gates ' + s;
    end else t:= 'with no active gates';

    mnemonic[currentCH-1].Add('OUT (0xA1) for '
               + FloatToStr(r) + ' us ('
               + IntToStr(Trunc(r*CLK)) + '=0x'+ Hex(Uwidth,2)+Hex(LWidth,8)+'+4 counts) '
               + t);

  end; {m}


  end; {if Yes_AD9858}

  end; {if not asyncFlag}

end;
{-------------------------------------------------}
procedure Tnmrppg.burst;
var
  c,k,m:integer;
  s,t,com:string;
begin
  if not (ch='(') then begin error(10); end;
  lineSet:=[];
  Uline:=0; Lline:=0;

  GETCH;
  Examine_Gates;

  if not asyncFlag then begin {ASYNCSYNC̊Ԃȉ͔΂}

    if Yes_AD9858 then begin error(601); end;
     {burst modeAD9858g䂪gĂG[}

    for k:=0 to 31 do begin
      if k in lineset then Lline:=Lline or (1 shl k);
      if k+32 in lineset then Uline:=Uline or (1 shl k);
    end;

    s:=''; c:=0;
    for k:=0 to 63 do begin
      if k in lineset then begin s:=s+IntToStr(k)+','; inc(c); end;
    end;

    if c>0 then begin
      if (s[Length(s)]=',') then Delete(s,length(s),1);
    end;

    com:='BA';

    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' ' + com + ' '
                                + '00'
                                + '00000000'
                                + ' '
                                + Hex(ULine,8)
                                + ' '
                                + Hex(LLine,8) );

  NOfLinesPerCommand[currentCH-1].Add('1');

  if c>0 then begin
    if c=1 then t:= 'with an active gate ' + s
    else t:= 'with active gates ' + s;
  end else t:= 'with no active gates';

  mnemonic[currentCH-1].Add('BURST (0x'+com+') '+ t);


  end; {if not asyncFlag}
end;
{-------------------------------------------------}
procedure TnmrPPG.delay;
var
  Uwidth:word;
  Lwidth:cardinal;
  q:double;
begin
  lineSet:=[];
  Uwidth:=0; Lwidth:=0;
  duration(Uwidth,Lwidth); {Ԑݒ}
  {̎_ch')'}
  checkCH(')');
  if erra then exit;

  if not asyncFlag then begin {ASYNCSYNC̊Ԃȉ͔΂}
    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' A1 '
                                + Hex(UWidth,2)
                                + Hex(LWidth,8)
                                + ' 00000000 00000000');
    NOfLinesPerCommand[currentCH-1].Add('1');
    q:=(Uwidth*4294967296)+Lwidth+4; {4294967296=2^32}
                                   {4: excess count consumed by FPGA pulser}

    mnemonic[currentCH-1].Add('OUT (0xA1) for '
               + FloatToStr(q/CLK) + ' us ('
               + FloatToStr(q) + '=0x'+ Hex(Uwidth,2)+Hex(LWidth,8)+'+4 counts) '
               + 'with no active gates');

  end; {if not asyncFlag}

  getch;  {)߂}
end;
{-------------------------------------------------}
procedure TnmrPPG.getLoopUnit(var N:lp_type);
var X:double;
begin
  Search('(');
  getch;
  Number(X);
  if erra then exit;

  if (X>2.0) or (X<1.0) then begin error(21); exit; end;
  N:=Trunc(X);

end;
{-------------------------------------------------}
procedure TnmrPPG.loop;
var LOOPUNIT:lp_type;
    Y: double;
    times: longInt;
    CM: string;
begin
  getLoopUnit(loopunit);

  if lp[loopunit] then begin
    err_str := IntToStr(loopunit);
    error(22);
    exit;
  end;

  search(';'); getch;

  Yes_loop:=true; Number(Y); Yes_loop:=false;
  if erra then exit;
  if (Y>65535.0) or (Y<1) then begin error(5); exit end;

  times:=trunc(Y);

  if checkCH(')') then begin
    getch;
    case loopUnit of
      1: CM:='11';
      2: CM:='12';
      else begin error(21); exit; end;
    end; // case

  if not asyncFlag then begin {ASYNCSYNC̊Ԃȉ͔΂}

    lp[loopunit]:=true;

    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' = '
                                + CM
                                + ' '
                                + Hex(times,10)
                                + ' 00000000 00000000');
    NOfLinesPerCommand[currentCH-1].Add('1');
    mnemonic[currentCH-1].Add('LOOP'+IntToStr(loopUnit)+' (0x'
                                + CM + ') with number of iterations '
                                + IntToStr(times)
                                + ' (0x' + Hex(times,10) + ')');
  end; {if not asyncFlag}
  end;
end;
{-------------------------------------------------}
procedure TnmrPPG.endLoop;
var
  loopUnit:lp_type;
  CM:string;
begin
  getLoopUnit(loopUnit);

  if checkCH(')') then begin
    getch;
    case loopUnit of
      1: CM:='19';
      2: CM:='1A';
      else begin error(21); exit; end;
    end; // case

    if not asyncFlag then begin {ASYNCSYNC̊Ԃȉ͔΂}

      if not lp[loopunit] then begin
        err_str := IntToStr(loopunit);
        dec(lnum);
        error(23);
        exit;
      end;

      lp[loopunit]:=false;

      compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                  + ' = '
                                  + CM
                                  + ' 0000000000 00000000 00000000');
      NOfLinesPerCommand[currentCH-1].Add('1');
      mnemonic[currentCH-1].Add('END_LOOP'+IntToStr(loopUnit)+' (0x' + CM + ')');
    end; {if not asyncFlag}

    if (lp[loopunit]) and (asyncFlag) then begin
   // else begin
      err_str:='';
      dec(lnum);
      error(24);
      exit;
    end;
  end;
end;
{-------------------------------------------------}
procedure TnmrPPG.async;
var k:integer;
    s:string;
begin
  s:='';
  Yomitobashi;
  if ch<>'(' then begin
    dec(lnum);
    error(228);
    exit;
  end;

  getch; {(肷}
  repeat
    if EndOfFile then begin error(226); end;
    if (ch<>' ') and (ch<>#9) then s:=s+ch;
    getch;
  until ch=')';
  getch; {)肷}

  k:=asyncPPGList.index(s);
  if k=-1 then begin
     {ȖOasynchronousV[PX݂͑ȂI}
    err_str:=s;
    error(224);
    exit;
  end;

  if currentCH=asyncPPGList.asyncPPG[k].channel then begin
    {ݏ̃`lƈv΁Aȉ̏sB}

    if asyncFlag=true then begin {asynchronousV[PXs}
      err_str:= '"'+ asyncPPGList.asyncPPG[k].name + '" '
               + 'here, because another asynchronous sequence '
               + 'for channel ' + IntToStr(currentCH) + ' is already active.';
      error(232);
      exit;
    end;

    asyncFlag:=true; {SYNCs܂ł̊trueɂ}

    with asyncPPGList do begin
      inc(asyncPPG[k].FNOfWhereIs);
      SetLength(asyncPPG[k].FWhereIS,asyncPPG[k].FNOfWhereIs);
      asyncPPG[k].FWhereIs[asyncPPG[k].Count-1]:=compiledPPG[currentCH-1].Count;
      asyncPPG[k].used:=true;
    end; {with asyncPPGList}

    compiledPPG[currentCH-1].Add('');
    NOfLinesPerCommand[currentCH-1].Add('1');
    mnemonic[currentCH-1].Add('');
    {uNsǉBƂŃWv悪m肵疄߂邩
     SzȂł}

  end; //if

  if currentCH=1 then begin;

    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' A1 0000000000 '
                                + '00000000 00000000'
                                ); {0 counts + excess 4 counts -> 4 counts}
                                   {4: excess count consumed by FPGA pulser}
    NOfLinesPerCommand[currentCH-1].Add('1');
    mnemonic[currentCH-1].Add('OUT (0xA1) for '
               + '4 CLK counts (ASYNC transition)'
              );

    end; // if currentch=1


end;
{-------------------------------------------------}
procedure TnmrPPG.sync;
var q:integer;
    s,t,u:string;
    sy2,sy3,ok:boolean;
begin

  sy2:=false; sy3:=false;

  Yomitobashi;
  if ch<>'(' then begin
    dec(lnum);
    error(229);
    exit;
  end;

  getch; {(肷}

  ok:=false;
  while not ok do begin

    s:='';
    repeat
      Yomitobashi;
      if not (Ord(ch) in [Ord('0')..Ord('9')]) then begin
        err_str:='';
        error(230);
        exit;
      end;
      s:=s+ch;
      getch;
    until Ord(ch) in [Ord(','),Ord(')')];

    q:=StrToInt(s);

    if not ((q=2) or (q=3)) then begin error(231); exit; end;

    if q=2 then begin
      if sy2=true then begin
        err_str:='2';
        error(234);
        exit;
      end;
      sy2:=true;
    end;

    if q=3 then begin
      if sy3=true then begin
        err_str:='3';
        error(234);
        exit;
      end;
      sy3:=true;
    end;

    if q=currentCH then begin
      if asyncFlag=false then begin
        err_str:=IntToStr(currentCH);
        error(233);
        exit;
      end else begin
        asyncFlag:=false;
      end;
    end;

    if ch=')' then ok:=true  {)ȂΎ}
    else begin {R}̏ꍇAɈ͂}
      getch;
      Yomitobashi;
    end;

  end; {while not ok}

  getch; {)肷}

  if currentCH=1 then begin;

  {TODO: SYNC PULSE}
    if sy2 then begin
      t:='40000000 00000000';
      u:='channel 2';
    end;
    if sy3 then begin
      t:='80000000 00000000';
      u:='channel 3';
    end;
    if (sy2) and (sy3) then begin
      t:='C0000000 00000000';
      u:='channels 2 and 3';
    end;

    compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' A1 0000000000 '
                                + t
                                ); {0 counts + excess 4 counts -> 4 counts}
                                   {4: excess count consumed by FPGA pulser}
    NOfLinesPerCommand[currentCH-1].Add('1');
    mnemonic[currentCH-1].Add('OUT (0xA1) for '
               + '4 CLK counts (SYNC pulse to '
               + u + ')');

    end; // if currentch=1



end;
{-------------------------------------------------}
procedure TnmrPPG.relax;
var
  i,k,c:integer;
 // N: double;
  Uwidth:word;
  Lwidth:cardinal;
  s,t:string;
  Uline,Lline:cardinal;
  q:double;
begin

  with Variables do begin
    i:=index('PD');
    if defined('PD') then begin
      if not asyncFlag then begin
        inc(Variable[i].FNOfWhereIs[currentCH-1]);
        SetLength(Variable[i].FWhereIs[currentCH-1],Variable[i].FNOfWhereIs[currentCH-1]);
        Variable[i].FWhereIs[currentCH-1][Variable[i].count(currentCH-1)-1]
          :=compiledPPG[currentCH-1].Count;

        SetLength(Variable[i].FGateIndex[currentCH-1],Variable[i].FNOfWhereIs[currentCH-1]);
        Variable[i].FGateIndex[currentCH-1][Variable[i].Count(currentCH-1)-1]
                       := -1;
         // ԕϐ(PD)̓Q[gƖ֌WȂ̂ŁAƂ肠_~[Ƃ-1ƂĂB

      end; {if not asyncFlag}

   //   N:=Variable[i].Value;

    end else begin
      err_str:='PD';
      error(0);
    end;
  end; // with

  // PD҂
  Jikann_Hennkann(auxParams.PD,Uwidth,LWidth);

  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' A1 '
                                + Hex(UWidth,2)
                                + Hex(LWidth,8)
                                + ' 00000000 00000000');
  NOfLinesPerCommand[currentCH-1].Add('1');
  mnemonic[currentCH-1].Add('OUT (0xA1) for a time interval set by PD='
               + FloatToStr(auxParams.PD/1000000) + ' s.');


  // TOGGLEPHASE
  lineSet:=[];
  Jikann_Hennkann(1.0,Uwidth,LWidth);

  q:=(Uwidth*4294967296)+Lwidth+4; {4294967296=2^32}
                                   {4: excess count consumed by FPGA pulser}
  {qɂ̍s̃NbNJEgi[Ă}


  s:='F'+IntToStr(currentCH)+'TOGGLEPHASE';
  k:=gates.index(s);
  if k=-1 then begin err_str:=s; error(36); end;
  lineset:=gates.gate[k].output;

  Uline:=0; Lline:=0;
  for k:=0 to 31 do begin
    if k in lineset then Lline:=Lline or (1 shl k);
    if k+32 in lineset then Uline:=Uline or (1 shl k);
  end;

  s:=''; c:=0;
  for k:=0 to 63 do begin
    if k in lineset then begin s:=s+IntToStr(k)+','; inc(c); end;
  end;

  if c>0 then begin
    if (s[Length(s)]=',') then Delete(s,length(s),1);
  end;

  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                  + ' ='
                                  + ' A1 '
                                  + Hex(UWidth,2)
                                  + Hex(LWidth,8)
                                  + ' '
                                  + Hex(ULine,8)
                                  + ' '
                                  + Hex(LLine,8) );
  NOfLinesPerCommand[currentCH-1].Add('1');
  if c>0 then begin
    if c=1 then t:= 'with an active gate ' + s
    else t:= 'with active gates ' + s;
  end else t:= 'with no active gates';

  mnemonic[currentCH-1].Add('OUT (0xA1) for '
             + FloatToStr(q/CLK) + ' us ('
             + FloatToStr(q) + '=0x'+ Hex(Uwidth,2)+Hex(LWidth,8)+'+4 counts) '
             + t);




  //
  // ʐMI܂ő҂
             // when X"A3" => -- WAIT_LOW
             // latchGReg <= '1';
             // trigWReg <= argReg(5 downto 0);
             // stateReg <= WL;
  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' A3 '
                                + '0000000002'
                                + ' 00000000 00000000');
  NOfLinesPerCommand[currentCH-1].Add('1');
  mnemonic[currentCH-1].Add('WAIT_LOW (0xA3) for the RCVR_BUSY flag');






  {END_LOOP0}
  {LOOP0͐ώZp̃[vJE^}
  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' = 18 0000000000 00000000 00000000');
  NOfLinesPerCommand[currentCH-1].Add('1');
  mnemonic[currentCH-1].Add('END_LOOP0 (0x18)');



  {NbV}
  {RȂNA(ώZ)=1̂Ƃpulser肭IȂ悤}
  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' = '
                                + 'A1 00000186A0 00000000 00000000');
                                   {10^5JEgB10^5=$186A0}
  NOfLinesPerCommand[currentCH-1].Add('1');
  mnemonic[currentCH-1].Add('OUT (0xA1) for 100000=0x186A0 counts');


  {END PPG}
  compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' = 00 0000000000 00000000 00000000');
  NOfLinesPerCommand[currentCH-1].Add('1');
  mnemonic[currentCH-1].Add('END_PPG (0x00)');

  EndOfFile:=True;
  {Assemble̓pXvÕCIƂmB}
  {asynchronous̏i΁jI}

end;
{-------------------------------------------------}
(*****************************************************)
(************** RpC[̖{ *******************)
(*****************************************************)

{******************* SYMBOL̉*********************}
procedure TnmrPPG.symbol;
type
  SYM=(INITSY, PULSESY, BURSTSY, DELAYSY, LOOPSY,
       ENDSY, ASYNCSY, SYNCSY, ACCUMSY,
       RELAXSY, READYSY, EXTTRIGSY, WAITHSY,
       WAITLSY, IMPORTSY, GUARD);
const
  KEY: ARRAY[INITSY..IMPORTSY] of STRING
  =('INIT','PULSE','BURST','DELAY','LOOP','ENDLP','ASYNC','SYNC',
    'ACCUM','RELAX','READY','EXTTRIG','WAIT_H','WAIT_L','IMPORT');
  var
    S: STRING;
    I: SYM;
    testln:integer;
    FOUND: BOOLEAN;
begin
 S:='';  testln:=lnum;
 while (Ord(CH) IN [Ord('A')..Ord('Z'),Ord('_')]) and (testln=lnum) {and (not EndOfFile)} DO
  begin
   S := S + CH;
   GETCH;
  end;

 I := INITSY;
 while (I<=IMPORTSY) AND (S<>KEY[I]) DO INC(I);
  FOUND := (I<>GUARD);
  if FOUND and (not erra) then
   case I of
    INITSY: begin
        if asyncQ then begin
          dec(lnum);  {G[s\PYĂ悤Ȃ̂ŁA}uIH}
          error(220);
          exit;
        end;
        Init;
        if erra then exit;
      end;

    PULSESY: begin
        PULSE;
        if erra then exit;
      end;

    BURSTSY: begin
        BURST;
        if erra then exit;
      end;

    DELAYSY: begin
        DELAY;
        if erra then exit;
      end;

    LOOPSY:  begin
        LOOP;
        if erra then exit;
      end;

    ENDSY:   begin
        endLOOP;
        if erra then exit;
      end;

    ASYNCSY:
      begin
        if asyncQ then begin
          dec(lnum);  {G[s\PYĂ悤Ȃ̂ŁA}uIH}
          error(222);
          exit;
        end;

        begin ASYNC; if erra then exit end;

      end; // ASYNCSY

    SYNCSY:
      begin
        if asyncQ then begin
          dec(lnum);  {G[s\PYĂ悤Ȃ̂ŁA}uIH}
          error(223);
          exit;
        end;

        begin
          SYNC;
          if erra then exit;
        end;

      end;
    ACCUMSY: begin {ACCUM; if erra then exit;}  end;
    RELAXSY:
      begin
        if asyncQ then begin
          dec(lnum);  {G[s\PYĂ悤Ȃ̂ŁA}uIH}
          error(221);
          exit;
        end;

        if asyncFlag then begin
          dec(lnum);
          error(225);
          exit;
        end;

        RELAX;
        if erra then exit;
      end;

    READYSY:
      begin
        compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' FF '
                                + '0000000000'
                                + ' 00000000 00000000');

        NOfLinesPerCommand[currentCH-1].Add('1');

        mnemonic[currentCH-1].Add('READY (0xFF): Standby until all other rf channels '
                       +'have been ready for the next operation.'
                       );

      end;

    EXTTRIGSY:
      begin
        compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' A2 '  // WAIT_HIGH
                                + '0000000001'  // The 0th bit corresponds to the ext trig.
                                + ' 00000000 00000000');

        NOfLinesPerCommand[currentCH-1].Add('1');

        mnemonic[currentCH-1].Add('WAIT_HIGH (0xA2): Standby until '
                       +'the external triger has been detected.'
                       );

      end;

    WAITHSY:  // wait_H
      begin
        wait_hl(1);
        if erra then exit;
      end;

    WAITLSY:  // wait_L
      begin
        wait_hl(0);
        if erra then exit;
      end;

    IMPORTSY: // import
      begin
        import;
        if erra then exit;
      end;

   end
   else begin
     err_str:=s;
     lnum:=testln; {G[̏ꏊlnumɊi[}
     ERROR(66);
     exit;
   end;
end;

{******************** ASSEMBLE *************************}
procedure TnmrPPG.assemble; {ɂ͂ߕcĂȂ}
var ippp: 0..2;
    j,k: integer;
begin

  erra:=false;
  EndOfFile := FALSE;
  nextln;
  //cnum:=cnump; //getch;
  asyncQ:=false;    {async̃RpCp}
  asyncFlag:=false; {C̃RpCpBasyncJnsync܂ł̊trueɂȂ}

  lineOffset:=preamble.Count-phListArrayOffset-1;


//--------  SHAPED PULSE SETUP  ----------
  k:=0;
  while k<shapedPulseArray.Count do begin

    if shapedPulseArray.Shape[k].Channel=currentCH then begin
         // s`lshaped pulsẽR}hǉI

     // shape_kReg <= shape[k].order
      compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' 20 '
                                + '0000000' + hex(shapedPulseArray.Shape[k].order,3)
                                + ' 00000000 00000000');
      NOfLinesPerCommand[currentCH-1].Add('1');
      mnemonic[currentCH-1].Add(shapedPulseArray.Shape[k].Name
                               + ': '
                               + IntToStr(shapedPulseArray.Shape[k].order)
                               + 'th shaped pulse in channel '
                               + IntToStr(shapedPulseArray.Shape[k].channel)
                               );

      compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' 21 '
                                + '0000000000'
                                + ' 00000000 '+
                                   hex(shapedPulseArray.shape[k].Count,4)+
                                   hex(shapedPulseArray.shape[k].address,4));
      NOfLinesPerCommand[currentCH-1].Add('1');
      mnemonic[currentCH-1].Add(shapedPulseArray.Shape[k].Name
                               + ': starting address is '
                               + IntToStr(shapedPulseArray.Shape[k].address)
                               + ', number of steps is '
                               + IntToStr(shapedPulseArray.Shape[k].Count)
                               );

     // 24̒Oɂ20KvI
      compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' 20 '
                                + '0000000' + hex(shapedPulseArray.Shape[k].order,3)
                                + ' 00000000 00000000');
      NOfLinesPerCommand[currentCH-1].Add('1');
      mnemonic[currentCH-1].Add(shapedPulseArray.Shape[k].Name
                               + ': '
                               + IntToStr(shapedPulseArray.Shape[k].order)
                               + 'th shaped pulse in channel '
                               + IntToStr(shapedPulseArray.Shape[k].channel)
                               );

      for j:=0 to shapedPulseArray.Shape[k].nOfLines-1 do begin
        compiledPPG[currentCH-1].Add(hex(compiledPPG[currentCH-1].Count,3)
                                + ' ='
                                + ' 24 '

                                + shapedPulseArray.shape[k].genBin(j)

                                );

         NOfLinesPerCommand[currentCH-1].Add('1');

         mnemonic[currentCH-1].Add(shapedPulseArray.Shape[k].Name
                               + ': ' + IntToStr(j)
                               + 'th line of the shaped pulse ('

                    + IntToStr(min(shapedPulseArray.Shape[k].Count,j*5+5))
                    + 'th down to '
                    + IntToStr(j*5)
                    +'th data )'

                              );

       end; // j
    end; // if
    inc(k);
  end; // while shapedPulseArray.NOfLists(currentCH)
//----------------------------------------


  while not EndOfFile DO begin
    if erra then exit;
      case CH of
        'A'..'Z','_' : SYMBOL; {߂I}
        ' '      : GETCH;
        '{' : nextln; // ̍s̎c̓RgI(added by Takeda 2005.11.18)
                      // ꂪȂƂPڂ'{'ꂽƂ
                      // RpCG[ɂȂI
      else begin
        Error(6); if erra then exit;
      end; {else}
    end; {case}
  end;  {while}

  lineOffset:=preamble.Count + mainPPG.Count;
  asyncFlag:=false;

  for k:=0 to asyncPPGList.length-1 do begin
    if asyncPPGList.asyncPPG[k].channel=currentCH then begin

      if not asyncPPGList.asyncPPG[k].used then begin
      {asynchronousV[PX͌ĂяoĂȂ}
        Warnings.Add('[Warning] : An asynchronous sequence "'
                     + asyncPPGList.asyncPPG[k].name
                     +'" is not called by the main sequnece.');
      end; // if

      asyncPPGList.asyncPPG[k].address:=compiledPPG[currentCH-1].Count;
      {asyncPPGList.asyncPPG[k].addressɂ
       sasynchoronousV[PX̊JnAhXێ}
      asyncPPGList.itemIndex:=k;
      workingLines.Assign(asyncPPGList.asyncPPG[k].sequence);
      lnum:=0;
      EndOfFile:=false;
      nextln;
      asyncQ:=true;

      while not EndOfFile do begin
        if erra then exit;
          case CH of
            'A'..'Z','_' : SYMBOL; {߂I}
            ' '      : GETCH;
            '{'      : nextln; {̍s̎c̓RgI}
          else begin
            Error(6); if erra then exit;
          end; {else}
        end; {case}
      end; // while

      {ɖ[v̂߂̃WvR}hB
       [v̕AFPGAɂ点̂ŁA
       ̃vOŏKv͂ȂB}

      if asyncPPGList.asyncPPG[k].asyncMode=asContinue then begin

        compiledPPG[currentCH-1].Add(Hex(compiledPPG[currentCH-1].Count,3)
                                    + ' = B0 '
                                    + Hex(asyncPPGList.asyncPPG[k].address,10)
                                    + ' 00000000 00000000');

        {asyncPPGList.asyncPPG[k].addressɂ͂asynchoronousV[PX
         JnAhXێĂ}
        NOfLinesPerCommand[currentCH-1].Add('1');

        mnemonic[currentCH-1].Add('JUMP (0xB0) to '
                                 + Hex(asyncPPGList.asyncPPG[k].address,3));

      end else if asyncPPGList.asyncPPG[k].asyncMode=asStandby then begin
        compiledPPG[currentCH-1].Add(Hex(compiledPPG[currentCH-1].Count,3)
                                    + ' = B0 '
                                    + Hex(compiledPPG[currentCH-1].Count,10)
                                    + ' 00000000 00000000');
          // sɃWvBsyncVOi܂Ŗ[v
          // Fline̓b`ꂽ܂܂łI
        NOfLinesPerCommand[currentCH-1].Add('1');

        mnemonic[currentCH-1].Add('JUMP (0xB0) to '
                                 + Hex(compiledPPG[currentCH-1].Count-1,3));
                            // Count͂łɁ{PĂ̂ŁA}CiXPĕ\


      end;



       {
       ɁAasyncĂяoĂ郁C̃R}h
       m肷BƂƂ̓AhXsȂ̂ŁAł邵ȂI}

       for j:=0 to asyncPPGList.asyncPPG[k].Count-1 do begin

         compiledPPG[currentCH-1][asyncPPGList.asyncPPG[k].whereIs(j)]:=
           Hex(asyncPPGList.asyncPPG[k].whereIs(j),3)
           + ' = AA '
           + Hex(asyncPPGList.asyncPPG[k].address,10)
           + ' 00000000 00000000';
         mnemonic[currentCH-1][asyncPPGList.asyncPPG[k].whereIs(j)]:=
           'ASYNC (0xAA): jump to '
             + Hex(asyncPPGList.asyncPPG[k].address,3);
       end; // j

    end; // if
    lineOffset:=lineOffset+asyncPPGList.asyncPPG[k].sequence.Count;
  end; // k

 for ippp:=1 to 2 do
   begin
    if lp[ippp] then begin
     {[vJE^[ĂȂI}
      err_str:=IntToStr(ippp);
      error(22); exit
    end;
   end;
end; {ASSEMBLE}

(******************* Compile   ************************)
procedure TnmrPPG.compile;
var i,j,k,p,q: integer;
    blankCount: integer;
    found, foundStart, foundInit, foundRelax: boolean;
    r,s,t:string;
    asy: TAsyncPPG;
    asMode: TAsyncMode;
    sl: TStrings;
begin

  preamble.Clear;
  mainPPG.Clear;


  k:=0; foundStart:=false;
  {+++++  ppgLines'start'܂łmainPPGɃRs[  +++++}
  while (not foundStart) and (k<ppgLines.Count) do begin
    s:=ppgLines[k];
    preamble.Add(s);
    inc(k);
    if CompareText('start',Copy(TrimLeft(s),1,5))=0 then foundStart:=true;
  end; // while

  if foundStart=false then begin {'start'ȂI}
    error(198);
    exit;
  end;


  {+++++  ppgLines'init'܂łmainPPGɃRs[  +++++}
  foundInit:=false;
  while (not foundInit) and (k<ppgLines.Count) do begin
    s:=ppgLines[k];
    mainPPG.Add(s);
    inc(k);
    if CompareText('init',Copy(TrimLeft(s),1,4))=0 then foundInit:=true;
  end; // while

  if foundInit=false then begin {'init'ȂI}
    error(197);
    exit;
  end;

  {"Init"R}h̒O phase cycle|C^Zbg閽߂B}
  mainPPG.Insert(mainPPG.Count-1,
    'pulse(1u; F1PhasePointer_Init, F2PhasePointer_Init, F3PhasePointer_Init)');

  {+++++  ppgLines'relax'܂łmainPPGɃRs[  +++++}
  foundRelax:=false;
  while (not foundRelax) and (k<ppgLines.Count) do begin
    s:=ppgLines[k];
    mainPPG.Add(s);
    inc(k);
    if CompareText('relax',Copy(TrimLeft(s),1,5))=0 then foundRelax:=true;
  end; // while

  if foundRelax=false then begin {'relax'ȂI}
    error(199);
    exit;
  end;
  {+++++++++++++++++++++++++++++++++++++++++++++++++++++}

  blankCount:=0;  {G[s̊Ẑߋ󔒍sRgs̐𐔂}

  while k<ppgLines.Count do begin
    s:=ppgLines[k];
    if (s='') or (Copy(TrimLeft(s),1,1)='{') then begin inc(blankCount);
    end
    else if CompareText('async',Copy(TrimLeft(s),1,5))=0 then begin
       { async̓ǂݍ }
       t:=TrimLeft(Copy(TrimLeft(s),6,Length(TrimLeft(s))));

       { t󂩂ǂH }
       if t='' then begin
         lnum:=k+1;   {G[sk+1}
         err_str:=IntToStr(k+1); error(204); exit;
       end;

       { (邩`FbN}
       if t[1]<>'(' then begin
         lnum:=k+1;   {G[sk+1}
         err_str:=IntToStr(k+1); error(201); exit;
       end;

       q:=2; r:=''; found:=false;
       while q<=Length(t) do begin
         if t[q]=')' then begin found:=true; break; end;
         r:=r+UpperCase(t[q]);
         inc(q);
       end;
       if not found then begin
         lnum:=k+1;   {G[sk+1}
         err_str:=IntToStr(k+1); error(202); exit;
       end;
       r:=Trim(r); {ÕXy[X}

       inc(q);
       while q<=Length(t) do begin
         if t[q]='{' then break
         else if t[q]<>' ' then begin
           lnum:=k+1;   {G[sk+1}
           err_str:=t[q] + ' at line '+ IntToStr(k+1);
           error(203); exit;
         end;
         inc(q);
       end;

       q:=Pos(';',r);
       if (q=0) or (q=1) or (q=Length(r)) or (Length(r)<3) then begin
       { ";"AŏɗĂŌɗĂς}
         lnum:=k+1;   {G[sk+1}
         err_str:=IntToStr(k+1);
         error(204);exit;
       end;



       for i:=1 to Length(Trim(Copy(r,1,q-1))) do begin
         if not CharInSet(Trim(Copy(r,1,q-1))[i], ['0'..'9']) then begin {ȂI}
           lnum:=k+1;   {G[sk+1}
           error(210);
           exit;
         end;
       end;


       p:=StrToInt(Trim(Copy(r,1,q-1)));  {chʔԍ}



       if (p<>2) and (p<>3) then begin  {chʔԍ2 or 3łȂĂ͂ȂȂ}
         lnum:=k+1;   {G[sk+1}
         err_str:=IntToStr(k+1);
         error(205);exit;
       end;

       t:=Trim(Copy(r,q+1,Length(r)-q+1));
       if not CharInSet(t[1], ['A'..'Z','_']) then begin
         lnum:=k+1;   {G[sk+1}
         err_str:=IntToStr(k+1);
         error(208);
         exit;
       end;

       for q:=1 to Length(t) do begin
         if not CharInSet(t[q], ['A'..'Z','_',' ','0'..'9']) then begin
           lnum:=k+1;   {G[sk+1}
           err_str:=IntToStr(k+1);
           error(209);
           exit;
         end;
       end;

       {̎_łɃ`lԍAtɎʖZbg}

       {d``FbN}
       if asyncPPGList.index(t)>-1 then begin
         lnum:=k+1;   {G[sk+1}
         err_str:=t;
         error(207);
         exit;
       end;


       {A"end async" Ƃs܂ŕRs[}

       sl:=TStringList.Create;

       for q:=1 to blankCount+1 do begin sl.Add(''); end;
       {G[sô߁A󔒍sĂ}
       {+1́Aasync(?;??)R}h}

       found:=false;
       inc(k);
       while k<ppgLines.Count do begin
         s:=Trim(UpperCase(ppgLines[k]));
         if Copy(s,1,8)='CONTINUE' then begin
           found:=true; asMode:=asContinue; inc(k); break;
         end else if Copy(s,1,7)='STANDBY' then begin
           found:=true; asMode:=asStandby; inc(k); break;
         end else begin
           sl.Add(s);
           inc(k);
         end;
       end;
       sl.Add('');
       {continueR}h̃uNsĂ}

       if found=false then begin
         lnum:=k;
         err_str:=t;
         sl.Free;
         error(206);
         exit;
       end;

       asy:=TAsyncPPG.Create;
       asy.channel:=p;
       asy.name:=t;
       asy.asyncMode:=asMode;
       asy.sequence.Assign(sl);
       sl.Free;


       asyncPPGList.add(asy);
       asy.Free;

       blankCount:=1; {asynĉ߂ɃZbgĂ}

    end else if s<>#0 then begin {syntax error}
      err_str:=IntToStr(k+1);
      error(200);
      exit;
    end;
    inc(k);
  end; // while


  info.Clear;
  messages.Clear;
  Warnings.Clear;

  for j:=0 to NOfChannels-1 do begin
    compiledPPG[j].Clear;
    mnemonic[j].Clear;
  end;

  auxParams.definePD:=false;
  auxParams.defineNA:=false;
  auxParams.defineAL:=false;
  auxParams.defineDW:=false;

  acqPhaseDefined:=false;


  lineOffset:=0;  {G[݂s̊ZɎg}

  pathDefined:=false; {Ƃ肠Q[g`t@C̃pX͐ݒ肵ĂȂ}
  gateDefFileName:='';       {ƂĂ}

  workingLines.Assign(preamble);   {vAuRs[}
  lnum:=0;
  Initial;                         {}
  if erra then exit;

  currentCH:=1;
  repeat
    workingLines.Clear;




    //phListArray.gen; // ͋o[W̃R[hB
    phListArray.gen(currentCH);
      // New in Version 0.90.
      // A`lԂœƂKvB

    workingLines.Assign(phListArray.source);


    if currentCH=1 then begin // acq. phaseݒ肷̂CH1̂݁B
      acqPhaseList.gen;
      for j:=0 to acqPhaseList.source.Count-1 do begin
        workingLines.Add(acqPhaseList.source[j]);
      end; // j
    end;




    workingLines.Add('READY');  //Ń`lԂ̓ƂB





    phListArrayOffset:=phListArray.source.Count
                      + acqPhaseList.source.Count;

    for j:=0 to mainPPG.Count-1 do begin
      workingLines.Add(mainPPG[j]);    {CRs[}
    end;
    lnum:=0;
    ASSEMBLE;                        {}
    if compiledPPG[currentCH-1].Count > 2048 then begin
      err_str:=IntToStr(compiledPPG[currentCH-1].Count);
      error(7);
      exit;
    end;
    if erra then exit;
    inc(currentCH);
  until currentCH>NOfChannels;

  compiled:=true;
  // SăRpC肭΂܂ŗB
  // compiled trueɂB

end;
{-------------------------------------------------}
procedure TnmrPPG.updateVariable(variableName:string; num:double; uni:TUnitOfTime);
// Time variables only
var q:integer;
    us: string;
begin
  erra:=false;
  messages.Clear;

  if not compiled then begin
    erra:=true;
    Messages.Add('Pulse program has not been compiled yet.');
    exit;
  end;

  q:=Variables.index(variableName);
  if q=-1 then begin  // ϐȂIH
    erra:=true;
    Messages.Add('Cannot find the variable '+variableName+'.');
    exit;
  end;

  case Variables.Variable[q].typeOfVar of
    timeVariable: begin
                    updateTimeVariable(q,num,uni);
                    if erra=true then exit;
                    us:=Variables.Variable[q].unitOfTimeString;
                  end;
    else begin
      erra:=true;
      Messages.Add('updateVariable(variableName:string; '
                   +'num:double; uni:TUnitOfTime) is valid only '
                   +'for time variables.');
    end; //

  end; //case;


  Messages.Add(Variables.Variable[q].name
           + ' has been modified to '
           + FloatToStr(Variables.Variable[q].Numeric)
           + Variables.Variable[q].unitOfTimeString
           + '.' );

end;
//------------------------------------------------------------------------------
procedure TnmrPPG.updateVariable(variableName:string; num:double);
var q:integer;
    us: string;
begin
  erra:=false;
  messages.Clear;

  if not compiled then begin
    erra:=true;
    Messages.Add('Pulse program has not been compiled yet.');
    exit;
  end;

  q:=Variables.index(variableName);
  if q=-1 then begin  // ϐȂIH
    erra:=true;
    Messages.Add('Cannot find the variable '+variableName+'.');
    exit;
  end;

  case Variables.Variable[q].typeOfVar of
    timeVariable: begin
                    updateTimeVariable(q,num,microSeconds);
                    if erra=true then exit;
                    us:=Variables.Variable[q].unitOfTimeString;
                  end;
    loopVariable: begin
                    updateLoopVariable(q,Round(num));
                    if erra=true then exit;
                    us:='';
                  end;

    NAVariable: begin
                    updateLoopVariable(q,Round(num));
                    if erra=true then exit;
                    us:='';
                  end;

    intVariable: begin
                   updateIntVariable(q,Round(num));
                   if erra=true then exit;
                   us:='';
                  end;

    ampVariable: begin
                   updateAmpVariable(q,num);
                   if erra=true then exit;
                   us:='';
                  end;
    phaseVariable: begin
                   updatePhaseVariable(q,num);
                   if erra=true then exit;
                   us:='';
                  end;

    freqVariable: begin
                    updateFreqVariable(q,num);
                    if erra=true then exit;
                    us:='';
                  end;

    else begin
      erra:=true;
      Messages.Add('updateVariable(variableName:string; '
                   +'num:double; uni:TUnitOfTime) is valid only '
                   +'for time variables.');
    end; //


  end; //case;


  Messages.Add(Variables.Variable[q].name
           + ' has been modified to '
           + FloatToStr(Variables.Variable[q].Numeric)
           + Variables.Variable[q].unitOfTimeString
           + '.' );
end;
{-------------------------------------------------}
procedure TnmrPPG.updateTimeVariable(q:integer; d: double; ut:TunitOfTime);
var mul:double;
    a,b,f: double;
    p,m,ad,q2:integer;
    s,s1,s2:string;
    UWidth: word;
    LWidth: cardinal;
    nAD9858: integer;
begin
  mul:=1;

  case ut of
    nanoSeconds: begin
      mul:=1E-3;
      Variables.Variable[q].unitOfTimeString:='n';
    end;

    microSeconds: begin
      mul:=1;
      Variables.Variable[q].unitOfTimeString:='u';
    end;

    milliSeconds: begin
      mul:=1E3;
      Variables.Variable[q].unitOfTimeString:='m';
    end;

    Seconds: begin
      mul:=1E6;
      Variables.Variable[q].unitOfTimeString:='s';
    end;
  end; //case

  f:=d*mul;

  if (f<Variables.Variable[q].Min) or (f>Variables.Variable[q].Max)
   then begin
     erra:=true;
     Messages.Add('Out of range for variable ' + Variables.Variable[q].Name
                  + '.');
     exit;
  end;

  if Variables.Variable[q].Name='DW' then begin
    if (f<0.025) or (f>1000) then begin
      erra:=true;
      Messages.Add('Out of range for DW (25ns to 1 ms).');
      exit;
    end else begin
      f:=Round(f/0.025)*0.025;  // ۂ߁I
//      MainForm.VariableInfoMemo.Lines.Add('DW has been modified, and accordingly, ');
      Variables.Variable[q].Value:=f;
      Variables.Variable[q].Numeric:=f/mul;   //  d = (f/mul)
      Variables.Variable[q].unitOfTime:=ut;
      Variables.Variable[q].updated:=true;
      auxParams.DW:=f;
        {ŎIauxParams.ACQTIMEAbvf[g}
      q2:=Variables.index('ACQTIME');
      Variables.Variable[q2].Value:=auxParams.ACQTIME; // us
      mul:=1E3;
      Variables.Variable[q2].Numeric:=auxParams.ACQTIME/mul;  // ms
      Variables.Variable[q2].unitOfTime:=milliSeconds;
      Variables.Variable[q2].unitOfTimeString:='m';

//      if digitalFilterOption=dfAuto then begin
//        autosetDigitalFilter;
//      end;  // if digitalFilterOption

      {-- ACQTIMEupdate --}
      f:=auxParams.ACQTIME;
      f:=Round(CLK*f)/CLK;  // ۂ߁I
      if (CLK*f-4) > 4294967295 {2^32-1} then begin
          {CLK[MHz] * f[us] --> number of count}
          {4 --> excess 4 clocks}
        a:=Trunc((CLK*f-4)/4294967296);   // 2^32Ŋ
        b:=(CLK*f-4) - (a*4294967296);
      end else begin
        a:=0;
        b:=CLK*f-4;
      end;

      Uwidth:=Trunc(a);
      LWidth:=Trunc(b);

      updateCommand.Clear;

      for p:=0 to nOfChannels-1 do begin   // number of channels (3)
        updateCommand.Add('CH'+IntToStr(p+1));

        for m:=0 to Variables.Variable[q2].Count(p)-1 do begin
          ad:=Variables.Variable[q2].whereIs(p,m);
          s:=compiledPPG[p][ad];
          s1:=Copy(s,1,9);
          s2:=Copy(s,20,18);
          compiledPPG[p][ad]:=s1+ Hex(UWidth,2) + Hex(LWidth,8) + s2;
          updateCommand.Add(compiledPPG[p][ad]);
        end;

      end; // p

      updateCommand.Add('CH1');

    end;

  end else begin {DWłȂ}

  f:=Round(CLK*f)/CLK;  // ۂ߁I
  d:=f/mul;
  Variables.Variable[q].Numeric:=d;
  Variables.Variable[q].unitOfTime:=ut;
  Variables.Variable[q].Value:=f;
  Variables.Variable[q].updated:=true;

  if (CLK*f-4<0) then begin
     erra:=true;
     Messages.Add('Out of Range.');
     exit;
  end;



  updateCommand.Clear;

  for p:=0 to nOfChannels-1 do begin   // number of channels (3)
    updateCommand.Add('CH'+IntToStr(p+1));

    for m:=0 to Variables.Variable[q].Count(p)-1 do begin
      ad:=Variables.Variable[q].whereIs(p,m);

      nAD9858:=StrToInt(NOfLinesPerCommand[p][ad]);
      if nAD9858=5 then begin
        if f*CLK-40<4 then begin
          erra:=true;
          Messages.Add('Out of range (this line uses frequency setting, and at least 44 clock counts are required).');
          exit;
        end;

        Jikann_Hennkann(f-(40.0/CLK), UWidth,LWidth);
        s:=compiledPPG[p][ad+4];
        s1:=Copy(s,1,9);
        s2:=Copy(s,20,18);
        compiledPPG[p][ad+4]:=s1+ Hex(UWidth,2) + Hex(LWidth,8) + s2;
        updateCommand.Add(compiledPPG[p][ad+4]);
      end else begin
        Jikann_Hennkann(f, UWidth,LWidth);
        s:=compiledPPG[p][ad];
        s1:=Copy(s,1,9);
        s2:=Copy(s,20,18);
        compiledPPG[p][ad]:=s1+ Hex(UWidth,2) + Hex(LWidth,8) + s2;
        updateCommand.Add(compiledPPG[p][ad]);
      end;
    end;

  end; // p
  updateCommand.Add('CH1');

  end; // {DWłȂ}

end;
{-------------------------------------------------}
procedure TnmrPPG.updateLoopVariable(q:integer; d: integer);
var
    k,m,ad: integer;
    s,s1,s2:string;
begin
  {TODO: ͈̓`FbN}

    if (d<Variables.Variable[q].Min) or (d>Variables.Variable[q].Max)
      then begin
       erra:=true;
       Messages.Add('Out of Range for loop variable.');
       exit;
    end;

    Variables.Variable[q].Numeric:=d;
    Variables.Variable[q].Value:=d;
    Variables.Variable[q].updated:=true;

    updateCommand.Clear;

    for k:=0 to nOfChannels-1 do begin   // nch: number of channels (3)
      updateCommand.Add('CH'+IntToStr(k+1));

      for m:=0 to Variables.Variable[q].Count(k)-1 do begin
        ad:=Variables.Variable[q].whereIs(k,m);
        s:=compiledPPG[k][ad];
        s1:=Copy(s,1,9);
        s2:=Copy(s,20,18);
        updateCommand.Add(s1+ Hex(d,10) + s2);
      end; // m
    end; // k
    updateCommand.Add('CH1');
    updateCommand.Add('');


end;
{-------------------------------------------------}
procedure TnmrPPG.updateIntVariable(q:integer; d: integer);
var z,u,m,ad,gi,k:integer;
    q2:integer;
    kAd9858,nAD9858:integer;
    t,s,s1,s2,s3,ns:string;
    tbs:TlineBitsSet;
    ULine,LLine:cardinal;
    mul:double;
    f,a,b:double;
    p:integer;
    UWidth: shortInt;
    LWidth: LongInt;
begin

  if Variables.Variable[q].Name='AL' then begin
    if (d<1) or (d>16384) then begin
      erra:=true;
      Messages.Add('Out of Range. AL should be an integer from 1 to 16384.');
      exit;
    end else begin
//      MainForm.VariableInfoMemo.Lines.Add('AL has been modified, and accordingly, ');
      Variables.Variable[q].Value:=d;
      Variables.Variable[q].Numeric:=d;
      Variables.Variable[q].updated:=true;
      auxParams.AL:=d;
        {ŎIauxParams.ACQTIMEAbvf[g}
      q2:=Variables.index('ACQTIME');
      Variables.Variable[q2].Value:=auxParams.ACQTIME; // us
      mul:=1E3;
      Variables.Variable[q2].Numeric:=auxParams.ACQTIME/mul;  // ms
      Variables.Variable[q2].unitOfTime:=milliSeconds;
      Variables.Variable[q2].unitOfTimeString:='m';


      {-- ACQTIMEupdate --}
      f:=auxParams.ACQTIME;
      f:=Round(CLK*f)/CLK;  // ۂ߁I
      if (CLK*f-4) > 4294967295 {2^32-1} then begin
          {CLK[MHz] * f[us] --> number of count}
          {4 --> excess 4 clocks}
        a:=Trunc((CLK*f-4)/4294967296);   // 2^32Ŋ
        b:=(CLK*f-4) - (a*4294967296);
      end else begin
        a:=0;
        b:=CLK*f-4;
      end;

      Uwidth:=Trunc(a);
      LWidth:=Trunc(b);

      updateCommand.Clear;

      for p:=0 to nOfChannels-1 do begin   // number of channels (3)
        updateCommand.Add('CH'+IntToStr(p+1));

        for m:=0 to Variables.Variable[q2].Count(p)-1 do begin
          ad:=Variables.Variable[q2].whereIs(p,m);
          s:=compiledPPG[p][ad];
          s1:=Copy(s,1,9);
          s2:=Copy(s,20,18);
          compiledPPG[p][ad]:=s1+ Hex(UWidth,2) + Hex(LWidth,8) + s2;
          updateCommand.Add(compiledPPG[p][ad]);

        end;

      end; // p

      updateCommand.Add('CH1');



    end;
  end else begin  //ALłȂ

    // ͈̓`FbN
    if (d<Variables.Variable[q].Min) or (d>Variables.Variable[q].Max)
    then begin
      erra:=true;
      messages.add('Out of Range.');
      exit;
    end;

    Variables.Variable[q].Numeric:=d;
    Variables.Variable[q].Value:=d;
    Variables.Variable[q].updated:=true;

    updateCommand.Clear;


    for u:=0 to nOfChannels-1 do begin

      if Variables.Variable[q].Count(u)>0 then begin
        ns:=IntToStr(u+1);
        updateCommand.Add('CH'+ns);

        for m:=0 to Variables.Variable[q].Count(u)-1 do begin
          ad:=Variables.Variable[q].whereIs(u,m);
          gi:=Variables.Variable[q].gateIndex(u,m);


          nAD9858:=StrToInt(NOfLinesPerCommand[u][ad]);
          for kAD9858 := 0 to nAD9858-1 do begin
               //
               // Abvf[gׂӏɎg䂪gĂꍇ̏
               //
            s:=compiledPPG[u][ad+kAD9858];
            s1:=Copy(s,1,20);
            s2:=Copy(s,21,8); //  63..32
            s3:=Copy(s,30,8); //  31..0

            tbs:=extractTLineBitsSet(s2+s3);

            for k:=0 to gates.gate[gi].Length-1 do begin
              z:=gates.gate[gi].getAssignment(k);
              if z in tbs then tbs:=tbs-[z];
            end;

            tbs:=tbs+gates.gate[gi].output(d);

            Lline:=0; ULine:=0;
            for k:=0 to 31 do begin
              if k in tbs then Lline:=Lline or (1 shl k);
              if k+32 in tbs then Uline:=Uline or (1 shl k);
            end;

            t := s1 + Hex(ULine,8) + ' ' + Hex(LLine,8);
            compiledPPG[u][ad+kAD9858]:=t; // RpCꂽpXvOAbvf[g
            updateCommand.Add(t);

          end; // kAD9858


        end; // m;
      end; // if count>0
    end; // u

    updateCommand.Add('');

  end;  //ALłȂ

end;
{-------------------------------------------------}
procedure TnmrPPG.updateAmpVariable(q:integer; d: double);
var p,z,k,m,ad: integer;
    kAD9858,nAD9858:integer;
    s,s1,s2,s3,t,ns:string;
    tbs:TlineBitsSet;
    indexOfampgate:word;
    ULine,LLine:cardinal;
begin
  // ͈̓`FbN
  if (d<Variables.Variable[q].Min) or (d>Variables.Variable[q].Max)
  then begin
    erra:=true;
    Messages.Add('Out of Range.');
    exit;
  end;

  Variables.Variable[q].Numeric:=d;
  Variables.Variable[q].Value:=d;
  Variables.Variable[q].updated:=true;

  updateCommand.Clear;

 for p:=0 to nOfChannels-1 do begin

 if Variables.Variable[q].Count(p)>0 then begin
  ns:=IntToStr(p+1);
  updateCommand.Add('CH'+ns);

  for m:=0 to Variables.Variable[q].Count(p)-1 do begin
    ad:=Variables.Variable[q].whereIs(p,m);

    nAD9858:=StrToInt(NOfLinesPerCommand[p][ad]);
    for kAD9858 := 0 to nAD9858-1 do begin
          //
          // Abvf[gׂӏɎg䂪gĂꍇ̏
          //

      s:=compiledPPG[p][ad+kAD9858];
      s1:=Copy(s,1,20);
      s2:=Copy(s,21,8); //  63..32
      s3:=Copy(s,30,8); //  31..0

//      indexOfAmpGate:=gates.index('F'+ns+'AMP');
//       ꂾƃQ[g`t@CˑĂ܂̂ŁAȉ̂悤ɏC
//       2010.7.15
      indexOfAmpGate:=Variables.Variable[q].gateIndex(p,m);

      tbs:=extractTLineBitsSet(s2+s3);

      for k:=0 to gates.gate[indexOfAmpGate].Length-1 do begin
        z:=gates.gate[indexOfAmpGate].getAssignment(k);
        if z in tbs then tbs:=tbs-[z];
      end;

      tbs:=tbs+gates.gate[indexOfAmpGate].output(d);

      Lline:=0; ULine:=0;
      for k:=0 to 31 do begin
        if k in tbs then Lline:=Lline or (1 shl k);
        if k+32 in tbs then Uline:=Uline or (1 shl k);
      end;

      t := s1 + Hex(ULine,8) + ' ' + Hex(LLine,8);
      compiledPPG[p][ad+kAD9858]:=t; // RpCꂽpXvOAbvf[g
      updateCommand.Add(t);
    end; // kAD9858

  end; // m;
  end; // if count>0
 end; //p

  updateCommand.Add('');


end;
{-------------------------------------------------}
procedure TnmrPPG.updatePhaseVariable(q:integer; d: double);
var p,z,k,m,ad: integer;
    kAD9858,nAD9858: integer;
    s,s1,s2,s3,t,ns:string;
    tbs:TlineBitsSet;
    indexOfphasegate:word;
    ULine,LLine:cardinal;
begin
  // ʑ̂Ƃ͔͈̓`FbNȂB

  Variables.Variable[q].Numeric:=d;
  Variables.Variable[q].Value:=d;
  Variables.Variable[q].updated:=true;

  updateCommand.Clear;

 for p:=0 to nOfChannels-1 do begin

 if Variables.Variable[q].Count(p)>0 then begin
  ns:=IntToStr(p+1);
  updateCommand.Add('CH'+ns);

  for m:=0 to Variables.Variable[q].Count(p)-1 do begin
    ad:=Variables.Variable[q].whereIs(p,m);

    nAD9858:=StrToInt(NOfLinesPerCommand[p][ad]);
    for kAD9858 := 0 to nAD9858-1 do begin
          //
          // Abvf[gׂӏɎg䂪gĂꍇ̏
          //

      s:=compiledPPG[p][ad+kAD9858];
      s1:=Copy(s,1,20);
      s2:=Copy(s,21,8); //  63..32
      s3:=Copy(s,30,8); //  31..0


//      indexOfPhaseGate:=gates.index('F'+ns+'PHASE');
//       ꂾƃQ[g`t@CˑĂ܂̂ŁAȉ̂悤ɏC
//       2010.7.15
      indexOfPhaseGate:=Variables.Variable[q].gateIndex(p,m);


      tbs:=extractTLineBitsSet(s2+s3);

      for k:=0 to gates.gate[indexOfPhaseGate].Length-1 do begin
        z:=gates.gate[indexOfPhaseGate].getAssignment(k);
        if z in tbs then tbs:=tbs-[z];
      end;

      tbs:=tbs+gates.gate[indexOfPhaseGate].output(d);

      Lline:=0; ULine:=0;
      for k:=0 to 31 do begin
        if k in tbs then Lline:=Lline or (1 shl k);
        if k+32 in tbs then Uline:=Uline or (1 shl k);
      end;

      t := s1 + Hex(ULine,8) + ' ' + Hex(LLine,8);
      compiledPPG[p][ad+kAD9858]:=t; // RpCꂽpXvOAbvf[g
      updateCommand.Add(t);
   end; // kAD9858

  end; // m;
  end; // if count>0
 end; //p

  updateCommand.Add('');


end;
{-------------------------------------------------}
procedure TnmrPPG.updateFreqVariable(q:integer; d: double);
var p,z,k,m,ad: integer;
    s,s1,s2,s3,t,ns:string;
    tbs:TlineBitsSet;
    indexOfAD9858Gate:word;
    ULine,LLine:cardinal;
    kAD9858: integer;
    psAD9858: shortInt;
    od:double;
begin
  // freq͈̓`FbNȂB

  Variables.Variable[q].Numeric:=d;
  Variables.Variable[q].Value:=d;
  Variables.Variable[q].updated:=true;

  updateCommand.Clear;

 for p:=0 to nOfChannels-1 do begin

 if Variables.Variable[q].Count(p)>0 then begin
  ns:=IntToStr(p+1);
  updateCommand.Add('CH'+ns);

  for m:=0 to Variables.Variable[q].Count(p)-1 do begin
      ad:=Variables.Variable[q].whereIs(p,m);

      psAD9858:=variables.Variable[q].psAD9858(p,m);

      //
      // *********
      //
      indexOfAD9858Gate:=Variables.Variable[q].gateIndex(p,m);
      for kAD9858:= 1 to 5 do begin

        s:=compiledPPG[p][ad+kAD9858-1];
        s1:=Copy(s,1,20);
        s2:=Copy(s,21,8); //  63..32
        s3:=Copy(s,30,8); //  31..0

        tbs:=extractTLineBitsSet(s2+s3);

        for k:=0 to gates.gate[indexOfAD9858Gate].Length-1 do begin
          z:=gates.gate[indexOfAD9858Gate].getAssignment(k);
          if z in tbs then tbs:=tbs-[z];
        end;


            if psAD9858=-1 then
              od:=d*1e6
              {ps=-1Ȃsweep modeAItZbgȂMHzP}
            else
              od:=(d + gates.gate[indexOfAD9858Gate].offset)*1e6;
              {ItZbgāAMHzHzɕϊ}

        tbs:=tbs+gates.gate[indexOfAD9858Gate].output(psAD9858,od,kAD9858);

        Lline:=0; ULine:=0;
        for k:=0 to 31 do begin
          if k in tbs then Lline:=Lline or (1 shl k);
          if k+32 in tbs then Uline:=Uline or (1 shl k);
        end;

        t := s1 + Hex(ULine,8) + ' ' + Hex(LLine,8);
        compiledPPG[p][ad+kAD9858-1]:=t; // RpCꂽpXvOAbvf[g    ******
        updateCommand.Add(t);

      end; // kAD9858

      //
      //***********
      //

  end; // m;
  end; // if count>0
 end; //p

  updateCommand.Add('');

end;

{-------------------------------------------------}
{-------------------------------------------------}
{-------------------------------------------------}
{-------------------------------------------------}
{-------------------------------------------------}

(******************************************************)
{-------------------------------------------------}
{-------------------------------------------------}
{-------------------------------------------------}
{-------------------------------------------------}

{-------------------------------------------------}
constructor TVariables.Create(nch:cardinal);
begin
  FCount:=0;
  FNOfChannels:=nch;
  FNOfArray:=0;

  arrayCounter:=TarrayCounter.Create;

  SetLength(arrayedVariableIndex,0);

end;
{-------------------------------------------------}
destructor TVariables.Free;
begin
  while (FCount>0) do begin
    Variable[FCount-1].Free;
    dec(Fcount);
  end;

  arrayCounter.Free;

end;
{-------------------------------------------------}
function TVariables.defined(s:string):boolean;
var k:integer;
    find:boolean;
begin
  find:=false;
  for k:=0 to count-1 do begin
    if Variable[k].Name=s then find:=true;
  end; // k
  result:=find;
end;
{-------------------------------------------------}
function TVariables.index(s:string):integer;
var k,q:integer;
begin
  q:=-1;
  for k:=0 to count-1 do begin
    if Variable[k].Name=s then q:=k;
  end; // k
  result:=q;
end;
{-------------------------------------------------}
procedure TVariables.setArrayLength(k: Integer);
var q:integer;
begin
  FArrayLength:=k;
  for q:=0 to count-1 do begin
    Variable[q].arrayLength:=FArrayLength;
  end;
  // ꂾƂׂĂ̕ϐAarray̗Lɂ炸mۂĂ܂B
  //@܂IH
  // 2011.6.28
end;

{-------------------------------------------------}
procedure TVariables.setNOfArray(k:integer);
begin
  FNOfArray:=k;
end;
{-------------------------------------------------}
procedure TVariables.add(s:string);
begin
  if not defined(s) then begin
    inc(FCount);
    SetLength(Variable, FCount);
    Variable[FCount-1]:=TVariable.Create(FNOfChannels,s);
    Variable[FCount-1].arrayLength:=FarrayLength;
    SetLength(Variable[FCount-1].ArrayedValues,Variable[FCount-1].arrayLength);
  end;
end;

procedure TVariables.add(s:string; t:TType);
begin
  add(s);
  Variable[FCount-1].typeOfVar:=t;
  case t of
    timeVariable: begin
        Variable[FCount-1].FMin:=0.025;
        Variable[FCount-1].FMax:=1E10;
      end;
    loopVariable: begin
        Variable[FCount-1].FMin:=1;
        Variable[FCount-1].FMax:=65535;
      end;
    NAVariable: begin
        Variable[FCount-1].FMin:=1;
        Variable[FCount-1].FMax:=4294967295;  // 2^32-1
      end;
    ampVariable: begin
        Variable[FCount-1].FMin:=0;
        Variable[FCount-1].FMax:=100.0;
      end;
    intVariable: begin
        Variable[FCount-1].FMin:=-4294967295 div 2;
        Variable[FCount-1].FMax:=4294967295 div 2;
      end;
    phaseVariable: begin
        // ʑϐ̏ꍇA
        // Ƃ肠őŏl̐ݒ͂ȂB
        //
      end;
    freqVariable: begin
        //
        // Ƃ肠őŏl̐ݒ͂ȂB
        // sweep modeŕ̒lgƂ邵A[U[̐ӔCŐݒB

    end;
  end; // case
end;
{-------------------------------------------------}
procedure TVariables.delete(k:integer);
var i:integer;
begin

  for i:=k to count-2 do begin
    Variable[i]:=Variable[i+1];
  end; // i

  Variable[FCount-1].Free;
  dec(FCount);

end;
{-------------------------------------------------}
procedure TVariables.delete;
begin
  Free;
end;


{+++++++++++++++++++++++++++++++++++++++++++++++++}
{+++++++++++++++++++++++++++++++++++++++++++++++++}
{+++++++++++++++++++++++++++++++++++++++++++++++++}
{-------------------------------------------------}
constructor TVariable.Create(nch:cardinal);
var k:integer;
begin
  FarrayQ:=false;
  FNOfChannels:=nch;
  SetLength(FNOfWhereIs,FNOfChannels);
  SetLength(FWhereIs,FNOfChannels);
  SetLength(FGateIndex,FNOfChannels);
  SetLength(FPSAD9858,FNOfChannels);
  for k:=0 to FNOfChannels-1 do begin
    FNOfWhereIs[k]:=0;
    SetLength(FWhereIs[k],0);
    SetLength(FGateIndex[k],0);
    SetLength(FPSAD9858[k],0);
  end;
  SetLength(ArrayedValues,1);
  FCurrentArrayIndex:=0;
  KindOfArray:=Fixed;
  updated:=false;
end;

constructor TVariable.Create(nch:cardinal;s:string);
begin
  Create(nch);
  Name := s;
end;

{-------------------------------------------------}
function TVariable.whereIs(k,m:integer):integer;
begin
  if (m+1)>FNOfWhereIs[k] then whereIs:=-1
  else whereIs:=FWhereIs[k][m];
end;
{-------------------------------------------------}
function TVariable.gateIndex(k,m:integer):integer;
begin
  if (m+1)>FNOfWhereIS[k] then gateIndex:=-1
  else gateIndex:=FGateIndex[k][m];
end;
{-------------------------------------------------}
function TVariable.psAD9858(k,m:integer):shortInt;
begin
  //if (m+1)>FNOfWhereIS[k] then psAD9858:=-1
  //   -1肦邵A`FbNȗI
    psAD9858:=FPSAD9858[k][m];
end;
{-------------------------------------------------}
function TVariable.Count(k:integer):integer;
begin
  result:=FNOfWhereIs[k];
end;
{-------------------------------------------------}
procedure TVariable.setArrayLength(k: Integer);
begin
  FArrayLength:=k;
  setLength(arrayedValues,FArrayLength);
end;
{-------------------------------------------------}
procedure TVariable.setArrayQ(b: Boolean);
begin
  FArrayQ:=b;
  if arrayQ then begin
    setLength(arrayedValues, arrayLength);
  end;
end;
{-------------------------------------------------}
constructor TAsyncPPG.Create;
begin
  sequence:=TSTringList.Create;
  FNOfWhereIS:=0;
  SetLength(FWhereIs,0);
  address:=0;
  used:=false;
  asyncMode:=asContinue;
  // type TAsyncMode=(asContinue, asStandby);
end;
{-------------------------------------------------}
destructor TAsyncPPG.Free;
begin
  sequence.Free;
end;
{-------------------------------------------------}
function TAsyncPPG.whereIs(k:integer):integer;
begin
  if (k+1)>FNOfWhereIs then whereIs:=-1
  else whereIs:=FWhereIs[k];
end;
{-------------------------------------------------}
constructor TAsyncPPGList.Create;
begin
  length:=0;
  SetLength(AsyncPPG,length);
end;
{-------------------------------------------------}
procedure TAsyncPPGList.add(a:TAsyncPPG);
{ł͖Ȍd`FbN͂ȂBĂяoŋCĂ}
begin
  inc(length);
  SetLength(AsyncPPG,length);
  AsyncPPG[length-1]:=TAsyncPPG.Create;
  AsyncPPG[length-1].name:=a.name;
  AsyncPPG[length-1].channel:=a.channel;
  AsyncPPG[length-1].asyncMode:=a.asyncMode;
  AsyncPPG[length-1].sequence.Assign(a.sequence);
end;
{-------------------------------------------------}
procedure TAsyncPPGList.delete;
var i:integer;
begin
  for i:=0 to length-1 do begin
    AsyncPPG[i].Free;
  end;
end;
{-------------------------------------------------}
function TAsyncPPGList.index(name:string):integer;
var i,q:integer;
begin
  q:=-1;
  for i:=0 to length-1 do begin
    if name=AsyncPPG[i].name then q:=i;
  end; //i
  result:=q;
end;
{-------------------------------------------------}
destructor TAsyncPPGList.Free;
begin
  delete;
end;
{-------------------------------------------------}
{-------------------------------------------------}
{-------------------------------------------------}
end.
